/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 * Portions Copyright 2013 Jean-Pierre Andre
 */

/*
 * Copyright (c) 2003
 *	Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/dditypes.h>
#include <sys/stream.h>
#include <sys/termio.h>
#include <sys/thread.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/cmn_err.h>
#include <sys/stropts.h>
#include <sys/strtty.h>
#include <sys/kbio.h>
#include <sys/cred.h>
#include <sys/stat.h>
#include <sys/consdev.h>
#include <sys/kmem.h>
#include <sys/modctl.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/pci.h>
#include <sys/errno.h>
#include <sys/gld.h>
#include <sys/dlpi.h>
#include <sys/dls.h>
#include <sys/ethernet.h>

#include <inet/common.h>
#include <inet/nd.h>
#include <inet/mi.h>
#include <sys/strsun.h>

#include "queue.h"
#include "pe_var.h"
#include "resource_var.h"
#include "ntoskrnl_var.h"
#include "ndis_var.h"
#include "hal_var.h"
#include "cfg_var.h"
#include "if_ndisvar.h"
#include <inet/wifi_ioctl.h>

char _depends_on[] = "misc/ndisapi misc/net80211 misc/mac ip";

#ifndef LINKEDIMAGE
#define NDIS_IMAGE
#else
extern unsigned char drv_data[];
extern uint32_t drv_data_len;
#endif
#define NDIS_REGVALS

#include "ndis.h"

struct ndis_pci_type ndis_devs[] = {
#ifdef NDIS_PCI_DEV_TABLE
	NDIS_PCI_DEV_TABLE
#endif
	{ 0, 0, 0, NULL }
};

/* 0 - 30 dBm to mW conversion table */
static const uint16_t dBm2mW[] = {
	1, 1, 1, 1, 2, 2, 2, 2, 3, 3,
	3, 4, 4, 4, 5, 6, 6, 7, 8, 9,
	10, 11, 13, 14, 16, 18, 20, 22, 25, 28,
	32, 35, 40, 45, 50, 56, 63, 71, 79, 89,
	100, 112, 126, 141, 158, 178, 200, 224, 251, 282,
	316, 355, 398, 447, 501, 562, 631, 708, 794, 891,
	1000
};

static void *ndis_soft_state_p = NULL;
/* "Array of driver-private property names, terminated by a null pointer." */
static char *priv_props = (char*)NULL;
uint32_t ndis_debug; /* JPA was in kern_ndis.c */

int ndis_attach(dev_info_t *, ddi_attach_cmd_t);
static int ndis_detach(dev_info_t *, ddi_detach_cmd_t);
static int ndis_quiesce(dev_info_t *);

MSTYPE3(void ndis_txeof, ndis_miniport_block*, ndis_packet *, ndis_status);
MSTYPE3(void ndis_rxeof, ndis_miniport_block*, ndis_packet **, uint32_t);
MSTYPE4(void ndis_linksts, ndis_miniport_block*, ndis_status, void *, uint32_t);
MSTYPE1(void ndis_linksts_done, ndis_miniport_block*);

void iotest_ping(struct ndis_softc *sc);

static ndis_senddone_func ndis_txeof_wrap;
static ndis_rxeof_handler ndis_rxeof_wrap;
static ndis_status_handler ndis_linksts_wrap;
static ndis_statusdone_handler ndis_linksts_done_wrap;

static uint32_t ndis_intr(caddr_t arg);
static void ndis_tick		(void *);
static void ndis_ticktask	(void *);

static mblk_t *ndis_m_tx(void *arg, mblk_t *mp);
static int32_t ndis_m_start(void *);
static void ndis_m_stop  (void *);

static int ndis_tx_begin(struct ndis_softc *sc, mblk_t *mp);

static int  ndis_get_assoc      (struct ndis_softc *, ndis_wlan_bssid_ex **);
static void ndis_getstate_80211	(struct ndis_softc *);

uint32_t ieee80211_ieee2mhz(uint32_t chan, uint32_t  flags);
uint32_t ieee80211_mhz2ieee(uint32_t freq, uint32_t flags);
uint32_t ieee80211_chan2ieee(ieee80211com_t *isc, struct ieee80211_channel *ch);

#define NDIS_DRIVER_NAME 	"bcmndis"

DDI_DEFINE_STREAM_OPS(ndis_dev_ops, nulldev, nulldev, ndis_attach, ndis_detach,
    nodev, NULL, D_MP, NULL, ndis_quiesce);

static struct modldrv ndis_modldrv = {
	&mod_driverops,		/* Type of module.  This one is a driver */
	"Ndiswrapper 1.3.0",	/* short description */
	&ndis_dev_ops		/* driver specific ops */
};

static struct modlinkage modlinkage = {
	MODREV_1, (void *)&ndis_modldrv, NULL
};

uint8_t wifi_bcastaddr[]	= { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
static uint8_t wifi_ietfmagic[]	= { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };
static uint8_t wifi_ieeemagic[]	= { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8 };
extern int *ndis_i_pcnt;

#define	IEEE80211_IS_MULTICAST(addr)	\
	((((uint8_t *)addr)[0]) & 0x01)

/*
 *	mi_strlen and mi_mpprintf_putc borrowed from
 *	http://fxr.watson.org/fxr/source/common/inet/mi.c?v=OPENSOLARIS
 */

size_t
mi_strlen(const char *str)
{
	const char *cp = str;

	while (*cp != '\0')
		 cp++;
	return ((int)(cp - str));
}

int
mi_mpprintf_putc(char *cookie, int ch)
{
	MBLKP	mp = (MBLKP)cookie;

	while (mp->b_cont)
		mp = mp->b_cont;
	if (mp->b_wptr >= mp->b_datap->db_lim) {
		mp->b_cont = allocb(1024, BPRI_HI);
		mp = mp->b_cont;
		if (!mp)
			return (0);
	}
	*mp->b_wptr++ = (unsigned char)ch;
	return (1);
}

static boolean_t ndis_m_getcap(void *arg, mac_capab_t c, void *v)
{
	cmn_err(CE_WARN,"Unexpected ndis_m_getcap() arg 0x%lx\n",(long)arg);
	return (0);
}

static int ndis_m_open(void *arg)
{
	cmn_err(CE_WARN,"Unexpected ndis_m_open() arg 0x%lx\n",(long)arg);
	return (0);
}

static void ndis_m_close(void *arg)
{
	cmn_err(CE_WARN,"Unexpected ndis_m_close() arg 0x%lx\n",(long)arg);
}

static void ndis_m_propinf(void *arg, const char *s, mac_prop_id_t id,
			mac_prop_info_handle_t p)
{
	cmn_err(CE_WARN,"Unexpected ndis_m_propinf() arg 0x%lx\n",(long)arg);
}

static int
__mac_wifi_unicst_verify(const void *addr, void *pdata)
{
	/* If it's not a group address, then it's a valid unicast address. */
	return (IEEE80211_IS_MULTICAST(addr) ? EINVAL : 0);
}

/* ARGSUSED */
static int
__mac_wifi_multicst_verify(const void *addr, void *pdata)
{
	/* The address must be a group address. */
	if (!IEEE80211_IS_MULTICAST(addr))
		return (EINVAL);
	/* The address must not be the media broadcast address. */
	if (bcmp(addr, wifi_bcastaddr, sizeof (wifi_bcastaddr)) == 0)
		return (EINVAL);
	return (0);
}

static boolean_t
__mac_wifi_sap_verify(uint32_t sap, uint32_t *bind_sap, void *pdata)
{
	if (sap >= ETHERTYPE_802_MIN && sap <= ETHERTYPE_MAX) {
		if (bind_sap != NULL)
			*bind_sap = sap;
		return (B_TRUE);
	}

	if (sap <= ETHERMTU) {
		if (bind_sap != NULL)
			*bind_sap = DLS_SAP_LLC;
		return (B_TRUE);
	}
	return (B_FALSE);
}


static mblk_t *
mac_wifi_header_new(const void *saddr, const void *daddr, uint32_t sap,
    void *pdata, mblk_t *payload, size_t extra_len)
{
	struct ieee80211_frame	*wh;
	struct ieee80211_llc	*llc;
	mblk_t			*mp;
	wifi_data_t		*wdp = pdata;

	if (!__mac_wifi_sap_verify(sap, NULL, NULL))
		return (NULL);

	if ((mp = allocb(WIFI_HDRSIZE + extra_len, BPRI_HI)) == NULL)
		return (NULL);
	bzero(mp->b_rptr, WIFI_HDRSIZE + extra_len);

	/*
	 * Fill in the fixed parts of the ieee80211_frame.
	 */
	wh = (struct ieee80211_frame *)mp->b_rptr;
	mp->b_wptr += sizeof (struct ieee80211_frame);
	wh->i_fc[0] = IEEE80211_FC0_VERSION_0 | IEEE80211_FC0_TYPE_DATA;

	switch (wdp->wd_opmode) {
	case IEEE80211_M_STA:
		wh->i_fc[1] = IEEE80211_FC1_DIR_TODS;
		IEEE80211_ADDR_COPY(wh->i_addr1, wdp->wd_bssid);
		IEEE80211_ADDR_COPY(wh->i_addr2, saddr);
		IEEE80211_ADDR_COPY(wh->i_addr3, daddr);
		break;

	case IEEE80211_M_IBSS:
	case IEEE80211_M_AHDEMO:
		wh->i_fc[1] = IEEE80211_FC1_DIR_NODS;
		IEEE80211_ADDR_COPY(wh->i_addr1, daddr);
		IEEE80211_ADDR_COPY(wh->i_addr2, saddr);
		IEEE80211_ADDR_COPY(wh->i_addr3, wdp->wd_bssid);
		break;

	case IEEE80211_M_HOSTAP:
		wh->i_fc[1] = IEEE80211_FC1_DIR_FROMDS;
		IEEE80211_ADDR_COPY(wh->i_addr1, daddr);
		IEEE80211_ADDR_COPY(wh->i_addr2, wdp->wd_bssid);
		IEEE80211_ADDR_COPY(wh->i_addr3, saddr);
		break;
	}

	/*
	 * Fill in the fixed parts of the WEP-portion of the frame.
	 */
	if (wdp->wd_secalloc == WIFI_SEC_WEP) {
		wh->i_fc[1] |= IEEE80211_FC1_WEP;
		/*
		 * The actual contents of the WEP-portion of the packet
		 * are computed when the packet is sent -- for now, we
		 * just need to account for the size.
		 */
		mp->b_wptr += IEEE80211_WEP_IVLEN + IEEE80211_WEP_KIDLEN;
	}

	/*
	 * Fill in the fixed parts of the ieee80211_llc header.
	 */
	llc = (struct ieee80211_llc *)mp->b_wptr;
	mp->b_wptr += sizeof (struct ieee80211_llc);
	bcopy(wifi_ietfmagic, llc, sizeof (wifi_ietfmagic));
	llc->illc_ether_type = htons(sap);

	return (mp);
}


static mblk_t *
mac_wifi_header_fake(mblk_t *mp, void *pdata)
{
	struct ether_header	*ehp;
	mblk_t			*llmp;

	if (MBLKL(mp) < sizeof (struct ether_header))
		return (NULL);

	ehp = (struct ether_header *)mp->b_rptr;
	llmp = mac_wifi_header_new(&ehp->ether_shost, &ehp->ether_dhost,
	    ntohs(ehp->ether_type), pdata, NULL, 0);
	if (llmp == NULL)
		return (NULL);

	/*
	 * The plugin framework guarantees that we have the only reference
	 * to the mblk_t, so we can safely modify it.
	 */
	mp->b_rptr += sizeof (struct ether_header);
	llmp->b_cont = mp;
	return (llmp);
}

static int
mac_wifi_header_check(mblk_t *mp, void *pdata, mac_header_info_t *mhp)
{
	struct ieee80211_frame	*wh;
	struct ieee80211_llc	*llc;
	uchar_t			*llcp;

	if (MBLKL(mp) < sizeof (struct ieee80211_frame))
		return (EINVAL);

	wh = (struct ieee80211_frame *)mp->b_rptr;
	llcp = mp->b_rptr + sizeof (struct ieee80211_frame);

	/*
	 * When we receive frames from other hosts, the hardware will have
	 * already performed WEP decryption, and thus there will not be a WEP
	 * portion.  However, when we receive a loopback copy of our own
	 * packets, it will still have a WEP portion.  Skip past it to get to
	 * the LLC header.
	 */
	if (wh->i_fc[1] & IEEE80211_FC1_WEP)
		llcp += IEEE80211_WEP_IVLEN + IEEE80211_WEP_KIDLEN;

	if (mp->b_wptr - llcp < sizeof (struct ieee80211_llc))
		return (EINVAL);

	llc = (struct ieee80211_llc *)llcp;
	mhp->mhi_origsap = ntohs(llc->illc_ether_type);
	mhp->mhi_bindsap = mhp->mhi_origsap;
	mhp->mhi_pktsize = 0;
	mhp->mhi_hdrsize = llcp + sizeof (*llc) - mp->b_rptr;

	/*
	 * Verify the LLC header is one of the known formats.  As per MSFT's
	 * convention, if the header is using IEEE 802.1H encapsulation, then
	 * treat the LLC header as data.  As per DL_ETHER custom when treating
	 * the LLC header as data, set the mhi_bindsap to be DLS_SAP_LLC, and
	 * assume mhi_origsap contains the data length.
	 */
	if (bcmp(llc, wifi_ieeemagic, sizeof (wifi_ieeemagic)) == 0) {
		mhp->mhi_bindsap = DLS_SAP_LLC;
		mhp->mhi_hdrsize -= sizeof (*llc);
		mhp->mhi_pktsize = mhp->mhi_hdrsize + mhp->mhi_origsap;
	} else if (bcmp(llc, wifi_ietfmagic, sizeof (wifi_ietfmagic)) != 0) {
		return (EINVAL);
	}

	switch (wh->i_fc[1] & IEEE80211_FC1_DIR_MASK) {
	case IEEE80211_FC1_DIR_NODS:
		mhp->mhi_daddr = wh->i_addr1;
		mhp->mhi_saddr = wh->i_addr2;
		break;

	case IEEE80211_FC1_DIR_TODS:
		mhp->mhi_daddr = wh->i_addr3;
		mhp->mhi_saddr = wh->i_addr2;
		break;

	case IEEE80211_FC1_DIR_FROMDS:
		mhp->mhi_daddr = wh->i_addr1;
		mhp->mhi_saddr = wh->i_addr3;
		break;

	case IEEE80211_FC1_DIR_DSTODS:
		/* We don't support AP-to-AP mode yet */
		return (ENOTSUP);
	}

	if (__mac_wifi_unicst_verify(mhp->mhi_daddr, NULL) == 0)
		mhp->mhi_dsttype = MAC_ADDRTYPE_UNICAST;
	else if (__mac_wifi_multicst_verify(mhp->mhi_daddr, NULL) == 0)
		mhp->mhi_dsttype = MAC_ADDRTYPE_MULTICAST;
	else
		mhp->mhi_dsttype = MAC_ADDRTYPE_BROADCAST;

	return (0);
}

static mblk_t *
mac_wifi_header_backout(mblk_t *mp, void *pdata)
{
	mac_header_info_t	mhi;
	struct ether_header	eh;

	if (mac_wifi_header_check(mp, pdata, &mhi) != 0) {
		/*
		 * The plugin framework guarantees the header is properly
		 * formed, so this should never happen.
		 */
		return (NULL);
	}

	/*
	 * The plugin framework guarantees that we have the only reference to
	 * the mblk_t and the underlying dblk_t, so we can safely modify it.
	 */

	IEEE80211_ADDR_COPY(&eh.ether_dhost, mhi.mhi_daddr);
	IEEE80211_ADDR_COPY(&eh.ether_shost, mhi.mhi_saddr);
	eh.ether_type = htons(mhi.mhi_origsap);

	mp->b_rptr += mhi.mhi_hdrsize - sizeof (struct ether_header);
	bcopy(&eh, mp->b_rptr, sizeof (struct ether_header));
	return (mp);
}

kmutex_t	top_mutex;
#define TOP_LOCK()  mutex_enter(&top_mutex)
#define TOP_UNLOCK()  mutex_exit(&top_mutex)

/*
 *		Check the layout of Windows-defined structs.
 *
 *	This is useful to detect situations where the compiler does
 *	not use the same rules as Windows (mostly alignment issues)
 *
 *	As this only checks constants, a decent compiler produces no code
 */

#define ASSERT_EQ(expr, val)						\
	do {								\
		if ((expr)!=(val)) {					\
			cmn_err(CE_NOTE,				\
				"file %s: line %d: assertion failed:"	\
				" (%s == %s), expr is %ld\n",		\
				__FILE__,				\
				__LINE__,				\
				#expr, #val, (long)(expr) );		\
			return (expr);					\
		}							\
	} while (0);

int layout_check(void)
{
#ifdef __amd64__
	ASSERT_EQ(sizeof(struct cm_partial_resource_desc), 20);
	ASSERT_EQ(sizeof(struct cm_partial_resource_list), 28);
	ASSERT_EQ(sizeof(struct irp), 208);
	ASSERT_EQ(sizeof(struct io_stack_location), 72);

	ASSERT_EQ(offsetof(struct cm_partial_resource_desc,
				ut.cprd_generic.cprd_len), 12);
	ASSERT_EQ(offsetof(struct cm_partial_resource_desc,
				ut.cprd_port.cprd_len), 12);
	ASSERT_EQ(offsetof(struct cm_partial_resource_desc,
				ut.cprd_mem.cprd_len), 12);

	ASSERT_EQ(offsetof(struct irp, irp_currentstackloc), 67);
	ASSERT_EQ(offsetof(struct irp, irp_tail.irp_overlay.irp_csl), 184);

	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_read.isl_len), 8);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_read.isl_byteoff), 24);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_write.isl_len), 8);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_write.isl_byteoff), 24);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_ioctl.isl_obuflen), 8);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_ioctl.isl_ibuflen), 16);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_parameters.isl_ioctl.isl_iocode), 24);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_devobj), 40);
	ASSERT_EQ(offsetof(struct io_stack_location,
				isl_completionfunc), 56);
#endif
	return 0;
}

int _init(void)
{
	int e;
	extern mac_callbacks_t ndis_m_callbacks;

	cmn_err(CE_CONT,"bcmndis %s at 0x%lx ndis_m_callbacks at 0x%lx\n",
			__DATE__, (long)drv_data, (long)&ndis_m_callbacks);
	if (layout_check()) {
		cmn_err(CE_NOTE,"%s is badly compiled and fails layout checks,"
			" cannot run\n", NDIS_DRIVER_NAME);
		return (ENOEXEC);
	}

	if (e = ddi_soft_state_init(&ndis_soft_state_p, 
	    sizeof (struct ndis_softc), 1))
		return (e);
	
	mac_init_ops(&ndis_dev_ops, NDIS_DRIVER_NAME);
	if (e = mod_install(&modlinkage)){
		ddi_soft_state_fini(&ndis_soft_state_p);
		mac_fini_ops(&ndis_dev_ops);
	}
	mutex_init(&top_mutex, NULL, MUTEX_DRIVER, NULL);

	return (e);
}

int _fini(void)
{
	int e;
	if (e = mod_remove(&modlinkage))
		return (e);
	mac_fini_ops(&ndis_dev_ops);
	ddi_soft_state_fini(&ndis_soft_state_p);
	mutex_destroy(&top_mutex);
	return (e);
}

int _info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

static int32_t
ndis_cfg_essid(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	ndis_80211_ssid ssid;
	wldp_t *infp;
	char *value;
	uint16_t i;
	wldp_t *outfp;
	char *buf;
	int rval;
	int len;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,"ieee80211: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		if (mi_strlen((const char *)isc->ic_des_essid) == 0) {
			outfp->wldp_length = WIFI_BUF_OFFSET +
			    offsetof(wl_essid_t, wl_essid_essid) +
			    mi_strlen((const char *)
			    isc->ic_bss->in_essid);
			((wl_essid_t *)(outfp->wldp_buf))->wl_essid_length =
			    mi_strlen((const char *)
			    isc->ic_bss->in_essid);
			bcopy(isc->ic_bss->in_essid,
			    buf + WIFI_BUF_OFFSET +
			    offsetof(wl_essid_t, wl_essid_essid),
			    mi_strlen((const char *)
			    isc->ic_bss->in_essid));
		} else {
			outfp->wldp_length = WIFI_BUF_OFFSET +
			    offsetof(wl_essid_t, wl_essid_essid) +
			    mi_strlen((const char *)isc->ic_des_essid);
			((wl_essid_t *)(outfp->wldp_buf))->wl_essid_length =
			    mi_strlen((const char *)isc->ic_des_essid);
			bcopy(isc->ic_des_essid,
			    buf + WIFI_BUF_OFFSET +
			    offsetof(wl_essid_t, wl_essid_essid),
			    mi_strlen((const char *)
			    isc->ic_des_essid));
		}
		outfp->wldp_result = WL_SUCCESS;
		
	} else if (cmd == WLAN_SET_PARAM) {
		value = ((wl_essid_t *)(infp->wldp_buf))->wl_essid_essid;
		bzero(isc->ic_des_essid, IEEE80211_NWID_LEN);
		if (mi_strlen(value) == 0) {
			bzero(isc->ic_des_essid, IEEE80211_NWID_LEN);
			isc->ic_des_esslen = 0;
		} else {
			len = mi_strlen((const char *)value);
			if (len > IEEE80211_NWID_LEN)
				len = IEEE80211_NWID_LEN;
			bcopy(value, isc->ic_des_essid, len);
			isc->ic_des_esslen = len;
			len = sizeof(ndis_80211_ssid);
			bzero((char *)&ssid, len);
			ssid.ns_ssidlen = isc->ic_des_esslen;
			if (ssid.ns_ssidlen == 0)
				ssid.ns_ssidlen = 1;
			else
				bcopy(isc->ic_des_essid, ssid.ns_ssid, ssid.ns_ssidlen);
			NDISDBG((CE_CONT,"%s: bssid=%s\n",__func__, isc->ic_des_essid));

			outfp->wldp_length = WIFI_BUF_OFFSET;

			rval = ndis_set_info(sc, OID_802_11_SSID, (void *)&ssid, &len);
			if (rval){
				NDISDBG((CE_CONT,"set ssid failed: %d\n", rval));
				outfp->wldp_result = WL_HW_ERROR;
				goto done;
			}
			outfp->wldp_result = WL_SUCCESS;
		}		
	}

done:
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
}

static int
ndis_set_bssid(struct ndis_softc *sc, const void *wldp_buf)
{
	int len, rval;
	char *bssid = (char *)wldp_buf;
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	
	cmn_err(CE_WARN, "wl_set_bssid: "
	    "set bssid=%s\n",
	    ieee80211_macaddr_sprintf(wldp_buf));
	len = 6;
	rval = ndis_set_info(sc, OID_802_11_BSSID, (void *)bssid, &len);
	bcopy(wldp_buf, isc->ic_des_bssid, sizeof (wl_bssid_t));
	isc->ic_flags |= IEEE80211_F_DESBSSID;

	return (ENETRESET);
}

static void
ndis_get_bssid(struct ndis_softc *sc, void *wldp_buf)
{
	uint8_t bssid[6];
	int len;
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;

	if (isc->ic_flags & IEEE80211_F_DESBSSID)
		bcopy(isc->ic_des_bssid, bssid, 6);
	else{
		len = 6;
		(void) ndis_get_info(sc, OID_802_11_BSSID, (void *)bssid, &len);
	}
	bcopy(bssid, wldp_buf, sizeof (wl_bssid_t));
}

static int32_t
ndis_cfg_bssid(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	wldp_t *outfp;
	char *buf;
	
	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT, "ndis: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN));
		return (ENOMEM);
	}
	
	outfp = (wldp_t *)buf;
	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		ndis_get_bssid(sc, buf + WIFI_BUF_OFFSET);
		outfp->wldp_length =  WIFI_BUF_OFFSET + sizeof(wl_bssid_t);
		outfp->wldp_result = WL_SUCCESS;
	} else if (cmd == WLAN_SET_PARAM) {
		ndis_set_bssid(sc, infp->wldp_buf);
		outfp->wldp_result = WL_SUCCESS;
		outfp->wldp_length =  WIFI_BUF_OFFSET;
	} else {
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}


static int32_t
ndis_cfg_encryption(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	wldp_t *outfp;
	char *buf;
	uint32_t arg;
	int /* was uint32_t */ len=0;
	int rval=0;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
#ifdef _KERNEL
		cmn_err(CE_CONT,"ndis: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN);
#else
		printf("ndis: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN);
#endif
		return (ENOMEM);
	}
	
	outfp = (wldp_t *)buf;
	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));


	outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_encryption_t);
	    
	if (cmd == WLAN_GET_PARAM) {
		*(wl_encryption_t *)(outfp->wldp_buf) =
		    (isc->ic_flags & IEEE80211_F_PRIVACY) ? 1 : 0;
		outfp->wldp_length =  WIFI_BUF_OFFSET + sizeof(wl_encryption_t);
		outfp->wldp_result = WL_SUCCESS;
	} else if (cmd == WLAN_SET_PARAM) {

		if (*(wl_encryption_t *)(infp->wldp_buf) ==
		    WL_NOENCRYPTION) {
			isc->ic_flags &= ~IEEE80211_F_PRIVACY;
		} else if (*(wl_encryption_t *)(infp->wldp_buf) ==
		    WL_ENC_WEP) {
			isc->ic_flags |= IEEE80211_F_PRIVACY;
		}
		
		if (isc->ic_flags & IEEE80211_F_PRIVACY) {
			arg = NDIS_80211_WEPSTAT_ENABLED;
			len = sizeof(arg);
			rval = ndis_set_info(sc, OID_802_11_WEP_STATUS, (void *)&arg, &len);
			if (rval){
				NDISDBG((CE_CONT, "enable WEP failed: %d\n", rval));
				outfp->wldp_result = WL_HW_ERROR;
				goto done;
			}

		} else {
			arg = NDIS_80211_WEPSTAT_DISABLED;
			len = sizeof(arg);
			rval = ndis_set_info(sc, OID_802_11_WEP_STATUS, (void *)&arg, &len);
			if (rval){
				NDISDBG((CE_CONT, "disable WEP failed: %d\n", rval));
				outfp->wldp_result = WL_HW_ERROR;
				goto done;
			}			
		}
		
		NDISDBG((CE_CONT,"ndis_getset: set encryption=%d\n",
		    (isc->ic_flags & IEEE80211_F_PRIVACY) ? 1 : 0));

		outfp->wldp_result = WL_SUCCESS;
	}

done:
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
}


static int32_t
ndis_cfg_wepkeyindex(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	wldp_t *outfp;
	char *buf;
	int  len, rval = 0;
	ndis_80211_wep	wep;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,"ndis: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_wep_key_id_t);

	if (cmd == WLAN_GET_PARAM) {
		*(wl_wep_key_id_t *)(outfp->wldp_buf) =
		    isc->ic_crypto.cs_def_txkey;
		outfp->wldp_result = WL_SUCCESS;
	} else if (cmd == WLAN_SET_PARAM) {
		isc->ic_crypto.cs_def_txkey = *(wl_wep_key_id_t *)(infp->wldp_buf);
		NDISDBG((CE_CONT,"set wepkeyid=%d\n",
		    isc->ic_crypto.cs_def_txkey));
		for (i = 0; i < IEEE80211_WEP_NKID; i++) {
			if (isc->ic_nw_keys[i].wk_keylen) {
				bzero((char *)&wep, sizeof(wep));
				wep.nw_keylen = isc->ic_nw_keys[i].wk_keylen;
				
				wep.nw_keyidx = i;
				wep.nw_length = (sizeof(uint32_t) * 3)
				    + wep.nw_keylen;
				if (i == (isc->ic_crypto.cs_def_txkey))
					wep.nw_keyidx |= NDIS_80211_WEPKEY_TX;
				else
					wep.nw_keyidx &= ~NDIS_80211_WEPKEY_TX;

				bcopy(isc->ic_nw_keys[i].wk_key,
				    wep.nw_keydata, wep.nw_length);
				len = sizeof(wep);
				rval = ndis_set_info(sc,
				    OID_802_11_ADD_WEP, &wep, &len);	
				if (rval){
					NDISDBG((CE_CONT,
					    "set wepkey failed: %d\n", rval));
					outfp->wldp_result = WL_HW_ERROR;
					goto done;
				}
			}
		}
		outfp->wldp_result = WL_SUCCESS;
	}

done:
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
}


static int32_t
ndis_cfg_setwepkey(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	wldp_t *outfp;
	char *buf;
	int  len, rval = 0;
	ndis_80211_wep	wep;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,"ndis: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));
	
	outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_wep_key_tab_t);

	if (cmd == WLAN_GET_PARAM) {
		outfp->wldp_result = WL_WRITEONLY;
	} else if (cmd == WLAN_SET_PARAM) {
		wl_wep_key_t *p_wepkey_tab;
		p_wepkey_tab = (wl_wep_key_t *)(infp->wldp_buf);

		for (i = 0; i < 4; i++) {
			if (p_wepkey_tab[i].wl_wep_operation == WL_ADD) {
				isc->ic_nw_keys[i].wk_keylen =
				    p_wepkey_tab[i].wl_wep_length;
				NDISDBG((CE_CONT,"isc->ic_nw_keys[%d].wk_keylen=%d"
					,i,p_wepkey_tab[i].wl_wep_length));
				bcopy(p_wepkey_tab[i].wl_wep_key,
				    isc->ic_nw_keys[i].wk_key,
				    p_wepkey_tab[i].wl_wep_length);
			}
		}

		isc->ic_flags |=IEEE80211_F_PRIVACY; 

		for (i = 0; i < IEEE80211_WEP_NKID; i++) {
			if (isc->ic_nw_keys[i].wk_keylen) {
				bzero((char *)&wep, sizeof(wep));
				wep.nw_keylen = isc->ic_nw_keys[i].wk_keylen;
				
				wep.nw_keyidx = i;
				wep.nw_length = (sizeof(uint32_t) * 3)
				    + wep.nw_keylen;
				if (i == (isc->ic_crypto.cs_def_txkey))
					wep.nw_keyidx |= NDIS_80211_WEPKEY_TX;
				else
					wep.nw_keyidx &= ~NDIS_80211_WEPKEY_TX;

				bcopy(isc->ic_nw_keys[i].wk_key,
				    wep.nw_keydata, wep.nw_length);
				len = sizeof(wep);
				rval = ndis_set_info(sc,
				    OID_802_11_ADD_WEP, &wep, &len);	
				if (rval){
					NDISDBG((CE_CONT,
					    "set wepkey failed: %d\n", rval));
					outfp->wldp_result = WL_HW_ERROR;
					goto done;
				}
			}
		}
		outfp->wldp_result = WL_SUCCESS;
	}
done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
}


static int32_t
ndis_cfg_scan(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	wldp_t *outfp;
	char *buf;
	int len;
	uint32_t i;
	
	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			  "ndis: can not alloc so much memory!(%d)\n",
			  MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;
	bcopy(mp->b_rptr, buf,  sizeof(wldp_t));
	if(sc->ndis_link)
		goto outing;
	len = 0;
	(void) ndis_set_info(sc, OID_802_11_BSSID_LIST_SCAN,
	    			NULL, &len);
	NDIS_UNLOCK(sc);
	delay(300);
	NDIS_LOCK(sc);
outing:
	outfp->wldp_result = WL_SUCCESS;

	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;
}

void ndis_wpa_ie_parser(wl_ess_conf_t *ap, uint8_t *ie,
 				    size_t ie_len)
{
	uint8_t *pos = ie;
	uint8_t *end = ie + ie_len;
 
	if (ie_len < sizeof(struct ndis_80211_fixed_ies))
		return;

		pos += sizeof(struct ndis_80211_fixed_ies);
	while (pos + 1 < end && pos + 2 + pos[1] <= end) {
	 		uint8_t ielen = 2 + pos[1];
	 		if (ielen > IEEE80211_MAX_WPA_IE) {
				pos += ielen;
				continue;
			}
			if (pos[0] == GENERIC_INFO_ELEM && pos[1] >= 4 &&
			    memcmp(pos + 2, "\x00\x50\xf2\x01", 4) == 0) {
				ap->wl_ess_conf_reserved[0] = 1;
			} else if (pos[0] == RSN_INFO_ELEM) {
				ap->wl_ess_conf_reserved[0] = 1;
 			}
			pos += ielen;
	}
}

int
ndis_nettype_parser(wl_ess_conf_t *ap, ndis_wlan_bssid_ex *wb)
{
	if(ap == NULL || wb ==NULL)
		return -1;
	
	switch(wb->nwbx_nettype){
	case NDIS_80211_NETTYPE_11FH:
		NDISDBG((CE_CONT,"NDIS_80211_NETTYPE_11FH\n"));
		wl_fhss_t *fhss = (wl_fhss_t *)&((ap->wl_phy_conf).wl_phy_fhss_conf);
		fhss->wl_fhss_subtype = WL_FHSS;
		fhss->wl_fhss_dwelltime = wb->nwbx_config.nc_fhconfig.ncf_dwelltime;
		fhss->wl_fhss_channel = ieee80211_mhz2ieee(wb->nwbx_config.nc_dsconfig / 1000, 0);
		break;
	case NDIS_80211_NETTYPE_11DS:
		NDISDBG((CE_CONT,"NDIS_80211_NETTYPE_11DS\n"));
		wl_dsss_t *dsss = (wl_dsss_t *)&((ap->wl_phy_conf).wl_phy_dsss_conf);
		dsss->wl_dsss_subtype = WL_DSSS;
		dsss->wl_dsss_channel = ieee80211_mhz2ieee(wb->nwbx_config.nc_dsconfig / 1000, 0);
		break;
	case NDIS_80211_NETTYPE_11OFDM5:
		NDISDBG((CE_CONT,"NDIS_80211_NETTYPE_11OFDM5\n"));
		wl_ofdm_t *ofdm = (wl_ofdm_t *)&((ap->wl_phy_conf).wl_phy_ofdm_conf);
		ofdm->wl_ofdm_subtype = WL_OFDM;		
		ofdm->wl_ofdm_frequency = wb->nwbx_config.nc_dsconfig / 1000;
		break;				
	case NDIS_80211_NETTYPE_11OFDM24:
		NDISDBG((CE_CONT,"NDIS_80211_NETTYPE_11OFDM24\n"));
		wl_erp_t *erp = (wl_erp_t *)&((ap->wl_phy_conf).wl_phy_erp_conf);
		erp->wl_erp_subtype = WL_ERP;		
		erp->wl_erp_channel = ieee80211_mhz2ieee(wb->nwbx_config.nc_dsconfig / 1000, 0);
		break;
	default:
		cmn_err(CE_CONT,"unknown nettype 0x%08x\n", wb->nwbx_nettype);
		return (-1);
	}
	return (0);
}

static int32_t
ndis_cfg_esslist(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i,j;
	wldp_t *outfp;
	char *buf;
	uint8_t *db;
	uint32_t rval;
	int /* was uint32_t */ len;
	ndis_80211_bssid_list_ex *ess_list;
	ndis_wlan_bssid_ex *wb;
	int32_t /* was uint32_t */ rssi;
	
	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			  "ieee80211: can not alloc so much memory!(%d)\n",
			  MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof(wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		len = 8192;
		ess_list = (ndis_80211_bssid_list_ex *)
				kmem_zalloc(len, KM_NOSLEEP);
		if (ess_list == NULL) {
			NDISDBG((CE_CONT,
			     "ndis: can not alloc so much memory!(%d)\n",
			     len));
			return (ENOMEM);
		}
		rval = ndis_get_info(sc, OID_802_11_BSSID_LIST, ess_list, &len);
		NDISDBG((CE_CONT,"list len=%d, rval=%x\n",len,rval));
		if (rval) {
			NDISDBG((CE_CONT,
			    "get bss list failed: %d\n", rval));	
			kmem_free(ess_list, 8192);
			goto done;
		}

		wb = (ndis_wlan_bssid_ex *)&ess_list->nblx_bssid[0];

		for(i=0; i< ess_list->nblx_items; i++){
			wl_ess_conf_t *p_ess_conf;

			p_ess_conf = (wl_ess_conf_t *)
			    (buf + WIFI_BUF_OFFSET +
			    offsetof(wl_ess_list_t, wl_ess_list_ess) +
			     i * sizeof (wl_ess_conf_t));
			db = wb->nwbx_ssid.ns_ssid;
			NDIS11NDBG((CE_CONT,"SSID name %s\n",wb->nwbx_ssid.ns_ssid));
			bcopy(&wb->nwbx_ssid, &p_ess_conf->wl_ess_conf_essid,
			    sizeof(ndis_80211_ssid));
			bcopy(wb->nwbx_macaddr, p_ess_conf->wl_ess_conf_bssid, 6);
//			(p_ess_conf->wl_phy_conf).wl_phy_dsss_conf.wl_dsss_subtype = WL_DSSS;
			if(wb->nwbx_netinfra == NDIS_80211_NET_INFRA_IBSS)
				p_ess_conf->wl_ess_conf_bsstype = WL_BSS_IBSS;
			else if(wb->nwbx_netinfra == NDIS_80211_NET_INFRA_BSS)
				p_ess_conf->wl_ess_conf_bsstype = WL_BSS_BSS;
			//p_ess_conf->wl_ess_conf_authmode??
			p_ess_conf->wl_ess_conf_wepenabled = 
				(wb->nwbx_privacy)?WL_ENC_WEP : WL_NOENCRYPTION;
			NDIS11NDBG((CE_CONT, "wb->nwbx_privacy(none/wep) = %d\n", wb->nwbx_privacy));
			//wpa detection
			ndis_wpa_ie_parser(p_ess_conf, (uint8_t *)&wb->nwbx_ies[0], wb->nwbx_ielen);
			/*signal strength*/
			rssi = (wb->nwbx_rssi + 200)/12;
			if(rssi < 0)
				rssi = 0;
			else if(rssi > 15)
				rssi = 15;			
			p_ess_conf->wl_ess_conf_sl = rssi;
			//AP maxrate detect
			for (j = 0; j < sizeof(wb->nwbx_supportedrates); j++) {
		              if ((wb->nwbx_supportedrates[j] & 0x7f) >
		                  p_ess_conf->wl_supported_rates[0]) {
	                            p_ess_conf->wl_supported_rates[0] =
	                                         wb->nwbx_supportedrates[j] & 0x7f;
	                         }
	             	}
			NDIS11NDBG((CE_CONT, "maxrates = %d\n", p_ess_conf->wl_supported_rates[0]));
			//nettype
			NDIS11NDBG((CE_CONT,"wb->nwbx_nettype = 0x%x\n",wb->nwbx_nettype));
			ndis_nettype_parser(p_ess_conf, wb);
			wb = (ndis_wlan_bssid_ex *)((char *)wb + wb->nwbx_len);
		}
	
		((wl_ess_list_t *)(outfp->wldp_buf))->wl_ess_list_num = ess_list->nblx_items;
		outfp->wldp_length = WIFI_BUF_OFFSET +
			    offsetof(wl_ess_list_t, wl_ess_list_ess) +
	   		    (ess_list->nblx_items) * sizeof (wl_ess_conf_t);
		outfp->wldp_result = WL_SUCCESS;		
		
		kmem_free(ess_list, 8192);
	
	} else if (cmd == WLAN_SET_PARAM) {
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_READONLY;
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;
}


static int32_t
ndis_cfg_bss_type(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	uint32_t arg;
	int len;
	int rval;
	wldp_t *outfp;
	char *buf;
	uint16_t type=0;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			  "ieee80211: can not alloc so much memory!(%d)\n",
			  MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));
	outfp->wldp_length = WIFI_BUF_OFFSET + sizeof (wl_bss_type_t);
	
	if (cmd == WLAN_GET_PARAM) {
		len = sizeof(arg);
		rval = ndis_get_info(sc, OID_802_11_INFRASTRUCTURE_MODE, 
				&arg, &len);
		if(rval) {
			NDISDBG((CE_CONT,"get bss type failed: %d\n", rval));
		}
		
		if(arg == NDIS_80211_NET_INFRA_IBSS){
			type = WL_BSS_IBSS;
			isc->ic_opmode = IEEE80211_M_IBSS;
		}else if(arg == NDIS_80211_NET_INFRA_BSS){
			type = WL_BSS_BSS;
			isc->ic_opmode = IEEE80211_M_STA;
		}else if(arg == NDIS_80211_NET_INFRA_AUTO){
			type = WL_BSS_ANY;
			isc->ic_opmode = IEEE80211_M_STA;
		}else{
			NDISDBG((CE_CONT,"%s: Wrong infra mode 0x%x\n",
						__func__,arg));
		}

		(*(wl_bss_type_t *)(outfp->wldp_buf)) = type;
		outfp->wldp_result = WL_SUCCESS;
	} else if (cmd == WLAN_SET_PARAM) {
		type = (uint16_t)(*(wl_bss_type_t *)(infp->wldp_buf));
		if ((type != WL_BSS_BSS) &&
			(type != WL_BSS_IBSS) &&
			(type != WL_BSS_ANY)) {
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
			goto done;
		}

		if(type == WL_BSS_IBSS) {
			isc->ic_opmode = IEEE80211_M_IBSS;
			arg = NDIS_80211_NET_INFRA_IBSS;
		}else if( type == WL_BSS_BSS) { 
			isc->ic_opmode = IEEE80211_M_STA;
			arg = NDIS_80211_NET_INFRA_BSS;
		}else if( type == WL_BSS_ANY){
			isc->ic_opmode = IEEE80211_M_STA;
			arg = NDIS_80211_NET_INFRA_BSS;
		}
		NDISDBG((CE_CONT,"%s: set bss type to 0x%x\n",
						__func__,type));		
		rval = ndis_set_info(sc, OID_802_11_INFRASTRUCTURE_MODE, 
				(void *)&arg, &len);
		if (rval)
			NDISDBG((CE_CONT, 
					"%s: set infra failed: %d\n", __func__, rval));
		outfp->wldp_result = WL_SUCCESS;	
	} else {
		cmn_err(CE_CONT,"%s:Missed set parm\n",__func__);
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}

static int32_t
ndis_cfg_phy(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp = (wldp_t *)(mp->b_rptr);
	uint16_t i;
	int chanflag, rval=0;
	wldp_t *outfp;
	char *buf;
	uint32_t chan;
	int len;
	ndis_80211_config config;	
	ndis_wlan_bssid_ex	*bs=0;
	wl_dsss_t *dsss=0;
	wl_fhss_t *fhss=0;
	wl_ofdm_t *ofdm=0;
	wl_erp_t *erp=0;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			"ndis: can not alloc so much memory!(%d)\n",
			MAX_BUF_LEN));
		return (ENOMEM);
	}
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	outfp = (wldp_t *)buf;
	dsss = (wl_dsss_t *)(outfp->wldp_buf);
	fhss = (wl_fhss_t *)(outfp->wldp_buf);
	ofdm = (wl_ofdm_t *)(outfp->wldp_buf);
	erp = (wl_erp_t *)(outfp->wldp_buf);
	if (cmd == WLAN_GET_PARAM) {
		if (sc->ndis_link) {
			switch( sc->nettype ) {
			case NDIS_80211_NETTYPE_11FH:
				outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_fhss_t);
				fhss->wl_fhss_subtype = WL_FHSS;
				fhss->wl_fhss_dwelltime = 200; //just fake it;
				fhss->wl_fhss_channel = sc->phychan;
				break;
			case NDIS_80211_NETTYPE_11DS:
				outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_dsss_t);
				dsss->wl_dsss_subtype = WL_DSSS;
				dsss->wl_dsss_channel = sc->phychan;
				break;
			case NDIS_80211_NETTYPE_11OFDM5:
				outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_ofdm_t);
				ofdm->wl_ofdm_subtype = WL_OFDM;		
				chanflag = IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_DYN|IEEE80211_CHAN_CCK;
				ofdm->wl_ofdm_frequency = ieee80211_ieee2mhz(sc->phychan, chanflag);
				break;
			case NDIS_80211_NETTYPE_11OFDM24:
			default:
				outfp->wldp_length = WIFI_BUF_OFFSET + sizeof(wl_erp_t);
				erp->wl_erp_subtype = WL_ERP;		
				erp->wl_erp_channel = sc->phychan;
				break;
			}
			outfp->wldp_result = WL_SUCCESS;
		}else{
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
			goto done;
		}
	} else if (cmd == WLAN_SET_PARAM) {
		if(sc->ndis_link){
			NDISDBG((CE_CONT,"%s: don't set phy when link up\n",__func__));
			kmem_free(buf, MAX_BUF_LEN);
			return (EINVAL);
		}
			
		chan = (uint32_t)
			(((wl_phy_conf_t *)(infp->wldp_buf))->wl_phy_dsss_conf.wl_dsss_channel);		
		NDISDBG((CE_CONT,"%s: chan set to [%d]\n",__func__, chan)); 
		if (chan < 1 || chan > IEEE80211_CHAN_MAX) {
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
			goto done;
		}

		len = sizeof(config);
		bzero((char *)&config, len);
		config.nc_length = len;
		config.nc_fhconfig.ncf_length = sizeof(ndis_80211_config_fh);
		rval = ndis_get_info(sc, OID_802_11_CONFIGURATION, &config, &len); 

		/*
		 * Some drivers expect us to initialize these values, so
		 * provide some defaults.
		 */
		if (config.nc_beaconperiod == 0)
			config.nc_beaconperiod = 100;
		if (config.nc_atimwin == 0)
			config.nc_atimwin = 100;
		if (config.nc_fhconfig.ncf_dwelltime == 0)
			config.nc_fhconfig.ncf_dwelltime = 200;

		if (rval == 0 ) { //workaround
			if (isc->ic_sup_rates[IEEE80211_MODE_11G].ir_nrates)
					/* IEEE80211_CHAN_G */
				chanflag |= IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_DYN;
			if (chan <= 14)	/* IEEE80211_CHAN_B */
				chanflag |= IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_CCK;
			if (isc->ic_sup_rates[IEEE80211_MODE_11A].ir_nrates &&
			    chan > 14)	/* IEEE80211_CHAN_A */
				chanflag |= IEEE80211_CHAN_5GHZ|IEEE80211_CHAN_OFDM;
		
			NDISDBG((CE_CONT,"%s: try channel fixed to [%d]\n",__func__,chan));	
			isc->ic_ibss_chan = &isc->ic_sup_channels[chan]; //fix channel pointer
			NDISDBG((CE_CONT,"%s: channel fixed to [%d]\n",__func__,chan));	
			isc->ic_ibss_chan->ich_freq =
			    ieee80211_ieee2mhz(chan, chanflag);
			config.nc_dsconfig =
			    isc->ic_ibss_chan->ich_freq * 1000;
			isc->ic_bss->in_chan = isc->ic_ibss_chan;
			len = sizeof(config);
			config.nc_length = len;
			config.nc_fhconfig.ncf_length =
				    sizeof(ndis_80211_config_fh);
			rval = ndis_set_info(sc, OID_802_11_CONFIGURATION,
				    (void *)&config, &len);
			if (rval){
				NWAMDBG((CE_CONT,"%s: failed\n",__func__));
				NDISDBG((CE_CONT, "couldn't change "
					    "DS config to %ukHz: %d\n",
					    config.nc_dsconfig, rval));
				outfp->wldp_result = WL_NOTSUPPORTED;
			}else{
				NDISDBG((CE_CONT, "changed "
					    "DS config to %ukHz: %d\n",
					    config.nc_dsconfig, rval));
			}
		} else 
			NDISDBG((CE_CONT, "couldn't retrieve "
			    "channel info: %d\n", rval));
		
		outfp->wldp_result = WL_SUCCESS;	
	} else {
		NWAMDBG((CE_CONT,"%s: should be command\n",__func__));
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}
done:
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}

static int32_t
ndis_cfg_rates(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	unsigned int i;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;
	outfp->wldp_length = WIFI_BUF_OFFSET;
	outfp->wldp_result = WL_NOTSUPPORTED;
done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}

static int32_t
ndis_cfg_radio(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	uint16_t i;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_NOTE, "can not alloc so much memory!(%d)\n",
		    MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		*(wl_radio_t *)(outfp->wldp_buf) = B_TRUE;
		outfp->wldp_length = WIFI_BUF_OFFSET + sizeof (wl_radio_t);
		outfp->wldp_result = WL_SUCCESS;
	} else if (cmd == WLAN_SET_PARAM) {
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_NOTSUPPORTED;
	} else {
		NWAMDBG((CE_CONT,"%s: command error\n",__func__));
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}
	for (i = 0; i < (outfp->wldp_length); i++)
		(void) mi_mpprintf_putc((char *)mp, buf[i]);
	kmem_free(buf, MAX_BUF_LEN);
	return (0);
}

static int32_t
ndis_cfg_supported_rates(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	uint16_t i;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ieee80211: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	outfp->wldp_length = WIFI_BUF_OFFSET;
	outfp->wldp_result = WL_NOTSUPPORTED;	

	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;						    
}

static int32_t
ndis_cfg_power_mode(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i,powermode=0;
	uint32_t arg;
	int rval;
	int len;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		len = sizeof(arg);
		rval = ndis_get_info(sc, OID_802_11_POWER_MODE, &arg, &len);

		if (rval)
			NDISDBG((CE_CONT,
			    "get power mode failed: %d\n", rval));

		switch(arg){
		case NDIS_80211_POWERMODE_CAM:
			NDISDBG((CE_CONT,"%s:NDIS_80211_POWERMODE_CAM\n ",__func__));
			powermode = WL_PM_AM;
			break;
		case NDIS_80211_POWERMODE_MAX_PSP:
			NDISDBG((CE_CONT,"%s:NDIS_80211_POWERMODE_MAX_PSP\n ",__func__));
			powermode = WL_PM_MPS;
			break;
		case NDIS_80211_POWERMODE_FAST_PSP:
			NDISDBG((CE_CONT,"%s:NDIS_80211_POWERMODE_FAST_PSP\n ",__func__));
			powermode = WL_PM_FAST;
			break;
		}
					
		if (powermode == WL_PM_AM)
			isc->ic_flags &= ~IEEE80211_F_PMGTON;
		else
			isc->ic_flags |= IEEE80211_F_PMGTON;
		
		((wl_ps_mode_t *)(outfp->wldp_buf))->wl_ps_mode = powermode;

		outfp->wldp_length = WIFI_BUF_OFFSET + 
			              sizeof (wl_ps_mode_t);
		outfp->wldp_result = WL_SUCCESS;

	} else if (cmd == WLAN_SET_PARAM) {
/*		if(sc->ndis_link){	
			cmn_err(CE_WARN,"%s: don't set powermode when link up\n",__func__);
			kmem_free(buf, MAX_BUF_LEN);
			return (EINVAL);
		}
*/			
		powermode = (uint16_t)(((wl_ps_mode_t *)(infp->wldp_buf))->wl_ps_mode);
		if ((powermode != WL_PM_AM) &&
			(powermode != WL_PM_MPS) &&
			(powermode != WL_PM_FAST)) {
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
			goto done;
		}
		switch(powermode){
		case WL_PM_AM:
			arg = NDIS_80211_POWERMODE_CAM;
			break;
		case WL_PM_MPS:
			arg = NDIS_80211_POWERMODE_MAX_PSP;
			break;
		case WL_PM_FAST:
			arg = NDIS_80211_POWERMODE_FAST_PSP;
			break;
		}
		
		len = sizeof(arg);
		rval = ndis_set_info(sc, OID_802_11_POWER_MODE, 
				(void *)&arg, &len);
		if (rval){
			NDISDBG((CE_CONT,"set power failed: %d\n",rval)); 
			kmem_free(buf, MAX_BUF_LEN);
			return (EINVAL);
		}else{
			if (arg == NDIS_80211_POWERMODE_CAM)
				isc->ic_flags &= ~IEEE80211_F_PMGTON;
			else
				isc->ic_flags |= IEEE80211_F_PMGTON;
		}
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_SUCCESS;	
	} else {
		NDISDBG((CE_CONT,"%s: powermode failed\n",__func__));			
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}


static int32_t
ndis_cfg_auth_mode(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	uint32_t	arg;
	int rval;
	int len;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		len = sizeof(arg);
		rval = ndis_get_info(sc, OID_802_11_AUTHENTICATION_MODE,
				&arg, &len);
		if (rval)
			NDISINF((CE_CONT,"get authmode failed: %d\n", rval));

		isc->ic_bss->in_authmode = arg;

		if( arg == NDIS_80211_AUTHMODE_OPEN) {
			*(wl_authmode_t *)(outfp->wldp_buf) = WL_OPENSYSTEM;
			NDISDBG((CE_CONT,"WL_OEPNSYS mode\n"));
		}else if( arg == NDIS_80211_AUTHMODE_SHARED) {
			*(wl_authmode_t *)(outfp->wldp_buf) = WL_SHAREDKEY;
			NDISDBG((CE_CONT,"WL_SHAREDKEY mode\n"));
		}else if( (arg == NDIS_80211_AUTHMODE_WPA) ||
			  (arg == NDIS_80211_AUTHMODE_WPANONE) ||
			  (arg == NDIS_80211_AUTHMODE_WPAPSK) ||
			  (arg == NDIS_80211_AUTHMODE_AUTO) ){
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_HW_ERROR;
			NDISDBG((CE_NOTE,"Known authmodes in H/W :0x%8x\n", arg));
			goto done;
		}else {
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_HW_ERROR;
			NDISDBG((CE_NOTE,"Unknown authmodes in H/W :0x%8x\n", arg));
			goto done;
		}

		outfp->wldp_length = WIFI_BUF_OFFSET + 
			              sizeof (wl_authmode_t);
		outfp->wldp_result = WL_SUCCESS;

	} else if (cmd == WLAN_SET_PARAM) {
		arg = *(uintptr_t *)((wl_authmode_t *)(infp->wldp_buf));
		if ((arg != WL_OPENSYSTEM) &&
			(arg != WL_SHAREDKEY)) {
			NDISINF((CE_NOTE,"What's the parameter? : %d\n",arg));
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
			goto done;
		}
		
		switch(arg){
		case WL_OPENSYSTEM:
			arg = NDIS_80211_AUTHMODE_OPEN;			
			NDISDBG((CE_NOTE,"set auth to NDIS_80211_AUTHMODE_OPEN\n"));
			break;
		case WL_SHAREDKEY:
			arg = NDIS_80211_AUTHMODE_SHARED;			
			NDISDBG((CE_NOTE,"set auth to NDIS_80211_AUTHMODE_SHARED\n"));
			break;
		}

		len = sizeof(arg);
		rval = ndis_set_info(sc, OID_802_11_AUTHENTICATION_MODE, 
				(void *)&arg, &len);
		if (rval){
			NDISINF((CE_CONT, "set auth mode failed: %d\n", rval));
			kmem_free(buf, MAX_BUF_LEN);
			return (EINVAL);
		}else{
			isc->ic_bss->in_authmode = arg;
		}
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_SUCCESS;	
	} else {
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}


static int32_t
ndis_cfg_create_ibss(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	uint32_t arg;
	int rval;
	int ret;
	wldp_t *outfp;
	char *buf;
	int len;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		if(isc->ic_opmode == IEEE80211_M_IBSS)
			*(wl_create_ibss_t *)(outfp->wldp_buf) = B_TRUE;
		else if(isc->ic_opmode == IEEE80211_M_STA)
			*(wl_create_ibss_t *)(outfp->wldp_buf) = B_FALSE;
		else
			NDISDBG((CE_CONT,"%s: what's bss type now?\n",__func__));

		outfp->wldp_length = WIFI_BUF_OFFSET +
					sizeof(wl_create_ibss_t); 
		outfp->wldp_result = WL_SUCCESS;

	} else if (cmd == WLAN_SET_PARAM) {
		ret = (uint16_t)(*(wl_create_ibss_t *)(infp->wldp_buf));
		if (ret != 0 && ret != 1) {
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
			goto done;
		}
	
		if(ret){
			arg = NDIS_80211_NET_INFRA_IBSS;
			isc->ic_opmode = IEEE80211_M_IBSS;
		}else{
			arg = NDIS_80211_NET_INFRA_BSS;
			isc->ic_opmode = IEEE80211_M_STA;
		}
		len = sizeof(arg);
		rval = ndis_set_info(sc, OID_802_11_INFRASTRUCTURE_MODE, 
				(void *)&arg, &len);
		if (rval){
			NDISDBG((CE_CONT,"create ibss failed: %d\n", rval));
		}else
			NDISDBG((CE_CONT,"%s: create ibss successfully\n",__func__));
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_SUCCESS;	
	} else {
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}


static int32_t
ndis_cfg_rssi(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	ndis_wlan_bssid_ex	*bs = NULL;
	unsigned int i;
	wldp_t *outfp;
	char *buf;
	
	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		*(wl_rssi_t *)(outfp->wldp_buf) = sc->ndis_rssi;
		outfp->wldp_length = WIFI_BUF_OFFSET +
					sizeof(wl_rssi_t); 
		outfp->wldp_result = WL_SUCCESS;

	} else if (cmd == WLAN_SET_PARAM) {
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_READONLY;	
	} else {
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}

static int
ndis_cfg_linkstatus(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	uint16_t i;
	wldp_t *outfp;
	char *buf;
	int iret;
	ieee80211com_t *isc = (ieee80211com_t *)&(sc->ic);

	buf = kmem_zalloc(MAX_BUF_LEN, KM_SLEEP);
	outfp = (wldp_t *)buf;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	outfp->wldp_length = WIFI_BUF_OFFSET + sizeof (wl_linkstatus_t);
	outfp->wldp_result = WL_SUCCESS;

	if (cmd == WLAN_GET_PARAM) {
		if (!NDIS_INITIALIZED(sc)){
			*(wl_linkstatus_t *)(outfp->wldp_buf) = WL_NOTCONNECTED;
		}else{
			//NDIS_UNLOCK(sc);
			//delay(200);
			//NDIS_LOCK(sc);
			*(wl_linkstatus_t *)(outfp->wldp_buf) =
		    		(IEEE80211_S_RUN == isc->ic_state) ?
		    			WL_CONNECTED : WL_NOTCONNECTED;
		}
	} else if (cmd == WLAN_SET_PARAM) {
			outfp->wldp_length = WIFI_BUF_OFFSET;
			outfp->wldp_result = WL_NOTSUPPORTED;
	}

	for (i = 0; i < (outfp->wldp_length); i++)
		mi_mpprintf_putc((char *)mp, buf[i]);
	iret = (int)(outfp->wldp_result);
	kmem_free(buf, MAX_BUF_LEN);
	return (iret);
}

static int32_t
ndis_cfg_disconnect(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	wldp_t *infp;
	unsigned int i;
	int rval;
	int len=4;
	wldp_t *outfp;
	char *buf;
	ndis_80211_ssid ssid;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	if(mp) {
	    infp = (wldp_t *)mp->b_rptr;
	    bcopy(mp->b_rptr, buf,  sizeof (wldp_t));
	}
	outfp->wldp_length = WIFI_BUF_OFFSET; 
		
	if (cmd == WLAN_COMMAND) {
		NDISDBG((CE_CONT,"SET in disconnect\n"));
		//len = 0;
		//rval = ndis_set_info(sc, OID_802_11_DISASSOCIATE, NULL, &len);
		len = sizeof(ssid);
		memset(&ssid.ns_ssid[0], 0xff, 32);
		ssid.ns_ssidlen = 32;
		rval = ndis_set_info(sc, OID_802_11_SSID, (void *)&ssid, &len);
		sc->ic.ic_state = IEEE80211_S_INIT;
		sc->ic.ic_flags &= ~IEEE80211_F_PRIVACY;
		sc->ic.ic_bss->in_authmode = NDIS_80211_AUTHMODE_OPEN;
		sc->ic.ic_opmode = IEEE80211_M_STA;
		memset(&sc->ic.ic_des_essid, 0, IEEE80211_NWID_LEN);
		sc->ic.ic_des_esslen = 0;
		sc->ic.ic_crypto.cs_def_txkey = 0;
		memset(&sc->ic.ic_bss->in_essid,0,IEEE80211_NWID_LEN);
		sc->ic.ic_bss->in_esslen = 0;
		for (i = 0; i < IEEE80211_WEP_NKID; i++) {
			sc->ic.ic_nw_keys[i].wk_keylen = 0;
			bzero(&sc->ic.ic_nw_keys[i].wk_key,IEEE80211_KEYBUF_SIZE);
		}

		outfp->wldp_result = WL_SUCCESS;
	}else
		NWAMDBG((CE_CONT,"%s, not WLAN_COMMAND,error\n",__func__));
done:
	if(mp) {
	    for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	    }
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}

static int32_t
ndis_cfg_loaddefault(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	ieee80211com_t *isc = (ieee80211com_t *)&sc->ic;
	wldp_t *infp;
	uint16_t i;
	uint32_t arg;
	int rval;
	int /* was uint32_t */ len;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;

	infp = (wldp_t *)mp->b_rptr;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	if (cmd == WLAN_GET_PARAM) {
		outfp->wldp_length = WIFI_BUF_OFFSET; 
		outfp->wldp_result = WL_NOTSUPPORTED;
		NWAMDBG((CE_CONT,"%s: What do you want to get?\n",__func__));
	} else if (cmd == WLAN_SET_PARAM) {
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_NOTSUPPORTED;
		NWAMDBG((CE_CONT,"%s, What do you want to set?\n",__func__));
	} else if(cmd == WLAN_COMMAND){ 
		len = sizeof(arg);
		arg = NDIS_80211_RELOADDEFAULT_WEP;
		rval = ndis_set_info(sc, OID_802_11_RELOAD_DEFAULTS, 
				(void *)&arg, &len);
		if (rval)
			NDISINF((CE_CONT,"reload default failed: %d\n", rval)); 
		NWAMDBG((CE_CONT,"%s: WLAN_COMMAND\n",__func__));
		outfp->wldp_length = WIFI_BUF_OFFSET;
		outfp->wldp_result = WL_SUCCESS;		
	}else{
		NWAMDBG((CE_CONT,"%s: subcommand = 0x%x\n",
			__func__,cmd));
		kmem_free(buf, MAX_BUF_LEN);
		return (EINVAL);
	}

done:	
	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
						    
}

static int32_t
ndis_cfg_nodename(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{
	uint16_t i;
	wldp_t *outfp;
	char *buf;

	buf = kmem_zalloc(MAX_BUF_LEN, KM_NOSLEEP);
	if (buf == NULL) {
		NDISDBG((CE_CONT,
			   "ndis: can not alloc so much memory!(%d)\n",
			   MAX_BUF_LEN));
		return (ENOMEM);
	}
	outfp = (wldp_t *)buf;
	bcopy(mp->b_rptr, buf,  sizeof (wldp_t));

	outfp->wldp_length = WIFI_BUF_OFFSET; 
	outfp->wldp_result = WL_NOTSUPPORTED;

	for (i = 0; i < (outfp->wldp_length); i++) {
		mi_mpprintf_putc((char *)mp, buf[i]);
	}
	kmem_free(buf, MAX_BUF_LEN);	
	return WL_SUCCESS;	
}

static int32_t
ndis_getset(struct ndis_softc *sc, mblk_t *mp, uint32_t cmd)
{

	int ret = WL_SUCCESS;
	NDISIOCTL((CE_CONT,"--> %s :cmd = %s, subcmd == 0x%x\n",
		__func__, (cmd == WLAN_GET_PARAM)?"get":"set",
		((wldp_t *)mp->b_rptr)->wldp_id));	
	if(sc->id.id_state != ID_ATTACHED)
	    return EBUSY;
	switch(((wldp_t *)mp->b_rptr)->wldp_id) {
		case WL_SCAN:
			NDISIOCTL((CE_CONT,"ndis: WL_SCAN_DO\n"));
			ret = ndis_cfg_scan(sc, mp, cmd);
			break;
		case WL_ESSID:
			NDISIOCTL((CE_CONT,"ndis: WL_ESSID\n"));
			ret = ndis_cfg_essid(sc, mp, cmd);
			break;
		case WL_BSSID:
			NDISIOCTL((CE_CONT,"ndis: WL_BSSID\n"));
			ret = ndis_cfg_bssid(sc, mp, cmd);
			break;
		case WL_ENCRYPTION:
			NDISIOCTL((CE_CONT,"ndis: WL_ENCRYPTION\n"));
			ret = ndis_cfg_encryption(sc, mp, cmd);
			break;
		case WL_WEP_KEY_TAB:
			NDISIOCTL((CE_CONT,"ndis: WL_WEP_KEY_TAB\n"));
			ret = ndis_cfg_setwepkey(sc, mp, cmd);
			break;
		case WL_WEP_KEY_ID:
			NDISIOCTL((CE_CONT,"ndis: WL_WEP_KEY_ID\n"));
			ret = ndis_cfg_wepkeyindex(sc, mp, cmd);
			break;
		case WL_ESS_LIST:
			NDISIOCTL((CE_CONT,"ndis: WL_ESS_LIST\n"));
			ret = ndis_cfg_esslist(sc, mp, cmd);
			break;
		case WL_AUTH_MODE:
			NDISIOCTL((CE_CONT,"ndis: WL_AUTH_MODE\n"));
			ret = ndis_cfg_auth_mode(sc, mp, cmd); 
			break;
		case WL_POWER_MODE:
			NDISIOCTL((CE_CONT,"ndis: WL_POWER_MODE\n"));
			ret = ndis_cfg_power_mode(sc, mp, cmd); 
			break;
		case WL_BSS_TYPE:
			NDISIOCTL((CE_CONT,"ndis: WL_BSS_TYPE\n"));
			ret = ndis_cfg_bss_type(sc, mp, cmd); 
			break;
		case WL_PHY_CONFIG:
			NDISIOCTL((CE_CONT,"ndis: WL_PHY_CONFIG\n"));
			ret = ndis_cfg_phy(sc, mp, cmd); 
			break;
		case WL_DESIRED_RATES:
			NDISIOCTL((CE_CONT,"ndis: WL_DESIRED_RATES\n"));
			ret = ndis_cfg_rates(sc, mp, cmd);
			break;
		case WL_SUPPORTED_RATES:
			NDISIOCTL((CE_CONT,"ndis: WL_SUPPORTED_RATES\n"));
			ret = ndis_cfg_supported_rates(sc, mp, cmd); 
			break;
		case WL_CREATE_IBSS:
			NDISIOCTL((CE_CONT,"ndis: WL_CREATE_IBSS\n"));
			ret = ndis_cfg_create_ibss(sc, mp, cmd); 
			break;
		case WL_RSSI:
			NDISIOCTL((CE_CONT,"ndis: WL_RSSI\n"));
			ret = ndis_cfg_rssi(sc, mp, cmd); 
			break;
		case WL_LINKSTATUS:
			NDISIOCTL((CE_CONT,"ndis: WL_LINKSTATUS\n"));
			ret = ndis_cfg_linkstatus(sc, mp, cmd);
			break;
		case WL_RADIO:
			NDISIOCTL((CE_CONT,"ndis: WL_RADIO\n"));
			ret = ndis_cfg_radio(sc, mp, cmd); 
			break;
		case WL_DISASSOCIATE:
			NDISIOCTL((CE_CONT,"ndis: WL_DISASSOCIATE\n"));
			ret = ndis_cfg_disconnect(sc,mp,cmd);
			break;
		case WL_LOAD_DEFAULTS:
			NDISIOCTL((CE_CONT,"ndis: WL_LOAD_DEFAULTS\n"));
			ret = ndis_cfg_loaddefault(sc, mp, cmd); 
			break;
		case WL_NODE_NAME:
			NDISIOCTL((CE_CONT,"ndis: WL_NODE_NAME\n"));
			ret = ndis_cfg_nodename(sc, mp, cmd);
			break;
		default:
			NDISIOCTL((CE_CONT,"MISTAKE 0x%x",((wldp_t *)mp->b_rptr)->wldp_id));
			break;
	}
	NDISIOCTL((CE_CONT,"<--ndis_getset: cmd = 0x%x, subcmd = 0x%x, ret = 0x%x\n",
				cmd, ((wldp_t *)mp->b_rptr)->wldp_id, ret));
	return ret;
}


static void
ndis_wlan_ioctl(struct ndis_softc *sc, queue_t *wq, mblk_t *mp, uint32_t cmd)
{
	struct	iocblk	*iocp = (struct iocblk *)mp->b_rptr;
	wldp_t 	*infp;
	uint32_t len, ret;
	mblk_t	*mp1;

	/*
	 * sanity check
	 */
	if (iocp->ioc_count == 0 || !(mp1 = mp->b_cont)) {
		miocnak(wq, mp, 0, EINVAL);
		return;
	}

	/*
	 * assuming single data block
	 */
	if (mp1->b_cont) {
		freemsg(mp1->b_cont);
		mp1->b_cont = NULL;
	}

	/*
	 * we will overwrite everything
	 */
	mp1->b_wptr = mp1->b_rptr;

	infp = (wldp_t *)mp1->b_rptr;

	NDIS_LOCK(sc);
	ret = ndis_getset(sc, mp1, cmd);
	NDIS_UNLOCK(sc);
	len = msgdsize(mp1);
	miocack(wq, mp, len, ret);

}

static void ndis_m_ioctl(void *arg, queue_t *wq, mblk_t *mp)
{
	struct iocblk *iocp;
	uint32_t cmd, ret;
	struct ndis_softc *sc = (struct ndis_softc *)arg;
	boolean_t need_privilege = B_TRUE;

	/*
	 * Validate the command before bothering with the mutexen ...
	 */
	iocp = (struct iocblk *)mp->b_rptr;
	iocp->ioc_error = 0;
	cmd = iocp->ioc_cmd;
	switch (cmd) {
	default:
		NDISDBG((CE_CONT, "ndis_m_ioctl: unknown cmd 0x%x", cmd));
		miocnak(wq, mp, 0, EINVAL);
		return;
	case WLAN_GET_PARAM:
		need_privilege = B_TRUE;
		break;
	case WLAN_SET_PARAM:
	case WLAN_COMMAND:
		break;
	}
	/*
	 * Check net_config privilege
	 */
	if (need_privilege) {
		if (ret = secpolicy_net_config(iocp->ioc_cr, B_FALSE)) {
			miocnak(wq, mp, 0, ret);
			return;
		}
	}
	if(sc->id.id_state == ID_QUIESCED)
		return;
	
	ndis_wlan_ioctl(sc, wq, mp, cmd);
}


/*
 * Main transmit routine. To make NDIS drivers happy, we need to
 * transform mblk chains into NDIS packets and feed them to the
 * send packet routines. Most drivers allow you to send several
 * packets at once (up to the maxpkts limit). Unfortunately, rather
 * that accepting them in the form of a linked list, they expect
 * a contiguous array of pointers to packets.
 *
 * For those drivers which use the NDIS scatter/gather DMA mechanism,
 * we need to perform busdma work here. Those that use map registers
 * will do the mapping themselves on a buffer by buffer basis.
 */

static int
ndis_tx_begin(struct ndis_softc	*sc, mblk_t *mp)
{
	ndis_packet *p = NULL;
	int pcnt=0;
	mblk_t *mpx;

	if( pullupmsg(mp, -1) == 0){
		freemsg(mp);
		return GLD_FAILURE;
	}
	if(sc->ndis_txpending ){ 
		mpx = mac_wifi_header_backout(mp, NULL);
	
		sc->ic.ic_stats.is_tx_bytes += MBLKL(mpx);
		sc->ic.ic_stats.is_tx_frags += 1;		
		if (ndis_mtop(mpx, &sc->ndis_txarray[sc->ndis_txidx])) {
			NDISDBG((CE_CONT,"P5, idx[%d]alloced failed\n",sc->ndis_txidx));
			return(GLD_FAILURE);
		}

		/*
	 	* Save pointer to original mblk
	 	* so we can free it later.
	 	*/
		p = sc->ndis_txarray[sc->ndis_txidx];
		p->np_txidx = sc->ndis_txidx;
		p->np_m0 = mpx;
		p->np_oob.npo_status = NDIS_STATUS_PENDING;
		
		NDIS_INC(sc);
		sc->ndis_txpending--;
		pcnt++;
	}
	if(pcnt == 0){
		NDISDBG((CE_CONT,"No slot for tx\n"));
		return (GLD_FAILURE);
	}
	if (sc->ndis_maxpkts == 1){
		ndis_send_packet(sc, p);
	}else
		ndis_send_packets(sc, &p, 0x1);
	return(GLD_SUCCESS);
}

static int32_t
ndis_m_setmulti(void *arg, boolean_t on, const uint8_t *mca)
{

	struct ndis_softc *sc = (struct ndis_softc *)arg;
	uint32_t	retval;
	int /* was uint32_t */ len;
	uint32_t	mclist_len, mclistsz;
	uint8_t	*mc_p;
	uint32_t	packet_filter;
	unsigned int	i;

	NDIS_LOCK(sc);
	if((sc->id.id_state != ID_ATTACHED) || !NDIS_INITIALIZED(sc)) {
		NDIS_UNLOCK(sc);
		return EBUSY;
	}

	if(!sc->id.id_mcast) {	
		/*
		 * first init
		 */
		len = sizeof(mclistsz);
		ndis_get_info(sc, OID_802_3_MAXIMUM_LIST_SIZE, &mclistsz, &len);
	//	NDISINF((CE_CONT,"mclistsz=%d\n",mclistsz));
		mclist_len = ETHER_ADDR_LEN * mclistsz;
		sc->id.id_mcast = kmem_alloc(mclist_len, KM_NOSLEEP);

		if (sc->id.id_mcast == NULL) {
			NDISINF((CE_CONT,"NULL mcast\n"));
			sc->ndis_filter |= NDIS_PACKET_TYPE_ALL_MULTICAST;
			goto out;
		}
		bzero(sc->id.id_mcast, mclist_len);
		sc->id.id_mcast_count = 0;
		sc->id.id_mcast_len = mclist_len;
	}

	mc_p = sc->id.id_mcast;
	if(on) {
		for(i=0; i<sc->id.id_mcast_count; i++, mc_p+= (ETHER_ADDR_LEN)) {
			if(!ether_cmp(mca, mc_p))
				break;
		}
		if(i < (sc->id.id_mcast_count))  /* find the entry */
			goto out;
		
		mc_p = sc->id.id_mcast;

		/* find a free entry */
		if( (sc->id.id_mcast_count + 1) > (sc->id.id_mcast_len / ETHER_ADDR_LEN))
			goto out;
		mc_p += (ETHER_ADDR_LEN ) * (sc->id.id_mcast_count);
		ether_copy(mca, mc_p);
		sc->id.id_mcast_count ++;
	}else {
		for (i = 0; i < sc->id.id_mcast_count; i++, mc_p += (ETHER_ADDR_LEN )) {
			if (!ether_cmp(mca, mc_p))
				break;
		}
		if (i >= sc->id.id_mcast_count)  /* not found */
			goto out;

		/*
		 * delete the entry
		 */
		bzero(mc_p,ETHER_ADDR_LEN);
		sc->id.id_mcast_count --;
	}

	sc->ndis_filter |= NDIS_PACKET_TYPE_MULTICAST;

	len = (sc->id.id_mcast_count) * ETHER_ADDR_LEN;
	retval = ndis_set_info(sc, OID_802_3_MULTICAST_LIST, sc->id.id_mcast, &len);
	if (retval) {
		NDISINF((CE_CONT, "set mclist failed: %d\n", retval));
		sc->ndis_filter |= NDIS_PACKET_TYPE_ALL_MULTICAST;
		sc->ndis_filter &= ~NDIS_PACKET_TYPE_MULTICAST;
	}
	
out:
	packet_filter = sc->ndis_filter;
	len = sizeof(packet_filter);
	retval = ndis_set_info(sc, OID_GEN_CURRENT_PACKET_FILTER,
	    		&packet_filter, &len);
	if (retval) {
		NDISINF((CE_CONT, "set filter failed: %d\n", retval));
		NDIS_UNLOCK(sc);
		return (NDIS_FAILURE);
	}
	NDIS_UNLOCK(sc);
	return (NDIS_SUCCESS);

}

static int
ndis_m_prom(void *arg, boolean_t on)
{
	uint32_t		error, err;
	int /* was uint32_t */ len;
	struct ndis_softc *sc = (struct ndis_softc *)arg;
	err = GLD_SUCCESS;
	
	NDIS_LOCK(sc);
	if (on) {
//		sc->ndis_filter |= NDIS_PACKET_TYPE_ALL_MULTICAST;
		sc->ndis_filter |= NDIS_PACKET_TYPE_PROMISCUOUS;
	}else
		sc->ndis_filter &= ~NDIS_PACKET_TYPE_PROMISCUOUS;

	len = sizeof(sc->ndis_filter);
	error = ndis_set_info(sc, OID_GEN_CURRENT_PACKET_FILTER, &sc->ndis_filter, &len);
	if (error){
		NDISDBG((CE_CONT,"GLD_MAC_PROMIC_MULTI setting failed\n"));
		err = GLD_FAILURE;
	}
	NDIS_UNLOCK(sc);
	return (err);
}


static int
ndis_m_unicst(void *arg, const uint8_t *macaddr)
{
	NDISDBG((CE_CONT,"UNICAST Not Support yet\n"));
	return (GLD_SUCCESS);
}

/*ARGSUSED*/
static int
ndis_m_stat(void *arg, uint_t stat, uint64_t *val)
{
	struct ndis_softc *sc  = (struct ndis_softc *)arg;
	ieee80211com_t *ic = (ieee80211com_t *)&sc->ic;

	switch (stat) {
	case MAC_STAT_RBYTES:
		*val = ic->ic_stats.is_rx_bytes;
		break;
	case MAC_STAT_IPACKETS:
		*val = ic->ic_stats.is_rx_frags;
		break;
	case MAC_STAT_OBYTES:
		*val = ic->ic_stats.is_tx_bytes;
		break;
	case MAC_STAT_OPACKETS:
		*val = ic->ic_stats.is_tx_frags;
		break;
	default:
		return (DDI_ENOTSUP);
	}
	return (DDI_SUCCESS);
}

static mblk_t *
ndis_m_tx(void *arg, mblk_t *mp)
{
	struct ndis_softc *sc = (struct ndis_softc *)arg;
	mblk_t *next;

	NDIS_LOCK(sc);
	if(!NDIS_INITIALIZED(sc) || (sc->ndis_link == 0) ||
		(sc->id.id_state != ID_ATTACHED) || (sc->id.id_state == ID_QUIESCED)){
		NDIS_UNLOCK(sc);
		NDISDBG((CE_CONT,"ndis_m_tx : media not ready\n"));
		freemsgchain(mp);
		return (NULL);
	}

	while (mp != NULL) {
		next = mp->b_next;
		mp->b_next = NULL;

		if (ndis_tx_begin(sc, mp)!= 0) {
			mp->b_next = next;
			break;
		}

		mp = next;
	}
	NDIS_UNLOCK(sc);
	
	return (mp);
}

static int32_t
ndis_m_start(void *arg)
{
	struct ndis_softc	*sc = (struct ndis_softc *)arg;
	int			i, error;

	NDIS_LOCK(sc);
	if(sc->ndis_skip_start) {
		NDIS_UNLOCK(sc);
		return(GLD_SUCCESS);
	} else {
		sc->ndis_skip_start = 0x1;
	}
		
	/*
	 * Avoid reintializing the link unnecessarily.
	 * This should be dealt with in a better way by
	 * fixing the upper layer modules so they don't
	 * call ifp->if_init() quite as often.
	 */
	if (NDIS_INITIALIZED(sc)) {
		sc->ndis_skip_start = 0;
		NDIS_UNLOCK(sc);
		return(GLD_SUCCESS);
	}
	/*
	 * Cancel pending I/O and free all RX/TX buffers.
	 */
	NDIS_UNLOCK(sc);
	ndis_m_stop(sc);
	if (ndis_init_nic(sc) != DDI_SUCCESS) {
		sc->ndis_skip_start = 0;
		return(GLD_FAILURE);
	}
	/* Program the packet filter */

	sc->ndis_filter = NDIS_PACKET_TYPE_DIRECTED;

	sc->ndis_filter |= NDIS_PACKET_TYPE_BROADCAST;
	
	sc->ndis_filter &= ~NDIS_PACKET_TYPE_PROMISCUOUS;
		
	i = sizeof(sc->ndis_filter);

	NDIS_LOCK(sc);
	error = ndis_set_info(sc, OID_GEN_CURRENT_PACKET_FILTER,
	    (void *)&sc->ndis_filter, &i);

	if (error){
		cmn_err(CE_CONT, "set filter failed: %d\n", error);
		sc->ndis_skip_start = 0;
		NDIS_UNLOCK(sc);
		return(GLD_FAILURE);
	}

	sc->ndis_tickstop = 0;
	sc->ndis_txidx = 0;
	sc->ndis_txpending = sc->ndis_maxpkts;
	sc->ndis_link = 0;
        sc->ic.ic_state = IEEE80211_S_INIT;
        mac_link_update(sc->ic.ic_mach, LINK_STATE_DOWN);   

	//drv_usecwait(50);
	sc->ndis_skip_start = 0;
	/*
	 * Enable interrupts. We must grab the interrupt lock since the
	 * enable routine itself assumes that the loc is already held.
         */
	INTR_LOCK(sc);
	ndis_enable_intr(sc);
	INTR_UNLOCK(sc);
	NDIS_UNLOCK(sc);
	/*
	 * Some drivers don't set this value. The NDIS spec says
	 * the default checkforhang timeout is "approximately 2
	 * seconds." We use 3 seconds, because it seems for some
	 * drivers, exactly 2 seconds is too fast.
	 */

	sc->ndis_block->nmb_checkforhangsecs = 3;
	sc->ndis_tickstop = 0;
	if(sc->ndis_stat_ch == NULL) {
		ndis_tick(sc);
	}
	sc->ic.ic_txpowlimit = 0; 	
	return(GLD_SUCCESS);
}


/*
 * Stop the adapter and free any mbufs allocated to the
 * RX and TX lists.
 */

static void
ndis_m_stop(void *arg)
{
	struct ndis_softc *sc =(struct ndis_softc *)arg ;
	int rval;
	int len=4;
	ndis_80211_ssid ssid;
	
	NDIS_LOCK(sc);
	len = sizeof(ssid);
	memset(&ssid.ns_ssid[0], 0xff, 32);
	ssid.ns_ssidlen = 32;
	rval = ndis_set_info(sc, OID_802_11_SSID, (void *)&ssid, &len);
	
	NDIS_UNLOCK(sc);
	if(sc->ndis_stat_ch){
		untimeout(sc->ndis_stat_ch);
		sc->ndis_tickstop = 0x1;
		sc->ndis_stat_ch = NULL;
	}

	while(sc->ndis_intick)
		delay(20);
	ndis_halt_nic(sc);
	NDIS_LOCK(sc);
	sc->ndis_block->nmb_pendingreq = NULL;
	sc->ndis_link = 0;
	sc->ic.ic_state = IEEE80211_S_INIT;
        mac_link_update(sc->ic.ic_mach, LINK_STATE_DOWN);   
	NDIS_UNLOCK(sc);
	return;
}

mac_callbacks_t ndis_m_callbacks = {
	MC_IOCTL | MC_SETPROP | MC_GETPROP,
	ndis_m_stat,
	ndis_m_start,
	ndis_m_stop,
	ndis_m_prom,
	ndis_m_setmulti,
	ndis_m_unicst,
	ndis_m_tx,
	NULL,		/* Reserved, do not use */
	ndis_m_ioctl,	/* the order of ioctl and getcap changed as of nv105 */
	ndis_m_getcap,	/* getcap */
	ndis_m_open,	/* open */
	ndis_m_close,	/* close */
	ndis_setprop,	/* setprop */
	ndis_getprop,	/* getprop */
	ndis_m_propinf	/* propinfo */
};

/************************************************************
*
* All stuff for i80211. 
*
*************************************************************/

uint32_t
ieee80211_ieee2mhz(uint32_t chan, uint32_t  flags)
{
	if (flags & IEEE80211_CHAN_2GHZ) {	/* 2GHz band */
		if (chan == 14)
			return 2484;
		if (chan < 14)
			return 2407 + chan*5;
		else
			return 2512 + ((chan-15)*20);
	} else if (flags & IEEE80211_CHAN_5GHZ) {/* 5Ghz band */
		return 5000 + (chan*5);
	} else {				/* either, guess */
		if (chan == 14)
			return 2484;
		if (chan < 14)			/* 0-13 */
			return 2407 + chan*5;
		if (chan < 27)			/* 15-26 */
			return 2512 + ((chan-15)*20);
		return 5000 + (chan*5);
	}
}

/*
 * Convert channel to IEEE channel number.
 */
uint32_t
ieee80211_chan2ieee(ieee80211com_t *isc, struct ieee80211_channel *ch)
{
	if (isc->ic_sup_channels <= ch &&
	    ch <= &isc->ic_sup_channels[IEEE80211_CHAN_MAX])
		return (ch - isc->ic_sup_channels);
	else if (ch == IEEE80211_CHAN_ANYC)
		return (IEEE80211_CHAN_ANY);
	else {
		NDISDBG((CE_CONT,"wlan: invalid channel freq %u flags %x\n",
		    ch->ich_freq, ch->ich_flags));
		return (0);		/* XXX */
	}
}

/*
 * Convert MHz frequency to IEEE channel number.
 */
uint32_t
ieee80211_mhz2ieee(uint32_t freq, uint32_t flags)
{
	if (flags & IEEE80211_CHAN_2GHZ) {	/* 2GHz band */
		if (freq == 2484)
			return (14);
		if (freq < 2484)
			return ((freq - 2407) / 5);
		else
			return (15 + ((freq - 2512) / 20));
	} else if (flags & IEEE80211_CHAN_5GHZ) {	/* 5Ghz band */
		return ((freq - 5000) / 5);
	} else {				/* either, guess */
		if (freq == 2484)
			return (14);
		if (freq < 2484)
			return ((freq - 2407) / 5);
		if (freq < 5000)
			return (15 + ((freq - 2512) / 20));
		return ((freq - 5000) / 5);
	}
}

/* regs access attributes */
static ddi_device_acc_attr_t ndis_reg_accattr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC,
	DDI_DEFAULT_ACC
};
/*
 * Probe for an NDIS device. Check the PCI vendor and device
 * IDs against our list and return a device name if we find a match.
 */
static int
ndis_probe_pci(dev_info_t *dev, struct ndis_softc *sc)
{
	struct ndis_pci_type	*t;
	driver_object		*drv;
	uint32_t		tmp1,tmp2,tmp3;
	uint32_t		i=0;
	caddr_t			cfg_base;

	t = &ndis_devs[0];
	drv = windrv_lookup(sc->ndis_drv_data);

	if (drv == NULL)
		return (ENXIO);

	if(ddi_regs_map_setup(sc->id.id_dev, 0,&cfg_base,0,256,
		&ndis_reg_accattr,&sc->id.id_cfg_hdl) != DDI_SUCCESS){
		cmn_err(CE_CONT,"PCI Configure Head Setup failed\n");
		return (DDI_FAILURE);
	}

	while (t->ndis_name != NULL) {
		if (((tmp1 = ddi_get16(sc->id.id_cfg_hdl,(uint16_t *)(cfg_base + PCI_CONF_VENID))) 
			== t->ndis_vid) &&
		    ((tmp2 = ddi_get16(sc->id.id_cfg_hdl,(uint16_t *)(cfg_base + PCI_CONF_DEVID))) 
			== t->ndis_did) 
		     ){
			NDISINF((CE_CONT,"RealVendor = 0x%x,RealDevice = 0x%x,RealSubsys = 0x%x\n",
				tmp1,tmp2,tmp3));
			NDISINF((CE_CONT,"InfVendor = 0x%x,InfDevice= 0x%x, InfSubsys = 0x%x\n",
				t->ndis_vid,t->ndis_did,t->ndis_subsys));

			windrv_create_pdo(drv, dev);
			ddi_regs_map_free(&sc->id.id_cfg_hdl);	
			return (0);
		}
		i++;
		t = &ndis_devs[i];
	}
	ddi_regs_map_free(&sc->id.id_cfg_hdl);	
	return (ENXIO);
}

static int
ndis_attach_pci(struct ndis_softc *sc)
{
	struct ndis_pci_type	*t;
	int			devidx = 0, defidx = 0;
	caddr_t		cfg_base;

	t = ndis_devs;
	
	if(ddi_regs_map_setup(sc->id.id_dev, 0,&cfg_base,0,256,
		&ndis_reg_accattr,&sc->id.id_cfg_hdl)
		!= DDI_SUCCESS)
		return (DDI_FAILURE);

	ddi_put16(sc->id.id_cfg_hdl, (uint16_t *)(cfg_base + 0x4),
		PCI_COMM_MAE|PCI_COMM_ME|PCI_COMM_IO);

	while(t->ndis_name != NULL) {
		if ((ddi_get16(sc->id.id_cfg_hdl,(uint16_t *)(cfg_base + PCI_CONF_VENID)) 
			== t->ndis_vid) &&
		    (ddi_get16(sc->id.id_cfg_hdl,(uint16_t *)(cfg_base + PCI_CONF_DEVID)) 
			== t->ndis_did)) {
			if (t->ndis_subsys == 0)
				defidx = devidx;
			else {
				if (t->ndis_subsys == 
					ddi_get32(sc->id.id_cfg_hdl,
					(uint32_t *)(cfg_base+PCI_CONF_SUBVENID)));
					break;
			}
		}
		t++;
		devidx++;
	}

	if (ndis_devs[devidx].ndis_name == NULL)
		sc->ndis_devidx = defidx;
	else
		sc->ndis_devidx = devidx;

	sc->ndis_iftype = PCIBus;
	ddi_regs_map_free(&sc->id.id_cfg_hdl);
	return(DDI_SUCCESS);
}

int ndis_xmit(ieee80211com_t *ic, mblk_t *blk, uint8_t x)
{
	cmn_err(CE_WARN,"Unexpected ndis_xmit() call\n");
	return (ENOTSUP);
}

/*
 * Attach the interface. Allocate softc structures, do ifmedia
 * setup attach.
 */

int ndis_initted;

int
ndis_attach(dev_info_t *devinfo, ddi_attach_cmd_t cmd)
{
	uint8_t			eaddr[ETHER_ADDR_LEN];
	int			len;
	int			i;
	unsigned int		ui;
	char			strbuf[32];
	struct ndis_softc	*sc;
	int32_t 		instance;
	driver_object		*drv;
	device_object		*pdo;
	uint32_t		err;
	mac_register_t		*macp;
#ifndef LINKEDIMAGE
	vm_offset_t		data;
#endif
	wifi_data_t		wd = { 0 };

	TOP_LOCK();
	if(ndis_initted == 1) {
	    TOP_UNLOCK();
	    return DDI_SUCCESS;
	}
	ndis_initted = 1;

	instance = ddi_get_instance(devinfo);
	sc = ddi_get_soft_state(ndis_soft_state_p, instance);
	switch (cmd) {
	case DDI_RESUME:
	    if(sc != NULL) {
		if(sc->id.id_state == ID_SUSPENDED) {
		    sc->id.id_state = ID_ATTACHED;
		    sc->ndis_skip_start = 0;
		    ndis_m_start(sc);
		}
	    }
	    TOP_UNLOCK();
	    return (DDI_SUCCESS);
	case DDI_ATTACH:
	    if (sc != NULL) {
		TOP_UNLOCK();
		return DDI_SUCCESS;
	    }
	    break;
	default:
	    TOP_UNLOCK();
	    return DDI_FAILURE;
	}
	if (ddi_soft_state_zalloc(ndis_soft_state_p, instance) != DDI_SUCCESS) {
		NDISINF((CE_WARN,("ndis_attach: Unable to alloc softstate\n")));
		goto failure_0;
	}
	sc = ddi_get_soft_state(ndis_soft_state_p, instance);
	sc->id.id_dev = devinfo;

#ifdef LINKEDIMAGE
	if (*(uint32_t*)&drv_data[drv_data_len] != drv_data_len) {
		cmn_err(CE_WARN,"Badly compiled driver (size %ld, should be %ld)\n",
			(long)*(uint32_t*)&drv_data[drv_data_len],
			(long)drv_data_len);
	}
	windrv_load(1, (vm_offset_t)drv_data, 0);
	ndis_rxeof_wrap = ndis_rxeof;
	ndis_txeof_wrap = ndis_txeof;
	ndis_linksts_wrap = ndis_linksts;
	ndis_linksts_done_wrap = ndis_linksts_done;
	sc->ndis_drv_data = (vm_offset_t)drv_data;
#else
	data = (vm_offset_t) kmem_zalloc(drv_data_len, KM_NOSLEEP);
	bcopy((char *)drv_data, (char *)data, drv_data_len);
	windrv_load(1, data, 0);
	windrv_wrap((funcptr)ndis_rxeof, &ndis_rxeof_wrap);
	windrv_wrap((funcptr)ndis_txeof, &ndis_txeof_wrap);
	windrv_wrap((funcptr)ndis_linksts, &ndis_linksts_wrap);
	windrv_wrap((funcptr)ndis_linksts_done,&ndis_linksts_done_wrap);
	sc->ndis_drv_data = data;
#endif

        ndis_debug = ddi_prop_get_int(DDI_DEV_T_ANY, devinfo, DDI_PROP_DONTPASS,"debug-level",0);


	ndis_attach_pci(sc);
	if(ndis_probe_pci(devinfo,sc) != 0){
		NDISINF((CE_CONT,"probe pci failed use 0\n"));
		goto failure_2;		
	}

	/*mutex and interrupt resource */
	mutex_init(&sc->id.id_topmtx, NULL, MUTEX_DRIVER, NULL);

	if (ddi_get_iblock_cookie(devinfo, 0, &sc->id.id_iblock)
	    != DDI_SUCCESS) {
		NDISINF((CE_WARN,"ndis_attach: Can not get iblock cookie for INT\n"));
		goto failure_3;
	}

	mutex_init(&sc->id.id_intrmtx, NULL, MUTEX_DRIVER,(void *)sc->id.id_iblock);

	if (ddi_add_intr(devinfo, 0, &sc->id.id_iblock, 0, ndis_intr, (caddr_t)sc) != DDI_SUCCESS) {
		NDISINF((CE_WARN,"ndis_attach: Can not set intr for if_ndis driver\n"));
		goto failure_4;
	}

	sc->ndis_regvals = ndis_regvals;
	/* Create sysctl registry nodes */
	if(ndis_create_sysctls(sc) == EINVAL)
	{
		NDISINF((CE_CONT,"create sysctls failed\n"));
		goto failure_5;
	}
	drv = windrv_lookup(sc->ndis_drv_data);
	if (!drv || !drv->dro_driverext) {
		cmn_err(CE_CONT,"Could not find the loaded driver\n");
		goto failure_5;
	}
	pdo = windrv_find_pdo(drv, devinfo);

	/* Set up driver image in memory. */
	if (NdisAddDevice(drv, pdo, sc) != STATUS_SUCCESS)
	{
		NDISINF((CE_CONT,"load driver failed\n"));
		goto failure_6;
	}

	/* Tell the user what version of the API the driver is using. */
	cmn_err(CE_CONT, "NDIS API version: %d.%d\n",
	    sc->ndis_chars->nmc_version_major,
	    sc->ndis_chars->nmc_version_minor);

	/* Do resource conversion. */
	if(ndis_convert_res(sc) != DDI_SUCCESS)
	{
		NDISINF((CE_CONT,"convert resource failed\n"));
		goto failure_7;
	}

	/* Install our RX and TX interrupt handlers. */
	sc->ndis_block->nmb_senddone_func = ndis_txeof_wrap;
	sc->ndis_block->nmb_pktind_func = ndis_rxeof_wrap;
	
	/* Call driver's init routine. */
	if (ndis_init_nic(sc) != DDI_SUCCESS) {
		NDISINF((CE_CONT, "init handler failed\n"));
		goto failure_8;
	}

	cmn_err(CE_CONT,"%s: Congratulations, Attached Okay!\n",__func__);
	/*
	 * Create minor node of type DDI_NT_NET_WIFI
	 */
	(void) snprintf(strbuf, sizeof (strbuf), "%s%d",
	    NDIS_DRIVER_NAME, instance);
	err = ddi_create_minor_node(devinfo, strbuf, S_IFCHR,
	    instance + 1, DDI_NT_NET_WIFI, 0);
	if (err != DDI_SUCCESS)
		NDISDBG((CE_WARN,  "ndis_attach(): ddi_create_minor_node() failed\n"));

	/*
	 * Get station address from the driver.
	 */
	len = sizeof(eaddr);
	ndis_get_info(sc, OID_802_3_CURRENT_ADDRESS, &eaddr, &len);

	bcopy(eaddr, (char *)sc->ic.ic_macaddr, ETHER_ADDR_LEN);
	cmn_err(CE_CONT,"Wlan MAC:%02x:%02x:%02x:%02x:%02x:%02x\n",
		eaddr[0],eaddr[1],eaddr[2],eaddr[3],eaddr[4],eaddr[5]);

	/*
	 * Figure out if we're allowed to use multipacket sends
	 * with this driver, and if so, how many.
	 */

	if (sc->ndis_chars->nmc_sendsingle_func &&
	    sc->ndis_chars->nmc_sendmulti_func == NULL) {
		sc->ndis_maxpkts = 1;
	} else {
		len = sizeof(sc->ndis_maxpkts);
		ndis_get_info(sc, OID_GEN_MAXIMUM_SEND_PACKETS,
		    &sc->ndis_maxpkts, &len);
	}
	NDISDBG((CE_CONT,"Maxpkts are %d\n",sc->ndis_maxpkts));
	sc->ndis_txpending =sc->ndis_maxpkts;

	sc->ndis_txarray = kmem_zalloc(sizeof(ndis_packet *) *
	    sc->ndis_maxpkts, KM_NOSLEEP);
	
	/* Allocate a pool of ndis_packets for TX encapsulation. */

	{
		ndis_status st;

		MSCALL4(NdisAllocatePacketPool, &st, &sc->ndis_txpool,
		   sc->ndis_maxpkts+10, PROTOCOL_RESERVED_SIZE_IN_PACKET);

		if (st != NDIS_STATUS_SUCCESS) {
			sc->ndis_txpool = NULL;
			NDISINF((CE_CONT, "failed to allocate TX packet pool\n"));
			goto failure_9;
		}
	}

	for(i = 0; i < sc->ndis_maxpkts;i++){
		ndis_status /* was int */ status;
		MSCALL3(NdisAllocatePacket, &status, &sc->ndis_txarray[i],
				sc->ndis_txpool);
	}

	/*
	 * Provide initial settings for the WiFi plugin; whenever this
	 * information changes, we need to call mac_pdata_update()
	 */
  	wd.wd_secalloc = WIFI_SEC_NONE;
  	wd.wd_opmode = IEEE80211_M_STA;
	
	macp = mac_alloc(MAC_VERSION);
	if (macp == NULL) {
		NDISDBG((CE_NOTE, "%s: MAC version mismatch\n",__func__));
		goto failure_9;
	}

	macp->m_type_ident	= MAC_PLUGIN_IDENT_WIFI;
	macp->m_driver		= sc;
	macp->m_dip		= devinfo;
	macp->m_src_addr	= sc->ic.ic_macaddr;
	macp->m_callbacks	= &ndis_m_callbacks;
	macp->m_min_sdu		= 0;
	macp->m_max_sdu		= IEEE80211_MTU;
  	macp->m_pdata		= &wd;
  	macp->m_pdata_size	= sizeof (wd);
	macp->m_priv_props	= &priv_props;

	err = mac_register(macp, &sc->ic.ic_mach);
	mac_free(macp);
	if (err != 0) {
		NDISDBG((CE_NOTE, "%s: mac_register err\n",__func__));
		goto failure_9;
	}
	mac_link_update(sc->ic.ic_mach, LINK_STATE_DOWN);

	sc->ndis_oidcnt = 0;
	/* Get supported oid list. */
	ndis_get_supported_oids(sc, &sc->ndis_oids, &sc->ndis_oidcnt);

	/*
	 * See if the OID_802_11_CONFIGURATION OID is
	 * supported by this driver. If it is, then this an 802.11
	 * wireless driver, and we should set up media for wireless.
	 */
	for (i = 0; i < sc->ndis_oidcnt; i++) {
		if (sc->ndis_oids[i] == OID_802_11_CONFIGURATION) {
			sc->ndis_80211++;
			break;
		}
	}

	/* Do media setup */
	if (sc->ndis_80211) {
		struct ieee80211com	*ic = &sc->ic;
		ndis_80211_rates_ex	rates;
		struct ndis_80211_nettype_list *ntl;
		uint32_t		arg;
		int			r;

		ic->ic_xmit = ndis_xmit; /* mandatory before attach() */
		ieee80211_attach(ic);
	        ic->ic_phytype = IEEE80211_T_DS;
		ic->ic_opmode = IEEE80211_M_STA;
		ic->ic_state = IEEE80211_S_INIT;
		ic->ic_modecaps = (1<<IEEE80211_MODE_AUTO);
		len = 0;
		r = ndis_get_info(sc, OID_802_11_NETWORK_TYPES_SUPPORTED,
		    NULL, &len);
		if (r != ENOSPC)
			goto nonettypes;
		ntl = kmem_zalloc(len, KM_NOSLEEP);
		r = ndis_get_info(sc, OID_802_11_NETWORK_TYPES_SUPPORTED,
		    ntl, &len);
		if (r != 0) {
			kmem_free(ntl, len);
			goto nonettypes;
		}

		for (ui = 0; ui < ntl->ntl_items; ui++) {
			switch (ntl->ntl_type[ui]) {
			case NDIS_80211_NETTYPE_11FH:
			case NDIS_80211_NETTYPE_11DS:
				NDISDBG((CE_WARN,"Support 80211B \n"));
				ic->ic_modecaps |= (1<<IEEE80211_MODE_11B);
				break;
			case NDIS_80211_NETTYPE_11OFDM5:
				NDISDBG((CE_WARN,"Support 80211A \n"));
				ic->ic_modecaps |= (1<<IEEE80211_MODE_11A);
				break;
			case NDIS_80211_NETTYPE_11OFDM24:
				ic->ic_modecaps |= (1<<IEEE80211_MODE_11G);
				NDISDBG((CE_WARN,"Support 80211G\n"));
				break;
			default:
				NDISDBG((CE_WARN,"new nettype %d\n",ntl->ntl_type[i]));
				break;
			}
		}
		kmem_free(ntl, len);
nonettypes:
		len = sizeof(rates);
		bzero((char *)&rates, len);
		r = ndis_get_info(sc, OID_802_11_SUPPORTED_RATES,
		    (void *)rates, &len);
		if (r)
			NDISDBG((CE_WARN, "get rates failed: 0x%x\n", r));
		for(len = 0; len <16;  len++)
			NDISDBG((CE_WARN, "rate[len] = %d\n",rates[len]&0x7f));
		/*
		 * Since the supported rates only up to 8 can be supported,
		 * if this is not 802.11b we're just going to be faking it
		 * all up to heck.
		 */

#define TESTSETRATE(x, y)						\
	do {								\
		int			i;				\
		for (i = 0; i < ic->ic_sup_rates[x].ir_nrates; i++) {	\
			if (ic->ic_sup_rates[x].ir_rates[i] == (y))	\
				break;					\
		}							\
		if (i == ic->ic_sup_rates[x].ir_nrates) {		\
			ic->ic_sup_rates[x].ir_rates[i] = (y);		\
			ic->ic_sup_rates[x].ir_nrates++;		\
		}							\
	} while (0)

#define SETRATE(x, y)	\
	ic->ic_sup_rates[x].ir_rates[ic->ic_sup_rates[x].ir_nrates] = (y)
#define INCRATE(x)	\
	ic->ic_sup_rates[x].ir_nrates++

		ic->ic_curmode = IEEE80211_MODE_AUTO;
		if (ic->ic_modecaps & (1<<IEEE80211_MODE_11A))
			ic->ic_sup_rates[IEEE80211_MODE_11A].ir_nrates = 0;
		if (ic->ic_modecaps & (1<<IEEE80211_MODE_11B))
			ic->ic_sup_rates[IEEE80211_MODE_11B].ir_nrates = 0;
		if (ic->ic_modecaps & (1<<IEEE80211_MODE_11G))
			ic->ic_sup_rates[IEEE80211_MODE_11G].ir_nrates = 0;
		for (i = 0; i < len; i++) {
			switch (rates[i] & IEEE80211_RATE_VAL) {
			case 2:
			case 4:
			case 11:
			case 10:
			case 22:
				if (!(ic->ic_modecaps &
				    (1<<IEEE80211_MODE_11B))) {
					/* Lazy-init 802.11b. */
					ic->ic_modecaps |=
					    (1<<IEEE80211_MODE_11B);
					ic->ic_sup_rates[IEEE80211_MODE_11B].ir_nrates = 0;
				}
				SETRATE(IEEE80211_MODE_11B, rates[i]);
				INCRATE(IEEE80211_MODE_11B);
				break;
			default:
				if (ic->ic_modecaps & (1<<IEEE80211_MODE_11A)) {
					SETRATE(IEEE80211_MODE_11A, rates[i]);
					INCRATE(IEEE80211_MODE_11A);
				}
				if (ic->ic_modecaps & (1<<IEEE80211_MODE_11G)) {
					SETRATE(IEEE80211_MODE_11G, rates[i]);
					INCRATE(IEEE80211_MODE_11G);
				}
				break;
			}
		}

		/*
		 * If the hardware supports 802.11g, it most
		 * likely supports 802.11b and all of the
		 * 802.11b and 802.11g speeds, so maybe we can
		 * just cheat here.  Just how in the heck do
		 * we detect turbo modes, though?
		 */
		if (ic->ic_modecaps & (1<<IEEE80211_MODE_11B)) {
			TESTSETRATE(IEEE80211_MODE_11B,
			    IEEE80211_RATE_BASIC|2);
			TESTSETRATE(IEEE80211_MODE_11B,
			    IEEE80211_RATE_BASIC|4);
			TESTSETRATE(IEEE80211_MODE_11B,
			    IEEE80211_RATE_BASIC|11);
			TESTSETRATE(IEEE80211_MODE_11B,
			    IEEE80211_RATE_BASIC|22);
		}
		if (ic->ic_modecaps & (1<<IEEE80211_MODE_11G)) {
			TESTSETRATE(IEEE80211_MODE_11G, 47);
			TESTSETRATE(IEEE80211_MODE_11G, 72);
			TESTSETRATE(IEEE80211_MODE_11G, 96);
			TESTSETRATE(IEEE80211_MODE_11G, 108);
		}
		if (ic->ic_modecaps & (1<<IEEE80211_MODE_11A)) {
			TESTSETRATE(IEEE80211_MODE_11A, 47);
			TESTSETRATE(IEEE80211_MODE_11A, 72);
			TESTSETRATE(IEEE80211_MODE_11A, 96);
			TESTSETRATE(IEEE80211_MODE_11A, 108);
		}
#undef SETRATE
#undef INCRATE
		/*
		 * Taking yet more guesses here.
		 */
		for (i = 1; i < IEEE80211_CHAN_MAX; i++) {
			int chanflag = 0;

			if (ic->ic_sup_rates[IEEE80211_MODE_11G].ir_nrates)
					/* IEEE80211_CHAN_G */
				chanflag |= IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_DYN;
			if (i <= 14)	/* IEEE80211_CHAN_B */
				chanflag |= IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_CCK;
			if (ic->ic_sup_rates[IEEE80211_MODE_11A].ir_nrates &&
			    i > 14)	/* IEEE80211_CHAN_A */
				chanflag |= IEEE80211_CHAN_5GHZ|IEEE80211_CHAN_OFDM;
			if (chanflag == 0)
				break;
			ic->ic_sup_channels[i].ich_freq =
			    ieee80211_ieee2mhz(i, chanflag);
			ic->ic_sup_channels[i].ich_flags = chanflag;
		}

		i = sizeof(arg);
		r = ndis_get_info(sc, OID_802_11_WEP_STATUS, &arg, &i);
		if (arg != NDIS_80211_WEPSTAT_NOTSUPPORTED){
			ic->ic_caps |= IEEE80211_C_WEP;
			NDISDBG((CE_WARN,"wep support\n"));
		}
		i = sizeof(arg);
		r = ndis_get_info(sc, OID_802_11_TX_POWER_LEVEL, &arg, &i);
		if (r == 0){
			ic->ic_caps |= IEEE80211_C_TXPMGT;
			ic->ic_txpowlimit = 0; 
		}
		bcopy(eaddr, (char *)sc->ic.ic_macaddr, sizeof(eaddr));
		/* FIXME */
	//	ic->ic_ibss_chan = kmem_zalloc(sizeof(struct ieee80211channel),KM_NOSLEEP);
		ic->ic_ibss_chan = &ic->ic_sup_channels[1];
	//	ic->ic_ibss_chan = IEEE80211_CHAN_ANYC;
		ic->ic_bss = kmem_zalloc(sizeof(struct ieee80211_node),KM_NOSLEEP);
		ic->ic_bss->in_chan = ic->ic_ibss_chan;
	} 

	/* Override the status handler so we can detect link changes. */
	sc->ndis_block->nmb_status_func = ndis_linksts_wrap;
	sc->ndis_block->nmb_statusdone_func = ndis_linksts_done_wrap;
	
	/* We're done talking to the NIC for now; halt it and wait for plumb. */
	ndis_halt_nic(sc);
	sc->id.id_state = ID_ATTACHED;
	NDISDBG((CE_CONT,"Halt Wlan Adapter, Waiting for Plumb\n"));
	TOP_UNLOCK();
	return (DDI_SUCCESS);

failure_9:
failure_8:
failure_7:
failure_6:
	ndis_flush_sysctls(sc);
failure_5:
	ddi_remove_intr(devinfo,0,sc->id.id_iblock);
failure_4:
	mutex_destroy(&sc->id.id_intrmtx);
failure_3:	
	mutex_destroy(&sc->id.id_topmtx);
failure_2:
	ddi_soft_state_free(ndis_soft_state_p, ddi_get_instance(devinfo));
failure_1:
#ifndef LINKEDIMAGE
	kmem_free((void *)sc->ndis_drv_data, drv_data_len);
#endif
failure_0:
	TOP_UNLOCK();
	return(DDI_FAILURE);
}

/*
 * Shutdown hardware and free up resources. This can be called any
 * time after the mutex has been initialized. It is called in both
 * the error case in attach and the normal detach case so it needs
 * to be careful about only freeing resources that have actually been
 * allocated.
 */

static int
ndis_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	struct ndis_softc	*sc;
	int ret;
	int test = 5;

	sc = ddi_get_soft_state(ndis_soft_state_p, ddi_get_instance(dip));
	if (sc == NULL) return(DDI_FAILURE);
	NDIS_LOCK(sc);
	switch (cmd) {
	case DDI_SUSPEND: 
		ndis_initted = 0;
		if((sc->id.id_state == ID_ATTACHED) && NDIS_INITIALIZED(sc)) {
			sc->id.id_state = ID_SUSPENDED;
			if(sc->ic.ic_state == IEEE80211_S_RUN) {
				(void) ndis_cfg_disconnect(sc, NULL, WLAN_COMMAND);
				sc->ndis_link = 0;
				sc->ic.ic_state = IEEE80211_S_ASSOC;
				mac_link_update(sc->ic.ic_mach, LINK_STATE_DOWN);
				NDIS_UNLOCK(sc);
				ndis_m_stop(sc);
			} else
			    NDIS_UNLOCK(sc);
		} else {
			sc->id.id_state = ID_SUSPENDED;
			NDIS_UNLOCK(sc);
		}
		return(DDI_SUCCESS);

	case DDI_DETACH:
		NDIS_UNLOCK(sc);
		ndis_m_stop(sc);
		NDIS_LOCK(sc);
		ret = mac_unregister(sc->ic.ic_mach);
		if (ret) {
			NDIS_UNLOCK(sc);
			return (ret);
		}
		sc->id.id_state = ID_SUSPENDED; 
		if(sc->id.id_mcast)
			kmem_free(sc->id.id_mcast,sc->id.id_mcast_len);
		if (sc->ndis_txarray)
			kmem_free(sc->ndis_txarray, sizeof(ndis_packet *) *
	   			sc->ndis_maxpkts);
		ndis_unload_driver(sc);
		MSCALL1(ExFreePool, sc->ndis_txpool);
		ddi_remove_intr(dip,0,sc->id.id_iblock);
		NDIS_UNLOCK(sc);
		delay(50);
		mutex_destroy(&sc->id.id_topmtx);
freelock:
		NDISDBG((CE_CONT,"Free mutex intrmtx\n"));
		if(mutex_tryenter(&sc->id.id_intrmtx) == 0){
			mutex_exit(&sc->id.id_intrmtx);
			delay(5);
			if(test--)
				goto freelock;
		} else {
			delay(10);
			goto freelock;
		}
		mutex_destroy(&sc->id.id_intrmtx);
		windrv_unload(1, sc->ndis_drv_data, 0);

		ndis_flush_sysctls(sc);		

#ifdef LINKEDIMAGE
		ndis_rxeof_wrap = (ndis_rxeof_handler)NULL;
		ndis_txeof_wrap = (ndis_senddone_func)NULL;
		ndis_linksts_wrap = (ndis_status_handler)NULL;
		ndis_linksts_done_wrap = (ndis_statusdone_handler)NULL;
#else
		kmem_free((char *)sc->ndis_drv_data, drv_data_len);
		windrv_unwrap(ndis_rxeof_wrap);
		windrv_unwrap(ndis_txeof_wrap);
		windrv_unwrap(ndis_linksts_wrap);
		windrv_unwrap(ndis_linksts_done_wrap);
#endif
		ddi_remove_minor_node(dip, NULL);
		sc->id.id_state = ID_DETACHED;
		ddi_soft_state_free(ndis_soft_state_p, ddi_get_instance(dip));
		return(DDI_SUCCESS);
	default:
		return (DDI_FAILURE);
	}
}

static int
ndis_quiesce(dev_info_t *dip)
{
	struct ndis_softc	*sc;

	sc = ddi_get_soft_state(ndis_soft_state_p, ddi_get_instance(dip));
	if (sc) {
		ndis_halt_nic(sc);
		sc->id.id_state = ID_QUIESCED;
	}
	return(DDI_SUCCESS);
}

/*
 * A frame has been uploaded: pass the resulting mbuf chain up to
 * the higher level protocols.
 *
 * When handling received NDIS packets, the 'status' field in the
 * out-of-band portion of the ndis_packet has special meaning. In the
 * most common case, the underlying NDIS driver will set this field
 * to NDIS_STATUS_SUCCESS, which indicates that it's ok for us to
 * take posession of it. We then change the status field to
 * NDIS_STATUS_PENDING to tell the driver that we now own the packet,
 * and that we will return it at some point in the future via the
 * return packet handler.
 *
 * If the driver hands us a packet with a status of NDIS_STATUS_RESOURCES,
 * this means the driver is running out of packet/buffer resources and
 * wants to maintain ownership of the packet. In this case, we have to
 * copy the packet data into local storage and let the driver keep the
 * packet.
 */
MSENTRY3(void ndis_rxeof, adapter, packets, pktcnt)
	ndis_miniport_block	*adapter;
	ndis_packet		**packets;
	uint32_t		pktcnt;
{
	struct ndis_softc	*sc;
	ndis_miniport_block	*block;
	ndis_packet		*p;
	mblk_t		*m0, *m=NULL;
	unsigned int	i;
	wifi_data_t	wd={0};
	
	block = (ndis_miniport_block *)adapter;
	sc = (struct ndis_softc *)(block->nmb_ifp);

	if (sc->id.id_state == ID_QUIESCED) {
		RESTORE();
		return;
	}

//	NDISDBG((CE_CONT, "ndis_rxeof,pcount %d\n",pktcnt));
	for (i = 0; i < pktcnt; i++) {
		p = packets[i];
		/* Stash the softc here so ptom can use it. */
		p->np_softc = sc;
		if (ndis_ptom(&m0, p)) {
			NDISINF((CE_CONT,"ptom failed\n"));
			if (p->np_oob.npo_status == NDIS_STATUS_SUCCESS)
				ndis_return_packet(sc, p);
		} else {
				if (p->np_oob.npo_status == NDIS_STATUS_RESOURCES) 
					cmn_err(CE_CONT,"RXOF resource err\n");
				else
					p->np_oob.npo_status = NDIS_STATUS_PENDING;
				bcopy(&sc->ndis_wd, &wd, sizeof(wifi_data_t));
				m = mac_wifi_header_fake(m0, (void *)&wd);
				if(m == NULL){
					cmn_err(CE_CONT,"Null package!!!\n");
				}else{	
					sc->ic.ic_stats.is_rx_bytes += MBLKL(m);
					sc->ic.ic_stats.is_rx_frags += 1;					
					mac_rx(sc->ic.ic_mach, NULL, m);
					NDISDBG((CE_CONT, "mac_rx done! \n"));
				}
		}
	}
	RESTORE();
	return;
}

/*
 * A frame was downloaded to the chip. It's safe for us to clean up
 * the list buffers.
 */
MSENTRY3(void ndis_txeof, adapter, packet, status)
	ndis_miniport_block	*adapter;
	ndis_packet		*packet;
	ndis_status		status;

{
	struct ndis_softc	*sc;
	ndis_miniport_block	*block;
	int			idx;
	mblk_t		*m;

	//NDISDBG((CE_CONT, "ndis_txeof\n"));
	block = (ndis_miniport_block *)adapter;
	sc = (struct ndis_softc *)block->nmb_ifp;

	NDIS_LOCK(sc);
	m = packet->np_m0;
	idx = packet->np_txidx;
	//ndis_free_packet(packet);
	ndis_free_bufs(packet->np_private.npp_head);
	packet->np_private.npp_head = NULL;
	//NdisFreePacket(packet);
	freemsg(m);
	packet->np_m0 = NULL;

	//sc->ndis_txarray[idx] = NULL;
	sc->ndis_txpending++;
	NDIS_UNLOCK(sc);
	if(sc->ndis_txpending == 1)
		mac_tx_update(sc->ic.ic_mach);
	RESTORE();
	return;
}

MSENTRY4(void ndis_linksts, adapter, status, sbuf, slen)
	ndis_miniport_block	*adapter;
	ndis_status		status;
	void			*sbuf;
	uint32_t		slen;
{
	ndis_miniport_block	*block;
	struct ndis_softc	*sc;

	block = adapter;
	sc = (struct ndis_softc *)block->nmb_ifp;

	block = adapter;
	block->nmb_getstat = status;

	RESTORE();
	return;
}

MSENTRY1(void ndis_linksts_done, adapter)
	ndis_miniport_block	*adapter;
{
	ndis_miniport_block	*block;
	struct ndis_softc	*sc;

	block = adapter;
	sc = (struct ndis_softc *)block->nmb_ifp;

	NDISDBG((CE_CONT, "ndis_linksts_done\n"));

	if (!NDIS_INITIALIZED(sc)) {
		RESTORE();
		return;
	}

	switch (block->nmb_getstat) {
	case NDIS_STATUS_MEDIA_CONNECT:
		ndis_sched(ndis_ticktask, sc, NDIS_TASKTICK);
		NDISDBG((CE_CONT, "media connect\n"));
		break;
	case NDIS_STATUS_MEDIA_DISCONNECT:
		if (sc->ndis_link)
			ndis_sched(ndis_ticktask, sc, NDIS_TASKTICK);
		NDISDBG((CE_CONT, "media disconnect\n"));
		break;
	default:
		break;
	}

	RESTORE();
	return;
}

static uint32_t
ndis_intr(caddr_t arg)
{
	struct ndis_softc	*sc;
	int			is_our_intr = 0;
	int			call_isr = 0;
	ndis_miniport_interrupt	*intr;
/*
 * The handler is called with INTR_LOCK(sc) already held
 */	
	sc = ( struct ndis_softc *)arg;
	intr = sc->ndis_block->nmb_interrupt;
	
	if (sc->ndis_block->nmb_miniportadapterctx == NULL
		|| intr == NULL || (sc->id.id_state != ID_ATTACHED)) {
		return(DDI_INTR_UNCLAIMED);
	}
	
	if (sc->ndis_block->nmb_interrupt->ni_isrreq == TRUE) {
		ndis_isr(sc, &is_our_intr, &call_isr);
	} else {
		NDISINF((CE_CONT,"SO BAD\n"));
		ndis_disable_intr(sc);
		call_isr = 1;
	}
	if ((is_our_intr || call_isr)) {
	//	ndis_intrhandler(sc);
		MSCALL3(KeInsertQueueDpc,
			&sc->ndis_block->nmb_deviceobj->do_dpc, NULL, sc);
		return(DDI_INTR_CLAIMED);
	}
	return(DDI_INTR_UNCLAIMED);
}

static void
ndis_tick(xsc)
	void			*xsc;
{
	struct ndis_softc	*sc;
	sc = (struct ndis_softc	*)xsc;
	if (!NDIS_INITIALIZED(sc) || sc->ndis_tickstop)
		return;
	ndis_sched(ndis_ticktask, sc, NDIS_TASKTICK);
	sc->ndis_stat_ch = timeout(ndis_tick, sc,
	    drv_usectohz(1000000) * sc->ndis_block->nmb_checkforhangsecs);
	return;
}


static void
ndis_ticktask(xsc)
	void			*xsc;
{
	struct ndis_softc	*sc;
	ndis_media_state	linkstate;
	int			arg,error, len,i;
	wifi_data_t 	wd;
	
	sc = xsc;

	if (!NDIS_INITIALIZED(sc))
		return;

	if(sc->ndis_intick || (sc->id.id_state == ID_QUIESCED))
		return;

	sc->ndis_intick = 0x1;
	len = sizeof(linkstate);
	error = ndis_get_info(sc, OID_GEN_MEDIA_CONNECT_STATUS,
	    (void *)&linkstate, &len);
	if (error == EAGAIN) {
		sc->ndis_intick = 0;
		return;
	}

	NDIS_LOCK(sc);
	if (sc->ndis_link == 0 && linkstate == nmc_connected) {
		cmn_err(CE_CONT, "link up\n");
		sc->ndis_link = 1;
		sc->ic.ic_state = IEEE80211_S_RUN;
		if (sc->ndis_80211){
			ndis_getstate_80211(sc);		
			mac_link_update(sc->ic.ic_mach, LINK_STATE_UP);
			bcopy(&sc->ndis_wd, &wd, sizeof(wd));
			(void) mac_pdata_update(sc->ic.ic_mach, &wd, sizeof (wd));
		}
	} else if (sc->ndis_link == 1 && linkstate == nmc_disconnected) {
		cmn_err(CE_CONT, "link down\n");
		sc->ndis_link = 0;
		sc->ic.ic_state = IEEE80211_S_INIT;
		mac_link_update(sc->ic.ic_mach, LINK_STATE_DOWN);
	}

	if(sc->ndis_link == 0){
		sc->ndis_rssi = 0;
		/* Get TX power */
		if ((sc->ic.ic_caps & IEEE80211_C_TXPMGT) && (sc->ic.ic_txpowlimit == 0)) {
			len = sizeof(arg);
			ndis_get_info(sc, OID_802_11_TX_POWER_LEVEL, &arg, &len);
			for (i = 0; i < (sizeof(dBm2mW) / sizeof(dBm2mW[0])); i++)
				if (dBm2mW[i] >= arg)
					break;
			sc->ic.ic_txpowlimit = i;
			/* Set TX power */
			if (sc->ic.ic_txpowlimit < (sizeof(dBm2mW) / sizeof(dBm2mW[0]))) {
				arg = dBm2mW[sc->ic.ic_txpowlimit];
				len = sizeof(arg);
				ndis_set_info(sc, OID_802_11_TX_POWER_LEVEL, &arg, &len);
			}
		}
	}

	sc->ndis_intick = 0x0;
	NDIS_UNLOCK(sc);
	return;
}

static int
ndis_get_assoc(	struct ndis_softc *sc, ndis_wlan_bssid_ex **assoc)
{
	ndis_80211_bssid_list_ex	*bl;
	ndis_wlan_bssid_ex	*bs;
	ndis_80211_macaddr	bssid;
	int len, error, rssi;
	uint32_t i;

	NDISDBG((CE_CONT,"%s\n",__func__));
	if (!sc->ndis_link)
		return(ENOENT);

	len = sizeof(bssid);
	error = ndis_get_info(sc, OID_802_11_BSSID, (void *)&bssid, &len);
	if (error) {
		NDISDBG((CE_CONT,"assoc failed to get bssid\n"));
		return(ENOENT);
	}
	len = 8192;
	bl = kmem_zalloc(len, KM_NOSLEEP);
	error = ndis_get_info(sc, OID_802_11_BSSID_LIST, (void *)bl, &len);
	if (error) {
		kmem_free(bl, 8192);
		NDISDBG((CE_CONT,"bssid_list failed\n"));
		return (error);
	}

	bs = (ndis_wlan_bssid_ex *)&bl->nblx_bssid[0];
	for (i = 0; i < bl->nblx_items; i++) {
		if (bcmp(bs->nwbx_macaddr, bssid, sizeof(bssid)) == 0) {
			*assoc = kmem_zalloc(bs->nwbx_len, KM_NOSLEEP);
			if (*assoc == NULL) {
				kmem_free(bl, 8192);
				return(ENOMEM);
			}
			bcopy((char *)bs, (char *)*assoc, bs->nwbx_len);
			bcopy((char *)bs->nwbx_macaddr, (char *)sc->ndis_wd.wd_bssid,
				IEEE80211_ADDR_LEN);
			rssi = (bs->nwbx_rssi + 200)/12;
			if(rssi < 0)
				rssi = 0;
			else if(rssi > 15)
				rssi = 15;
			sc->ndis_rssi = rssi;
			sc->ndis_wd.wd_opmode = sc->ic.ic_opmode;
			if(sc->ic.ic_flags & IEEE80211_F_PRIVACY)
				sc->ndis_wd.wd_secalloc = WIFI_SEC_WEP;
			else
				sc->ndis_wd.wd_secalloc = WIFI_SEC_NONE;

			kmem_free(bl, 8192);
			return(0);
		}	
		bs = (ndis_wlan_bssid_ex *)((char *)bs + bs->nwbx_len);
	}

	kmem_free(bl, 8192);
	NDISINF((CE_CONT,"Associate Failed\n"));	
	return(ENOENT);
}

static void
ndis_getstate_80211(struct ndis_softc	*sc)
{
	struct ieee80211com	*ic;
	ndis_80211_ssid		ssid;
	ndis_80211_config	config;
	ndis_wlan_bssid_ex	*bs;
	int32_t			rval, len, i = 0;
	uint32_t		arg;

	ic = &sc->ic;

	NDISDBG((CE_CONT,"ndis_getstate_80211\n"));
	if (!NDIS_INITIALIZED(sc))
		return;

	/*
	 * If we're associated, retrieve info on the current bssid.
	 */
	if ((rval = ndis_get_assoc(sc, &bs)) == 0) {
		sc->nettype = bs->nwbx_nettype;
		switch(bs->nwbx_nettype) {
		case NDIS_80211_NETTYPE_11FH:
		case NDIS_80211_NETTYPE_11DS:
			ic->ic_curmode = IEEE80211_MODE_11B;
			break;
		case NDIS_80211_NETTYPE_11OFDM5:
			ic->ic_curmode = IEEE80211_MODE_11A;
			break;
		case NDIS_80211_NETTYPE_11OFDM24:
			ic->ic_curmode = IEEE80211_MODE_11G;
			break;
		default:
			NDISDBG((CE_CONT, "unknown nettype %d\n", arg));
			break;
		}
		kmem_free(bs,bs->nwbx_len);
	} else
		return;

	len = sizeof(ssid);
	bzero((char *)&ssid, len);
	rval = ndis_get_info(sc, OID_802_11_SSID, &ssid, &len);

	if (rval)
		NDISDBG((CE_CONT,"get ssid failed: %d\n", rval));
	else{
		bcopy(ssid.ns_ssid, ic->ic_bss->in_essid, ssid.ns_ssidlen);
		ic->ic_bss->in_esslen = ssid.ns_ssidlen;
	}

	len = sizeof(arg);
	rval = ndis_get_info(sc, OID_GEN_LINK_SPEED, &arg, &len);
	if (rval)
		NDISDBG((CE_CONT,"get link speed failed: %d\n", rval));
	NDISDBG((CE_CONT, "LinkSpeed = %dMbps\n", arg/10000));	

	if (ic->ic_modecaps & (1<<IEEE80211_MODE_11B)) {
		ic->ic_bss->in_rates = ic->ic_sup_rates[IEEE80211_MODE_11B];
		for (i = 0; i < ic->ic_bss->in_rates.ir_nrates; i++) {
			if ((uint32_t)(ic->ic_bss->in_rates.ir_rates[i] &
			    IEEE80211_RATE_VAL) == arg / 5000)
				break;
		}
	}

	if (i == ic->ic_bss->in_rates.ir_nrates &&
	    ic->ic_modecaps & (1<<IEEE80211_MODE_11G)) {
		ic->ic_bss->in_rates = ic->ic_sup_rates[IEEE80211_MODE_11G];
		for (i = 0; i < ic->ic_bss->in_rates.ir_nrates; i++) {
			if ((uint32_t)(ic->ic_bss->in_rates.ir_rates[i] &
			    IEEE80211_RATE_VAL) == arg / 5000)
				break;
		}
	}

	if (i == ic->ic_bss->in_rates.ir_nrates)
		NDISDBG((CE_CONT,"no matching rate for: %d\n", arg / 5000));
	else
		ic->ic_bss->in_txrate = i;

	if (ic->ic_caps & IEEE80211_C_PMGT) {
		len = sizeof(arg);
		rval = ndis_get_info(sc, OID_802_11_POWER_MODE, &arg, &len);

		if (rval)
			NDISDBG((CE_CONT, "get power mode failed: %d\n", rval));
		if (arg == NDIS_80211_POWERMODE_CAM)
			ic->ic_flags &= ~IEEE80211_F_PMGTON;
		else
			ic->ic_flags |= IEEE80211_F_PMGTON;
	}

	len = sizeof(config);
	bzero((char *)&config, len);
	config.nc_length = len;
	config.nc_fhconfig.ncf_length = sizeof(ndis_80211_config_fh);
	rval = ndis_get_info(sc, OID_802_11_CONFIGURATION, &config, &len);   
	if (rval == 0) { 
		int chan;

		chan = ieee80211_mhz2ieee(config.nc_dsconfig / 1000, 0);
		if (chan < 0 || chan >= IEEE80211_CHAN_MAX) {
				NDISDBG((CE_CONT,"current channel "
				    "(%uMHz) out of bounds\n", 
				    config.nc_dsconfig / 1000));
			ic->ic_bss->in_chan = &ic->ic_sup_channels[1];
			sc->phychan = 1; // by default.
		} else{
			ic->ic_bss->in_chan = &ic->ic_sup_channels[chan];
			sc->phychan = chan;
			NDISDBG((CE_CONT,"%s: connected channel is [%d]\n",__func__, chan));	
		}
	} else
		NDISDBG((CE_CONT,"couldn't retrieve "
		    "channel info: %d\n", rval));

	return;
}


