/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*-
 * Copyright (c) 2003
 *	Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD: /repoman/r/ncvs/src/sys/dev/if_ndis/if_ndisvar.h,v 1.19 2005/02/16 05:41:17 wpaul Exp $
 */
#ifndef _IF_NDISVAR_H_
#define _IF_NDISVAR_H_

#include <sys/net80211.h>
#include <sys/mac_provider.h>
#include <sys/mac_wifi.h>

#define NDIS_DEFAULT_NODENAME	"FreeBSD NDIS node"
#define NDIS_NODENAME_LEN	32
#define ETHER_ADDR_LEN 6

#define NDISINF(x) \
        if (ndis_debug & NDISINFMSG ) cmn_err x

#define NDISDBG(x) \
        if (ndis_debug & NDISDBGMSG ) cmn_err x

#define NWAMDBG(x) \
        if (ndis_debug & NDSINWAMSG ) cmn_err x

#define NDIS11NDBG(x) \
	if (ndis_debug & NDIS80211N ) cmn_err x

#define NDISIOCTL(x) \
	if (ndis_debug & NDISDIOCTL ) cmn_err x

#define NDISRATES(x) \
	if (ndis_debug & NDISDRATES ) cmn_err x

#define NDISCFGPHY(x) \
	if (ndis_debug & NDISDCFGPHY ) cmn_err x

#define NDISPROP(x) \
	if (ndis_debug & NDISBRUSSELS) cmn_err x

struct ndis_pci_type {
	uint16_t		ndis_vid;
	uint16_t		ndis_did;
	uint32_t		ndis_subsys;
	char			*ndis_name;
};

struct ndis_pccard_type {
	const char		*ndis_vid;
	const char		*ndis_did;
	char			*ndis_name;
};

typedef struct {
	ddi_acc_handle_t	acc_hdl;	/* handle for memory */
	caddr_t			mem_va;		/* CPU VA of memory */
	uint32_t		nslots;		/* number of slots */
	uint32_t		size;		/* size per slot */
	size_t			alength;	/* allocated size */
						/* >= product of above */

	ddi_dma_handle_t	dma_hdl;	/* DMA handle */
	offset_t		offset;		/* relative to handle */
	ddi_dma_cookie_t	cookie;		/* associated cookie */
	uint32_t		ncookies;	/* must be 1 */
	uint32_t		token;		/* arbitrary identifier */
} my_dma_area_t;					/* 0x50 (80) bytes */

struct ndis_shmem {
	ddi_dma_attr_t	*accattr;
	my_dma_area_t	bf_dma;
	void			*ndis_saddr;
	uint32_t 		shm_len;
	physaddr		phyaddr;
	struct ndis_shmem	*ndis_next;
};

struct ndis_cfglist {
	ndis_cfg		ndis_cfg;
        TAILQ_ENTRY(ndis_cfglist)	link;
};

TAILQ_HEAD(nch, ndis_cfglist);

struct ntmemlist {
	uintptr_t		*addr;
	uint32_t		len;
	uint32_t		tag;
    TAILQ_ENTRY(ntmemlist)	link;
};

TAILQ_HEAD(nml, ntmemlist);


#define NDIS_INITIALIZED(sc)	(sc->ndis_block->nmb_miniportadapterctx != NULL)

#define NDIS_INC(x)		\
	(x)->ndis_txidx = ((x)->ndis_txidx + 1) % (x)->ndis_maxpkts 
#define arpcom ic.ic_ac

typedef enum ieee80211_idstate {
    ID_DETACHED=0,
    ID_QUIESCED=1,
    ID_SUSPENDED=2,
    ID_ATTACHED=3
} ieee80211_idstate_t;

typedef struct ndis_softc softc_t;
typedef struct ieee80211ddi {
	dev_info_t		*id_dev;
	ddi_acc_handle_t	id_cfg_hdl;	/* DDI I/O handle */
	ddi_acc_handle_t	id_io_hdl;	/* DDI I/O handle */
	ddi_softint_handle_t	id_softint;
	ddi_iblock_cookie_t	id_iblock;
	kmutex_t		id_topmtx;
	kmutex_t		id_intrmtx;
	ieee80211_idstate_t	id_state;

	uint8_t			*id_mcast;
	uint32_t		id_mcast_count;
	uint32_t		id_mcast_len;
/*
 * callbacks
 */
	void			(*id_intr)(softc_t *);
	int			(*id_attach)(softc_t *);
	int			(*id_detach)(softc_t *);
	int			(*id_get)(softc_t *, int, void **);
	int			(*id_set)(softc_t *, int, void *);
	int			(*id_cmd)(softc_t *, int);
	int			(*id_xmt)(softc_t *, mblk_t *, uint8_t);
} ieee80211ddi_t;

#define IEEE80211_MAX_WPA_IE		40	/* IEEE802.11i */
#define GENERIC_INFO_ELEM	0xdd
#define RSN_INFO_ELEM	0x30

int
ndis_setprop(void *ic_arg, const char *pr_name, mac_prop_id_t wldp_pr_num,
    uint_t wldp_length, const void *wldp_buf);
int
ndis_getprop(void *ic_arg, const char *pr_name, mac_prop_id_t wldp_pr_num,
    uint_t wldp_length, void *wldp_buf);

struct ndis_softc {
	ieee80211com_t		ic;		/* interface info */
	ieee80211ddi_t		id;

	void			*ioport_ndis;
	ddi_acc_handle_t	ndis_PCI_bar1_handle;
	void			*ioreg_ndis[6];
	ddi_acc_handle_t	ndis_PCI_bar2_handle[6];
	vm_offset_t		ndis_drv_data;

	wifi_data_t		ndis_wd;
	u_long			ndis_hwassist;
	uint32_t		ndis_v4tx;
	uint32_t		ndis_v4rx;
	int			ndis_rescnt;
	int			ndis_unit;
	ndis_miniport_block	*ndis_block;
	ndis_miniport_characteristics	*ndis_chars;
	interface_type		ndis_type;
	int			ndis_maxpkts;
	ndis_oid		*ndis_oids;
	int			ndis_oidcnt;
	int			ndis_txidx;
	int			ndis_txpending;
	ndis_packet		**ndis_txarray;
	ndis_handle		ndis_txpool;
	int			ndis_sc;
	ndis_cfg		*ndis_regvals;
	struct nch		ndis_cfglist_head;
	int			ndis_80211;
	int			ndis_link;
	uint32_t		ndis_filter;
	int			ndis_if_flags;
	int			ndis_skip;
	int			ndis_tickstop;
	int			ndis_intick;

	int			ndis_devidx;
	interface_type		ndis_iftype;

	struct ndis_shmem	*ndis_shlist;

	int			ndis_mmapcnt;
	timeout_id_t		ndis_stat_ch;
	int32_t			ndis_promisc;
	int			ndis_skip_gstat;
	int			ndis_skip_start;
	uint32_t		ndis_rssi;
	uint32_t		phychan;
	uint32_t		nettype;
	ndis_wlan_bssid_ex	ndis_assoc;
};
#define NDIS_LOCK(_sc)		mutex_enter(&(_sc)->id.id_topmtx)
#define NDIS_UNLOCK(_sc)	mutex_exit(&(_sc)->id.id_topmtx)
#define INTR_LOCK(sc)		mutex_enter(&(sc)->id.id_intrmtx)
#define INTR_UNLOCK(sc)		mutex_exit(&(sc)->id.id_intrmtx)

#ifndef GLD_SUCCESS
#define NDIS_SUCCESS 0x0
#define NDIS_FAILURE 0X1
#define GLD_SUCCESS NDIS_SUCCESS
#define GLD_FAILURE NDIS_FAILURE
#else
#define NDIS_SUCCESS 0x0
#define NDIS_FAILURE 0X1
#endif

#endif
