/*
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 * Portions Copyright 2013 Jean-Pierre Andre
 */

/*-
 * Copyright (c) 2003
 *	Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/unistd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/systm.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/proc.h>

#include <sys/asm_linkage.h>
#include <sys/panic.h>
#include <sys/regset.h>
#include <sys/reboot.h>
#include <sys/psw.h>

#include <sys/callo.h>
#include <sys/sunddi.h>
#include <sys/ddi.h>
#include <sys/atomic.h>
#include <sys/condvar.h>
#include <sys/thread.h>
#include <sys/cmn_err.h>

#include "queue.h"
#include "pe_var.h"
#include "resource_var.h"
#include "ntoskrnl_var.h"
#include "ndis_var.h"
#include "hal_var.h"
#include "cfg_var.h"
#include "if_ndisvar.h"

#define __regparm __attribute__((regparm(3)))

typedef unsigned long ACCESS_MASK;
typedef uint64_t vm_paddr_t;

MSTYPE3(uint8_t RtlEqualUnicodeString, ndis_unicode_string *,
	ndis_unicode_string *, uint8_t);
MSTYPE2(void RtlCopyUnicodeString, ndis_unicode_string *,
	ndis_unicode_string *);
MSTYPE3(ndis_status RtlUnicodeStringToAnsiString, ndis_ansi_string *,
	ndis_unicode_string *, uint8_t);
MSTYPE3(ndis_status RtlAnsiStringToUnicodeString, ndis_unicode_string *,
	ndis_ansi_string *, uint8_t);
MSTYPE5(irp *IoBuildSynchronousFsdRequest, uint32_t, device_object *,
	 void *, uint32_t, uint64_t *, nt_kevent *, io_status_block *);
MSTYPE5(irp *IoBuildAsynchronousFsdRequest, uint32_t,
	device_object *, void *, uint32_t, uint64_t *, io_status_block *);
MSTYPE5(irp *IoBuildDeviceIoControlRequest, uint32_t,
	device_object *, void *, uint32_t, void *, uint32_t,
	uint8_t, nt_kevent *, io_status_block *);
MSTYPE2(irp *IoAllocateIrp, uint8_t, uint8_t);
MSTYPE2(void IoReuseIrp, irp *, uint32_t);
MSTYPE1(void IoFreeIrp, irp *);
MSTYPE3(void IoInitializeIrp, irp *, uint16_t, uint8_t);
MSTYPE2(irp *IoMakeAssociatedIrp, irp *, uint8_t);
MSTYPE5(uint32_t KeWaitForMultipleObjects, uint32_t,
	nt_dispatch_header **, uint32_t, uint32_t, uint32_t, uint8_t,
	int64_t *, wait_block *);
static void ntoskrnl_wakeup(void *);
static void ntoskrnl_timercall(void *);
static void ntoskrnl_run_dpc(void *);
MSTYPE2(void WRITE_REGISTER_USHORT, uint16_t *, uint16_t);
MSTYPE1(uint16_t READ_REGISTER_USHORT, uint16_t *);
MSTYPE2(void WRITE_REGISTER_ULONG, uint32_t *, uint32_t);
MSTYPE1(uint32_t READ_REGISTER_ULONG, uint32_t *);
MSTYPE2(void WRITE_REGISTER_UCHAR, uint8_t *, uint8_t);
MSTYPE1(uint8_t READ_REGISTER_UCHAR, uint8_t *);
MSTYPE2(int64_t _allmul, int64_t, int64_t);
MSTYPE2(int64_t _alldiv, int64_t, int64_t);
MSTYPE2(int64_t _allrem, int64_t, int64_t);
__regparm static int64_t _allshr(int64_t, uint8_t);
__regparm static int64_t _allshl(int64_t, uint8_t);
MSTYPE2(uint64_t _aullmul, uint64_t, uint64_t);
MSTYPE2(uint64_t _aulldiv, uint64_t, uint64_t);
MSTYPE2(uint64_t _aullrem, uint64_t, uint64_t);
__regparm static uint64_t _aullshr(uint64_t, uint8_t);
__regparm static uint64_t _aullshl(uint64_t, uint8_t);
static slist_entry *ntoskrnl_pushsl(slist_header *, slist_entry *);
static slist_entry *ntoskrnl_popsl(slist_header *);
MSTYPE5(void ExInitializePagedLookasideList, paged_lookaside_list *,
	lookaside_alloc_func, lookaside_free_func,
	uint32_t, size_t, uint32_t, uint16_t);
MSTYPE1(void ExDeletePagedLookasideList, paged_lookaside_list *);
MSTYPE5(void ExInitializeNPagedLookasideList, npaged_lookaside_list *,
	lookaside_alloc_func, lookaside_free_func,
	uint32_t, size_t, uint32_t, uint16_t);
MSTYPE1(void ExDeleteNPagedLookasideList, npaged_lookaside_list *);
__fastcall static slist_entry
	*InterlockedPushEntrySList(REGARGS2(slist_header *head,
	slist_entry *entry));
__fastcall static slist_entry *InterlockedPopEntrySList(REGARGS1(slist_header
	*head));
__fastcall static slist_entry
	*ExInterlockedPushEntrySList(REGARGS2(slist_header *head,
	slist_entry *entry), kspin_lock *lock);
__fastcall static slist_entry
	*ExInterlockedPopEntrySList(REGARGS2(slist_header *head,
	kspin_lock *lock));
MSTYPE1(uint16_t ExQueryDepthSList, slist_header *);
__fastcall static uint32_t
	InterlockedIncrement(REGARGS1(volatile uint32_t *addend));
__fastcall static uint32_t
	InterlockedDecrement(REGARGS1(volatile uint32_t *addend));
__fastcall static void
	ExInterlockedAddLargeStatistic(REGARGS2(uint64_t *addend, uint32_t));
MSTYPE2(uint32_t MmSizeOfMdl, void *, size_t);
MSTYPE1(void MmBuildMdlForNonPagedPool, mdl *);
MSTYPE2(void *MmMapLockedPages, mdl *, uint8_t);
MSTYPE5(void *MmMapLockedPagesSpecifyCache, mdl *,
	uint8_t, uint32_t, void *, uint32_t, uint32_t);
MSTYPE2(void MmUnmapLockedPages, void *, mdl *);
MSTYPE3(size_t RtlCompareMemory, const void *, const void *, size_t);
MSTYPE2(void RtlInitAnsiString, ndis_ansi_string *, char *);
MSTYPE2(void RtlInitString, ndis_ansi_string *, char *);
MSTYPE2(void RtlInitUnicodeString, ndis_unicode_string *,
	uint16_t *);
MSTYPE1(void RtlFreeUnicodeString, ndis_unicode_string *);
MSTYPE1(void RtlFreeAnsiString, ndis_ansi_string *);
MSTYPE3(ndis_status RtlUnicodeStringToInteger, ndis_unicode_string *,
	uint32_t, uint32_t *);
static int atoi (const char *);
static long atol (const char *);
static int rand(void);
static void srand(unsigned int);
static void ntoskrnl_time(uint64_t *);
MSTYPE2(uint8_t IoIsWdmVersionAvailable, uint8_t, uint8_t);
static void ntoskrnl_thrfunc(void *);
MSTYPE5(ndis_status PsCreateSystemThread, ndis_handle *,
	uint32_t, void *, ndis_handle, void *, void *, void *);
MSTYPE1(ndis_status PsTerminateSystemThread, ndis_status);
MSTYPE5(ndis_status IoGetDeviceProperty, device_object *, uint32_t,
	uint32_t, void *, uint32_t *);
MSTYPE2(void KeInitializeMutex, kmutant *, uint32_t);
MSTYPE2(uint32_t KeReleaseMutex, kmutant *, uint8_t);
MSTYPE1(uint32_t KeReadStateMutex, kmutant *);
MSTYPE5(ndis_status ObReferenceObjectByHandle, ndis_handle,
	uint32_t, void *, uint8_t, void **, void **);
#if defined(__amd64__) & defined(LINKEDIMAGE)
MSTYPE1(void ObfDereferenceObject, void*);
#else
__fastcall static void ObfDereferenceObject(REGARGS1(void *object));
#endif
MSTYPE1(uint32_t ZwClose, ndis_handle);
static void *ntoskrnl_memset(void *, int, size_t);
static funcptr ntoskrnl_findwrap(funcptr);
#if defined(__amd64__) & defined(LINKEDIMAGE)
MSTYPE5(uint32_t DbgPrint, char*, uint64_t, uint64_t, uint64_t, uint64_t, uint64_t);
#else
static uint32_t DbgPrint(char *, ...);
#endif
MSTYPE0(void DbgBreakPoint);
MSTYPE0(void dummy_ntoskrnl);
MSTYPE1(vm_paddr_t MmGetPhysicalAddress, vm_offset_t);
MSTYPE1(uint8_t KeAcquireSpinLockRaiseToDpc, kspin_lock*);
MSTYPE3(void MmFreeContiguousMemorySpecifyCache, void*, uint32_t, uint32_t);
MSTYPE5(void *MmAllocateContiguousMemorySpecifyCache, uint32_t, uint64_t, uint64_t, uint64_t, uint32_t);
MSTYPE5(void KeBugCheckEx, uint32_t, u_long, u_long, u_long, u_long);
MSTYPE4(uint32_t IoGetDeviceObjectPointer, struct unicode_string*,
			ACCESS_MASK, void**, struct device_object**);

static void NtoskernelPageCache(uint32_t, void**);
static void *myExAllocatePoolWithTag(uint32_t, size_t, uint32_t);

kmutex_t ntoskrnl_dispatchlock,sched_lock, register_mtx;
kmutex_t Txbuflock, Iobuflock;
extern kmutex_t ndis_poollist_mtx;
static kspin_lock ntoskrnl_global;
static kspin_lock ntoskrnl_cancellock;
static int ntoskrnl_kth = 0;
static struct nt_objref_head ntoskrnl_reflist;
kcondvar_t ntoskrnl_cv;
extern struct ndis_softc	*curSC;
#define LOCK_REG(x)	mutex_enter(x)
#define UNLOCK_REG(x)	mutex_exit(x)

#if defined(__amd64__) & defined(LINKEDIMAGE)
MSTYPE3(void *win_memcpy, void*, const void *, size_t);
MSTYPE3(int win_strncmp, const char*, const char*, size_t);
MSTYPE3(char *win_strncpy, char *, const char *, size_t);
MSTYPE2(char *win_strchr, const char*, char);
MSTYPE2(char *win_strrchr, const char*, char);
#endif

#if defined(__amd64__) & defined(LINKEDIMAGE)

/*
 *	va_list in Window is much different from the one in Unix,
 *	so we assume no more than 8 arguments are needed.
 *	Do not define a prototype to avoid type checking.
 */

MSENTRY5(int win_sprintf, buf, fmt, a1, a2, a3, a4, a5, a6)
char *buf;
const char *fmt;
void *a1, *a2, *a3, *a4, *a5, *a6;
{
	int sz;

// TODO translate the format (%I64d -> %lld)
	sz = snprintf(buf, INT_MAX, fmt, a1, a2, a3, a4, a5, a6);
	RESTORE();
	return (sz);
}
 
#else /* defined(__amd64__) & defined(LINKEDIMAGE) */

int
win_sprintf(char *buf, const char *fmt, ...)
{
	int retravl;
	va_list args;

	__FUNCNAME__;
	va_start(args, fmt);
	retravl = vsnprintf(buf, INT_MAX, fmt, args);
	va_end(args);

	return (retravl);
}

#endif /* defined(__amd64__) & defined(LINKEDIMAGE) */

MSENTRY3(void *win_memcpy, s1, s2, n)
void *s1;
const void *s2;
size_t n;
{
	__FUNCNAME__;
	bcopy(s2, s1, n);
	RESTORE();
	return (s1);
}

MSENTRY2(void win_bzero, s1, n)
	void *s1;
	size_t n;
{
	__FUNCNAME__;
	bzero(s1, n);
	RESTORE();
}

#if defined(__amd64__) & defined(LINKEDIMAGE)

MSENTRY3(int win_strncmp, s1, s2, cnt)
	const char *s1;
	const char *s2;
	size_t cnt;
{
	int cmp;

	cmp = strncmp(s1, s2, cnt);
	RESTORE();
	return (cmp);
}

MSENTRY3(char *win_strncpy, d, s, cnt)
	char *d;
	const char *s;
	size_t cnt;
{
	char *p;

	p = strncpy(d, s, cnt);
	RESTORE();
	return (p);
}

MSENTRY2(char *win_strrchr, s, c)
	const char *s;
	char c;
{
	char *p;

	p = strrchr(s, c);
	RESTORE();
	return (p);
}

MSENTRY2(char *win_strchr, s, c)
	const char *s;
	char c;
{
	char *p;

	p = strchr(s, c);
	RESTORE();
	return (p);
}

#endif

#define MDL_ZONE_PAGES 16
#define MDL_ZONE_SIZE (sizeof(mdl) + (sizeof(vm_offset_t) * MDL_ZONE_PAGES))
#define MDL_ZONE_ALLOCED		0x8000	/* BSD private */

static kmem_cache_t *mdl_cache,*tx_mdl_cache;

struct nml nt_mem_list;
struct nml nt_mem_list_used;
struct nml tx_mem_list;
struct nml tx_mem_list_used;

#define NT_BUF_SIZE (1024*3)
#define BUF_LIST_SIZE 256

int
ntoskrnl_libinit()
{
	image_patch_table	*patch;
  	int i=0;
	struct ntmemlist 	*e;

	__FUNCNAME__;
		    
	mutex_init(&ntoskrnl_dispatchlock, "ntoskrnl dispatch lock", 
		MUTEX_DEFAULT, NULL);
	mutex_init(&sched_lock, "ntos lock mutex",
	   MUTEX_DEFAULT, NULL);
        mutex_init(&register_mtx, "register access lock mutex",
           MUTEX_DEFAULT, NULL);
        mutex_init(&Txbuflock, "tx buffer lock mutex",
           MUTEX_DEFAULT, NULL);
        mutex_init(&Iobuflock, "tx buffer lock mutex",
           MUTEX_DEFAULT, NULL);		
	cv_init(&ntoskrnl_cv, "ntos cond lock", CV_DRIVER, NULL);
	
	KiInitializeSpinLock(&ntoskrnl_global);
	KiInitializeSpinLock(&ntoskrnl_cancellock);
	TAILQ_INIT(&ntoskrnl_reflist);

	patch = ntoskrnl_functbl;
	while (patch->ipt_func != NULL) {
		windrv_wrap((funcptr)patch->ipt_func,
		    (funcptr *)&patch->ipt_wrap);
		patch++;
	}
	
	mdl_cache = kmem_cache_create("ndis_mdl", MDL_ZONE_SIZE*1024, 0, NULL,NULL,NULL,
				      NULL, NULL,0);
	tx_mdl_cache = kmem_cache_create("ndis_mdl_tx", MDL_ZONE_SIZE*1024, 0, NULL,NULL,NULL,
				      NULL, NULL,0);
	
	TAILQ_INIT(&nt_mem_list);
	TAILQ_INIT(&nt_mem_list_used);
	TAILQ_INIT(&tx_mem_list);
	TAILQ_INIT(&tx_mem_list_used);
	for(i = 0; i< BUF_LIST_SIZE; i++){
		e = (struct ntmemlist *)kmem_zalloc(sizeof(struct ntmemlist),KM_SLEEP);
		e->addr = (uintptr_t *)kmem_zalloc(NT_BUF_SIZE,KM_SLEEP);
		e->len = NT_BUF_SIZE;
		if(!e || !e->addr)
			panic("should not be\n");
		TAILQ_INSERT_TAIL(&nt_mem_list, e, link);
	}
	for(i = 0; i< BUF_LIST_SIZE; i++){
		e = (struct ntmemlist *)kmem_zalloc(sizeof(struct ntmemlist),KM_SLEEP);
		e->addr = (uintptr_t *)kmem_zalloc(NT_BUF_SIZE,KM_SLEEP);
		e->len = NT_BUF_SIZE;
		if(!e || !e->addr)
			panic("should not be\n");
		TAILQ_INSERT_TAIL(&tx_mem_list, e, link);
	}		
	return(0);
}

int
ntoskrnl_libfini()
{
	image_patch_table	*patch;
	struct ntmemlist *e;
	int len=1;

	__FUNCNAME__;

	patch = ntoskrnl_functbl;
	while (patch->ipt_func != NULL) {
		windrv_unwrap(patch->ipt_wrap);
		patch++;
	}
	if (mdl_cache)
		kmem_cache_destroy(mdl_cache);
	if (tx_mdl_cache)
		kmem_cache_destroy(tx_mdl_cache);

	cv_destroy(&ntoskrnl_cv);
	mutex_destroy(&sched_lock);
	mutex_destroy(&ntoskrnl_dispatchlock);
	mutex_destroy(&register_mtx);
	mutex_destroy(&Txbuflock);
	mutex_destroy(&Iobuflock);
	
	while(!TAILQ_EMPTY(&nt_mem_list)){
		e = TAILQ_FIRST(&nt_mem_list);
		TAILQ_REMOVE(&nt_mem_list,e,link);
		if((uintptr_t *)e->addr){
			kmem_free(e->addr, e->len);
			e->addr = NULL;
		}
		if(e){
			kmem_free(e, sizeof(struct ntmemlist));
			//cmn_err(CE_CONT,"free nt_mem_list %d\n", len);
		}
		len++;
	}
	len = 1;
	while(!TAILQ_EMPTY(&nt_mem_list_used)){
		e = TAILQ_FIRST(&nt_mem_list_used);
		TAILQ_REMOVE(&nt_mem_list_used,e,link);
		if((uintptr_t *)e->addr){
			kmem_free(e->addr, e->len);
			e->addr = NULL;
		}
		if(e){
			kmem_free(e, sizeof(struct ntmemlist));
			//cmn_err(CE_CONT,"free nt_mem_list_used %d\n", len);
		}
		len++;
	}
	len = 1;
	while(!TAILQ_EMPTY(&tx_mem_list)){
		e = TAILQ_FIRST(&tx_mem_list);
		TAILQ_REMOVE(&tx_mem_list,e,link);
		if((uintptr_t *)e->addr){
			kmem_free(e->addr, e->len);
			e->addr = NULL;
		}
		if(e){
			kmem_free(e, sizeof(struct ntmemlist));
			//cmn_err(CE_CONT,"free tx_mem_list %d\n", len);
		}
		len++;
	}
	len = 1;
	while(!TAILQ_EMPTY(&tx_mem_list_used)){
		e = TAILQ_FIRST(&tx_mem_list_used);
		TAILQ_REMOVE(&tx_mem_list_used,e,link);
		if((uintptr_t *)e->addr){
			kmem_free(e->addr, e->len);
			e->addr = NULL;
		}
		if(e){
			kmem_free(e, sizeof(struct ntmemlist));
			//cmn_err(CE_CONT,"free tx_mem_list_used %d\n", len);
		}
		len++;	
	}	
	return(0);
}

/*
 * We need to be able to reference this externally from the wrapper;
 * GCC only generates a local implementation of memset.
 */
static void *
ntoskrnl_memset(buf, ch, size)
	void			*buf;
	int			ch;
	size_t			size;
{
	unsigned char *t;
	__FUNCNAME__;
	if ((unsigned char)ch == '\0')
		bzero(buf, size);
	else {
		for (t = (unsigned char *)buf; size > 0; size--)
			*t++ = (unsigned char)ch;
	}
	return (buf);
}

MSENTRY3(uint8_t RtlEqualUnicodeString, str1, str2, caseinsensitive)
	ndis_unicode_string	*str1;
	ndis_unicode_string	*str2;
	uint8_t			caseinsensitive;
{
	int			i;

	__FUNCNAME__;
	if (str1->us_len != str2->us_len) {
		RESTORE();
		return(FALSE);
	}

	for (i = 0; i < str1->us_len; i++) {
		if (caseinsensitive == TRUE) {
			if (toupper((char)(str1->us_buf[i] & 0xFF)) !=
			    toupper((char)(str2->us_buf[i] & 0xFF))) {
				RESTORE();
				return(FALSE);
			}
		} else {
			if (str1->us_buf[i] != str2->us_buf[i]) {
				RESTORE();
				return(FALSE);
			}
		}
	}

	RESTORE();
	return(TRUE);
}

MSENTRY2(void RtlCopyUnicodeString, dest, src)
	ndis_unicode_string	*dest;
	ndis_unicode_string	*src;
{

	__FUNCNAME__;
	if (dest->us_maxlen >= src->us_len)
		dest->us_len = src->us_len;
	else
		dest->us_len = dest->us_maxlen;
//	win_memcpy(dest->us_buf, src->us_buf, dest->us_len);
	memcpy(dest->us_buf, src->us_buf, dest->us_len);
	RESTORE();
	return;
}

MSENTRY3(ndis_status RtlUnicodeStringToAnsiString, dest, src, allocate)
	ndis_ansi_string	*dest;
	ndis_unicode_string	*src;
	uint8_t			allocate;
{
	char			*astr = NULL;

	__FUNCNAME__;
	if (dest == NULL || src == NULL) {
		RESTORE();
		return(NDIS_STATUS_FAILURE);
	}

	if (allocate == TRUE) {
		if (ndis_unicode_to_ascii(src->us_buf, src->us_len, &astr)) {
			RESTORE();
			return(NDIS_STATUS_FAILURE);
		}
		dest->nas_buf = astr;
		dest->nas_len = dest->nas_maxlen = strlen(astr);
	} else {
		dest->nas_len = src->us_len / 2; /* XXX */
		if (dest->nas_maxlen < dest->nas_len)
			dest->nas_len = dest->nas_maxlen;
		ndis_unicode_to_ascii(src->us_buf, dest->nas_len * 2,
		    &dest->nas_buf);
	}
	RESTORE();
	return (NDIS_STATUS_SUCCESS);
}

MSENTRY3(ndis_status RtlAnsiStringToUnicodeString, dest, src, allocate)
	ndis_unicode_string	*dest;
	ndis_ansi_string	*src;
	uint8_t			allocate;
{
	uint16_t		*ustr = NULL;

	__FUNCNAME__;
	if (dest == NULL || src == NULL) {
		RESTORE();
		return(NDIS_STATUS_FAILURE);
	}

	if (allocate == TRUE) {
		if (ndis_ascii_to_unicode(src->nas_buf, &ustr)) {
			RESTORE();
			return(NDIS_STATUS_FAILURE);
		}
		dest->us_buf = ustr;
		dest->us_len = dest->us_maxlen = strlen(src->nas_buf) * 2;
	} else {
		dest->us_len = src->nas_len * 2; /* XXX */
		if (dest->us_maxlen < dest->us_len)
			dest->us_len = dest->us_maxlen;
		ndis_ascii_to_unicode(src->nas_buf, &dest->us_buf);
	}
	RESTORE();
	return (NDIS_STATUS_SUCCESS);
}

MSENTRY3(void *ExAllocatePoolWithTag, pooltype, len, tag)
	uint32_t		pooltype;
	size_t			len;
	uint32_t		tag;
{
	struct ntmemlist 	*e=NULL;
	void *ret;

	mutex_enter(&Iobuflock);
	if(TAILQ_EMPTY(&nt_mem_list) || (len > NT_BUF_SIZE)){
		//cmn_err(CE_CONT,"%s:len is%d\n", __func__, len);
		e = kmem_zalloc(sizeof(struct ntmemlist),KM_SLEEP);
		e->addr = kmem_zalloc(len, KM_SLEEP);
		e->len = len;
		e->tag = tag;
		/* new home of e is here:*/
		TAILQ_INSERT_HEAD(&nt_mem_list_used, e, link);
		mutex_exit(&Iobuflock);
		ret = (void*)(e->addr);
		RESTORE();
		return (ret);
	}

	e = TAILQ_FIRST(&nt_mem_list);

	if(!e || !e->addr)
		panic("No memory now!\n");

	TAILQ_REMOVE(&nt_mem_list,e,link);
	
	/* new home of e is here:*/
	TAILQ_INSERT_HEAD(&nt_mem_list_used, e, link);
	//memset(e->addr, 0, NT_BUF_SIZE);
	mutex_exit(&Iobuflock);
	ret = (void *)(e->addr);
	RESTORE();
	return (ret);
}

/* local relay to the above (many local calls) */
static void *myExAllocatePoolWithTag(pooltype, len, tag)
	uint32_t		pooltype;
	size_t			len;
	uint32_t		tag;
{
	return ((void*)MSCALL3(ExAllocatePoolWithTag, pooltype, len, tag));
}


// TODO many calls, relay useful
MSENTRY1(void ExFreePool, buf)
	void			*buf;
{
	struct ntmemlist 	*e=NULL;
	int hit=0;
	
	__FUNCNAME__;
	if(!buf){
		cmn_err(CE_CONT,"%s:Free Null pointer?\n",__func__);
		RESTORE();
		return;
	}

	mutex_enter(&Iobuflock);
	TAILQ_FOREACH(e,&nt_mem_list_used,link){
		if((uintptr_t *)e->addr == (uintptr_t *)buf){
			hit = TRUE;
			break;
		}
	}

	if(hit){
		TAILQ_REMOVE(&nt_mem_list_used,e,link);
		if(e->len > NT_BUF_SIZE){
			//cmn_err(CE_CONT,"Free big mem %d\n", e->len);
			kmem_free(e->addr, e->len);
			kmem_free(e, sizeof(struct ntmemlist));
			mutex_exit(&Iobuflock);	
			RESTORE();
			return;
		}
		/* new home of e is here:*/
		TAILQ_INSERT_HEAD(&nt_mem_list, e, link);
	}else{
		cmn_err(CE_CONT,"I don't know this address!\n");
	}
	mutex_exit(&Iobuflock);
	RESTORE();
	return;
}

MSENTRY2(void ExFreePoolWithTag, buf, tag)
	void			*buf;
	uint32_t		tag;
{
	struct ntmemlist 	*e=NULL;
	int hit=0;
	
	__FUNCNAME__;
	if(!buf){
		cmn_err(CE_CONT,"%s:Free Null pointer?\n",__func__);
		RESTORE();
		return;
	}

	mutex_enter(&Iobuflock);
		/* Assume freeing with null tag means according to buf */
	TAILQ_FOREACH(e,&nt_mem_list_used,link){
		if (tag ? e->tag == tag : (void*)e->addr == buf) {
			hit = TRUE;
			break;
		}
	}

	if(hit){
		TAILQ_REMOVE(&nt_mem_list_used,e,link);
		if(e->len > NT_BUF_SIZE){
			//cmn_err(CE_CONT,"Free big mem %d\n", e->len);
			kmem_free(e->addr, e->len);
			kmem_free(e, sizeof(struct ntmemlist));
			mutex_exit(&Iobuflock);	
			RESTORE();
			return;
		}
		/* new home of e is here:*/
		TAILQ_INSERT_HEAD(&nt_mem_list, e, link);
	}else{
		cmn_err(CE_CONT,"I don't know this address!\n");
	}
	mutex_exit(&Iobuflock);
	RESTORE();
	return;
}

void *
TxExAllocatePoolWithTag(pooltype, len, tag)
	uint32_t		pooltype;
	size_t			len;
	uint32_t		tag;
{
	struct ntmemlist 	*e=NULL;
	
	mutex_enter(&Txbuflock);
	if(len > NT_BUF_SIZE || TAILQ_EMPTY(&tx_mem_list) ){
		//cmn_err(CE_WARN,"%s:len is%d\n", __func__, len);
		e = kmem_zalloc(sizeof(struct ntmemlist),KM_SLEEP);
		e->addr = kmem_zalloc(len, KM_SLEEP);
		e->len = len;
		e->tag = tag;
		/* new home of e is here:*/
		TAILQ_INSERT_HEAD(&tx_mem_list_used, e, link);
		mutex_exit(&Txbuflock);
		return((void *)(e->addr));
	}

	e = TAILQ_FIRST(&tx_mem_list);

	if(!e || !e->addr)
		panic("No memory now!\n");

	TAILQ_REMOVE(&tx_mem_list,e,link);
	
	/* new home of e is here:*/
	TAILQ_INSERT_HEAD(&tx_mem_list_used, e, link);
	mutex_exit(&Txbuflock);
	return((void *)(e->addr));
}

void
TxExFreePool(buf)
	void			*buf;
{
	struct ntmemlist 	*e=NULL;
	int hit=0;
	
	__FUNCNAME__;
	if(!buf){
		cmn_err(CE_CONT,"%s:Free Null pointer?\n",__func__);
		return;
	}

	//cmn_err(CE_CONT,"%s:called\n",__func__);
	mutex_enter(&Txbuflock);
	TAILQ_FOREACH(e,&tx_mem_list_used,link){
		if((uintptr_t *)e->addr == (uintptr_t *)buf){
			cmn_err(CE_CONT,"%s:hit\n",__func__);
			hit = TRUE;
			break;
		}
	}

	if(hit){
		TAILQ_REMOVE(&tx_mem_list_used,e,link);
		if(e->len != NT_BUF_SIZE){
			//cmn_err(CE_CONT,"Free random mem %d\n", e->len);
			kmem_free(e->addr, e->len);
			kmem_free(e, sizeof(struct ntmemlist));
			mutex_exit(&Txbuflock);
			return;
		}
		//cmn_err(CE_CONT,"rehome NT mem %d\n", e->len);
		/* new home of e is here:*/
		TAILQ_INSERT_HEAD(&tx_mem_list, e, link);
	}else{
		cmn_err(CE_CONT,"I don't know this address!\n");
	}

	mutex_exit(&Txbuflock);
	return;
}

MSENTRY4(uint32_t IoAllocateDriverObjectExtension, drv, clid, extlen, ext)
	driver_object		*drv;
	void			*clid;
	uint32_t		extlen;
	void			**ext;
{
	custom_extension	*ce;

	__FUNCNAME__;
	ce = myExAllocatePoolWithTag(NonPagedPool, sizeof(custom_extension)
	    + extlen, 0);

	if (ce == NULL)
		return(STATUS_INSUFFICIENT_RESOURCES);

	ce->ce_clid = clid;
	INSERT_LIST_TAIL((&drv->dro_driverext->dre_usrext), (&ce->ce_list));

	*ext = (void *)(ce + 1);

	RESTORE();
	return(STATUS_SUCCESS);
}

MSENTRY2(void *IoGetDriverObjectExtension, drv, clid)
	driver_object		*drv;
	void			*clid;
{
	list_entry		*e;
	custom_extension	*ce;

	__FUNCNAME__;
	e = drv->dro_driverext->dre_usrext.nle_flink;
	while (e != &drv->dro_driverext->dre_usrext) {
		ce = (custom_extension *)e;
		if (ce->ce_clid == clid)
			return((void *)(ce + 1));
		e = e->nle_flink;
	}

	cmn_err(CE_CONT,"GetDriverObjectEx NULL\n");
	RESTORE();
	return(NULL);
}


MSENTRY5(uint32_t IoCreateDevice, drv, devextlen, devname, devtype,
					devchars, exclusive, newdev)
	driver_object		*drv;
	uint32_t		devextlen;
	unicode_string		*devname;
	uint32_t		devtype;
	uint32_t		devchars;
	uint8_t			exclusive;
	device_object		**newdev;
{
	device_object		*dev;


	__FUNCNAME__;
	dev = myExAllocatePoolWithTag(NonPagedPool, sizeof(device_object), 0);
	if (dev == NULL)
		return(STATUS_INSUFFICIENT_RESOURCES);

	dev->do_type = devtype;
	dev->do_drvobj = drv;
	dev->do_currirp = NULL;
	dev->do_flags = 0;

	if (devextlen) {
		dev->do_devext = myExAllocatePoolWithTag(NonPagedPool,
		    devextlen, 0);

		if (dev->do_devext == NULL) {
			MSCALL1(ExFreePool, dev);
			return(STATUS_INSUFFICIENT_RESOURCES);
		}
		bzero(dev->do_devext, devextlen);
	} else
		dev->do_devext = NULL;

	dev->do_size = sizeof(device_object) + devextlen;
	dev->do_refcnt = 1;
	dev->do_attacheddev = NULL;
	dev->do_nextdev = NULL;
	dev->do_devtype = devtype;
	dev->do_stacksize = 1;
	dev->do_alignreq = 1;
	dev->do_characteristics = devchars;
	dev->do_iotimer = NULL;
	MSCALL3(KeInitializeEvent, &dev->do_devlock, EVENT_TYPE_SYNC, TRUE);

	/*
	 * Vpd is used for disk/tape devices,
	 * but we don't support those. (Yet.)
	 */
	dev->do_vpb = NULL;

	dev->do_devobj_ext = myExAllocatePoolWithTag(NonPagedPool,
	    sizeof(devobj_extension), 0);

	if (dev->do_devobj_ext == NULL) {
		if (dev->do_devext != NULL)
			MSCALL1(ExFreePool, dev->do_devext);
		MSCALL1(ExFreePool, dev);
		return(STATUS_INSUFFICIENT_RESOURCES);
	}

	dev->do_devobj_ext->dve_type = 0;
	dev->do_devobj_ext->dve_size = sizeof(devobj_extension);
	dev->do_devobj_ext->dve_devobj = dev;

	/*
	 * Attach this device to the driver object's list
	 * of devices. Note: this is not the same as attaching
	 * the device to the device stack. The driver's AddDevice
	 * routine must explicitly call IoAddDeviceToDeviceStack()
	 * to do that.
	 */

	if (drv->dro_devobj == NULL) {
		drv->dro_devobj = dev;
		dev->do_nextdev = NULL;
	} else {
		dev->do_nextdev = drv->dro_devobj;
		drv->dro_devobj = dev;
	}

	*newdev = dev;

	RESTORE();
	return(STATUS_SUCCESS);
}

MSENTRY1(void IoDeleteDevice, dev)
	device_object		*dev;
{
	device_object		*prev;

	__FUNCNAME__;
	if (dev == NULL)
		return;

	if (dev->do_devobj_ext != NULL)
		MSCALL1(ExFreePool, dev->do_devobj_ext);

	if (dev->do_devext != NULL)
		MSCALL1(ExFreePool, dev->do_devext);

	/* Unlink the device from the driver's device list. */

	prev = dev->do_drvobj->dro_devobj;
	if (prev == dev)
		dev->do_drvobj->dro_devobj = dev->do_nextdev;
	else {
		while (prev->do_nextdev != dev)
			prev = prev->do_nextdev;
		prev->do_nextdev = dev->do_nextdev;
	}

	MSCALL1(ExFreePool, dev);

	RESTORE();
	return;
}

MSENTRY1(device_object *IoGetAttachedDevice, dev)
	device_object		*dev;
{
	device_object		*d;

	__FUNCNAME__;
	if (dev == NULL) {
		RESTORE();
		return (NULL);
	}

	d = dev;

	while (d->do_attacheddev != NULL)
		d = d->do_attacheddev;

	RESTORE();
	return (d);
}

MSENTRY4(uint32_t IoGetDeviceObjectPointer, name, desired_access, file_obj,
				dev_obj)
	struct unicode_string *name;
	ACCESS_MASK desired_access;
	void **file_obj;
	struct device_object **dev_obj;
{
	int i=0;
	char buffer[256];

	__FUNCNAME__;
	cmn_err(CE_CONT,"%s: what are you looking for?\n",__func__);
	for (i=0; (2*i<name->us_len) && (i<255); i++)
		buffer[i] = name->us_buf[i];
	buffer[i] = 0;
	cmn_err(CE_CONT,"%s (size %d)\n",buffer,(int)name->us_len);
	if (file_obj)
		*(char**)file_obj = "Fake file object";
	if (dev_obj)
		*dev_obj = (struct device_object*)kmem_zalloc(
					sizeof(struct device_object),KM_SLEEP);
	
	RESTORE();
	return (STATUS_SUCCESS);
}

MSENTRY5(irp *IoBuildSynchronousFsdRequest, func, dobj, buf, len, off, event, status)
	uint32_t		func;
	device_object		*dobj;
	void			*buf;
	uint32_t		len;
	uint64_t		*off;
	nt_kevent		*event;
	io_status_block		*status;
{
	irp			*ip;

	__FUNCNAME__;
	ip = (irp*)MSCALL6(IoBuildAsynchronousFsdRequest,
				func, dobj, buf, len, off, status);
	if (ip == NULL) {
		RESTORE();
		return(NULL);
	}
	ip->irp_usrevent = event;

	RESTORE();
	return(ip);
}

MSENTRY5(irp *IoBuildAsynchronousFsdRequest, func, dobj, buf, len, off, status)
	uint32_t		func;
	device_object		*dobj;
	void			*buf;
	uint32_t		len;
	uint64_t		*off;
	io_status_block		*status;
{
	irp			*ip;
	io_stack_location	*sl;

	__FUNCNAME__;
	ip = (irp*)MSCALL2(IoAllocateIrp, dobj->do_stacksize, TRUE);
	if (ip == NULL) {
		RESTORE();
		return(NULL);
	}

	ip->irp_usriostat = status;
	ip->irp_tail.irp_overlay.irp_thread = NULL;

	sl = IoGetNextIrpStackLocation(ip);
	sl->isl_major = func;
	sl->isl_minor = 0;
	sl->isl_flags = 0;
	sl->isl_ctl = 0;
	sl->isl_devobj = dobj;
	sl->isl_fileobj = NULL;
	sl->isl_completionfunc = NULL;

	ip->irp_userbuf = buf;

	if (dobj->do_flags & DO_BUFFERED_IO) {
		ip->irp_assoc.irp_sysbuf =
		    myExAllocatePoolWithTag(NonPagedPool, len, 0);
		if (ip->irp_assoc.irp_sysbuf == NULL) {
			MSCALL1(IoFreeIrp, ip);
			RESTORE();
			return(NULL);
		}
		bcopy(buf, ip->irp_assoc.irp_sysbuf, len);
	}

	if (dobj->do_flags & DO_DIRECT_IO) {
		ip->irp_mdl = (mdl*)MSCALL5(IoAllocateMdl, buf, len,
				FALSE, FALSE, ip);
		if (ip->irp_mdl == NULL) {
			if (ip->irp_assoc.irp_sysbuf != NULL)
				MSCALL1(ExFreePool, ip->irp_assoc.irp_sysbuf);
			MSCALL1(IoFreeIrp, ip);
			RESTORE();
			return(NULL);
		}
		ip->irp_userbuf = NULL;
		ip->irp_assoc.irp_sysbuf = NULL;
	}

	if (func == IRP_MJ_READ) {
		sl->isl_parameters.isl_read.isl_len = len;
		if (off != NULL)
			sl->isl_parameters.isl_read.isl_byteoff = *off;
		else
			sl->isl_parameters.isl_read.isl_byteoff = 0;
	}

	if (func == IRP_MJ_WRITE) {
		sl->isl_parameters.isl_write.isl_len = len;
		if (off != NULL)
			sl->isl_parameters.isl_write.isl_byteoff = *off;
		else
			sl->isl_parameters.isl_write.isl_byteoff = 0;
	}	

	RESTORE();
	return(ip);
}

MSENTRY5(irp *IoBuildDeviceIoControlRequest, iocode, dobj, ibuf, ilen,
		obuf, olen, isinternal, event, status)
	uint32_t		iocode;
	device_object		*dobj;
	void			*ibuf;
	uint32_t		ilen;
	void			*obuf;
	uint32_t		olen;
	uint8_t			isinternal;
	nt_kevent		*event;
	io_status_block		*status;
{

	irp			*ip;
	io_stack_location	*sl;
	uint32_t		buflen;

	__FUNCNAME__;
	cmn_err(CE_CONT,"%s called\n",__func__);

	ip = (irp*)MSCALL2(IoAllocateIrp, dobj->do_stacksize, TRUE);
	if (ip == NULL) {
		RESTORE();
		return(NULL);
	}
	ip->irp_usrevent = event;
	ip->irp_usriostat = status;
	ip->irp_tail.irp_overlay.irp_thread = NULL;

	sl = IoGetNextIrpStackLocation(ip);
	sl->isl_major = isinternal == TRUE ?
	    IRP_MJ_INTERNAL_DEVICE_CONTROL : IRP_MJ_DEVICE_CONTROL;
	sl->isl_minor = 0;
	sl->isl_flags = 0;
	sl->isl_ctl = 0;
	sl->isl_devobj = dobj;
	sl->isl_fileobj = NULL;
	sl->isl_completionfunc = NULL;
	sl->isl_parameters.isl_ioctl.isl_iocode = iocode;
	sl->isl_parameters.isl_ioctl.isl_ibuflen = ilen;
	sl->isl_parameters.isl_ioctl.isl_obuflen = olen;

	switch(IO_METHOD(iocode)) {
	case METHOD_BUFFERED:
		if (ilen > olen)
			buflen = ilen;
		else
			buflen = olen;
		if (buflen) {
			ip->irp_assoc.irp_sysbuf =
			    myExAllocatePoolWithTag(NonPagedPool, buflen, 0);
			if (ip->irp_assoc.irp_sysbuf == NULL) {
				MSCALL1(IoFreeIrp, ip);
				RESTORE();
				return(NULL);
			}
		}
		if (ilen && ibuf != NULL) {
			bcopy(ibuf, ip->irp_assoc.irp_sysbuf, ilen);
			bzero((char *)ip->irp_assoc.irp_sysbuf + ilen,
			    buflen - ilen);
		} else
			bzero(ip->irp_assoc.irp_sysbuf, ilen);
		ip->irp_userbuf = obuf;
		break;
	case METHOD_IN_DIRECT:
	case METHOD_OUT_DIRECT:
		if (ilen && ibuf != NULL) {
			ip->irp_assoc.irp_sysbuf =
			    myExAllocatePoolWithTag(NonPagedPool, ilen, 0);
			if (ip->irp_assoc.irp_sysbuf == NULL) {
				MSCALL1(IoFreeIrp, ip);
				RESTORE();
				return(NULL);
			}
			bcopy(ibuf, ip->irp_assoc.irp_sysbuf, ilen);
		}
		if (olen && obuf != NULL) {
			ip->irp_mdl = (mdl*)MSCALL5(IoAllocateMdl,obuf, olen,
					FALSE, FALSE, ip);
			/*
			 * Normally we would MmProbeAndLockPages()
			 * here, but we don't have to in our
			 * imlementation.
			 */
		}
		break;
	case METHOD_NEITHER:
		ip->irp_userbuf = obuf;
		sl->isl_parameters.isl_ioctl.isl_type3ibuf = ibuf;
		break;
	default:
		break;
	}

	/*
	 * Ideally, we should associate this IRP with the calling
	 * thread here.
	 */

	RESTORE();
	return (ip);

}

MSENTRY2(irp *IoAllocateIrp, stsize, chargequota)
	uint8_t			stsize;
	uint8_t			chargequota;
{
	irp			*ip;

	__FUNCNAME__;
	ip = myExAllocatePoolWithTag(NonPagedPool, IoSizeOfIrp(stsize), 0);
	if (ip == NULL) {
		RESTORE();
		return (NULL);
	}

	MSCALL3(IoInitializeIrp, ip, IoSizeOfIrp(stsize), stsize);

	RESTORE();
	return (ip);
}

MSENTRY2(irp *IoMakeAssociatedIrp, ip, stsize)
	irp			*ip;
	uint8_t			stsize;
{
	irp			*associrp;

	__FUNCNAME__;
	associrp = (irp*)MSCALL2(IoAllocateIrp, stsize, FALSE);
	if (associrp == NULL) {
		RESTORE();
		return(NULL);
	}

	mutex_enter(&ntoskrnl_dispatchlock);
	associrp->irp_flags |= IRP_ASSOCIATED_IRP;
	associrp->irp_tail.irp_overlay.irp_thread =
	    ip->irp_tail.irp_overlay.irp_thread;
	associrp->irp_assoc.irp_master = ip;
	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return(associrp);
}

MSENTRY1(void IoFreeIrp, ip)
	irp			*ip;
{
	__FUNCNAME__;
	MSCALL1(ExFreePool, ip);
	RESTORE();
	return;
}

MSENTRY3(void IoInitializeIrp, io, psize, ssize)
	irp			*io;
	uint16_t		psize;
	uint8_t			ssize;
{
	__FUNCNAME__;
	bzero((char *)io, sizeof(irp));
	io->irp_size = psize;
	io->irp_stackcnt = ssize;
	io->irp_currentstackloc = ssize;
	INIT_LIST_HEAD(&io->irp_thlist);
	io->irp_tail.irp_overlay.irp_csl =
	    (io_stack_location *)(io + 1) + ssize;

	RESTORE();
	return;
}

MSENTRY2(void IoReuseIrp, ip, status)
	irp			*ip;
	uint32_t		status;
{
	uint8_t			allocflags;

	__FUNCNAME__;
	allocflags = ip->irp_allocflags;
	MSCALL3(IoInitializeIrp, ip, ip->irp_size, ip->irp_stackcnt);
	ip->irp_iostat.isb_status = status;
	ip->irp_allocflags = allocflags;

	RESTORE();
	return;
}

#if defined(__amd64__) & defined(LINKEDIMAGE)
MSENTRY2(uint32_t IofCallDriver, dobj, ip)
	device_object *dobj;
	irp *ip;
#else
__fastcall uint32_t
IofCallDriver(REGARGS2(device_object *dobj, irp *ip))
#endif
{
	driver_object		*drvobj;
	io_stack_location	*sl;
	uint32_t		status;
	driver_dispatch		disp;

	drvobj = dobj->do_drvobj;

	__FUNCNAME__;
	if (ip->irp_currentstackloc <= 0)
		panic("IofCallDriver(): out of stack locations");

	IoSetNextIrpStackLocation(ip);
	sl = IoGetCurrentIrpStackLocation(ip);

	sl->isl_devobj = dobj;

	if (drvobj) {
		disp = drvobj->dro_dispatch[sl->isl_major];
		status = MSCALL2(disp, dobj, ip);
	} else {
		cmn_err(CE_WARN,"No drvobj, isl_major %d\n",(int)sl->isl_major);
		status = STATUS_INVALID_DEVICE_REQUEST;
	}

	RESTORE();
	return(status);
}

#if defined(__amd64__) & defined(LINKEDIMAGE)
MSENTRY2(void IofCompleteRequest, ip, prioboost)
	irp *ip;
	uint8_t prioboost;
#else
__fastcall void
IofCompleteRequest(REGARGS2(irp *ip, uint8_t prioboost))
#endif
{
	uint32_t		i;
	uint32_t		status;
	device_object		*dobj;
	io_stack_location	*sl;
	completion_func		cf;

	__FUNCNAME__;
	ip->irp_pendingreturned =
	    IoGetCurrentIrpStackLocation(ip)->isl_ctl & SL_PENDING_RETURNED;
	sl = (io_stack_location *)(ip + 1);

	for (i = ip->irp_currentstackloc; i < (uint32_t)ip->irp_stackcnt; i++) {
		if (ip->irp_currentstackloc < ip->irp_stackcnt - 1) {
			IoSkipCurrentIrpStackLocation(ip);
			dobj = IoGetCurrentIrpStackLocation(ip)->isl_devobj;
		} else
			dobj = NULL;

		if (sl[i].isl_completionfunc != NULL &&
		    ((ip->irp_iostat.isb_status == STATUS_SUCCESS &&
		    sl->isl_ctl & SL_INVOKE_ON_SUCCESS) ||
		    (ip->irp_iostat.isb_status != STATUS_SUCCESS &&
		    sl->isl_ctl & SL_INVOKE_ON_ERROR) ||
		    (ip->irp_cancel == TRUE &&
		    sl->isl_ctl & SL_INVOKE_ON_CANCEL))) {
			cf = sl->isl_completionfunc;
			status = MSCALL3(cf, dobj, ip, sl->isl_completionctx);
			if (status == STATUS_MORE_PROCESSING_REQUIRED) {
				RESTORE();
				return;
			}
		}

		if (IoGetCurrentIrpStackLocation(ip)->isl_ctl &
		    SL_PENDING_RETURNED)
			ip->irp_pendingreturned = TRUE;
	}

	/* Handle any associated IRPs. */

	if (ip->irp_flags & IRP_ASSOCIATED_IRP) {
		uint32_t		masterirpcnt;
		irp			*masterirp;
		mdl			*m;

		masterirp = ip->irp_assoc.irp_master;
		masterirpcnt = FASTCALL1(InterlockedDecrement,
		    &masterirp->irp_assoc.irp_irpcnt);

		while ((m = ip->irp_mdl) != NULL) {
			ip->irp_mdl = m->mdl_next;
			MSCALL1(IoFreeMdl, m);
		}
		MSCALL1(IoFreeIrp, ip);
#if defined(__amd64__) & defined(LINKEDIMAGE)
		if (masterirpcnt == 0)
			MSCALL2(IofCompleteRequest, masterirp, IO_NO_INCREMENT);
#else
		if (masterirpcnt == 0)
			IoCompleteRequest(masterirp, IO_NO_INCREMENT);
#endif
		RESTORE();
		return;
	}

	/* With any luck, these conditions will never arise. */

	if (ip->irp_flags & (IRP_PAGING_IO|IRP_CLOSE_OPERATION)) {
		if (ip->irp_usriostat != NULL)
			*ip->irp_usriostat = ip->irp_iostat;
		if (ip->irp_usrevent != NULL)
			MSCALL3(KeSetEvent, ip->irp_usrevent, prioboost, FALSE);
		if (ip->irp_flags & IRP_PAGING_IO) {
			if (ip->irp_mdl != NULL)
				MSCALL1(IoFreeMdl, ip->irp_mdl);
			MSCALL1(IoFreeIrp, ip);
		}
	}

	RESTORE();
	return;
}

MSENTRY2(device_object *IoAttachDeviceToDeviceStack, src, dst)
	device_object		*src;
	device_object		*dst;
{
	device_object		*attached;

	__FUNCNAME__;
	mutex_enter(&ntoskrnl_dispatchlock);

	attached = (device_object*)MSCALL1(IoGetAttachedDevice, dst);
	attached->do_attacheddev = src;
	src->do_attacheddev = NULL;
	src->do_stacksize = attached->do_stacksize + 1;

	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return(attached);
}

MSENTRY1(void IoDetachDevice, topdev)
	device_object		*topdev;
{
	device_object		*tail;

	__FUNCNAME__;
	mutex_enter(&ntoskrnl_dispatchlock);

	/* First, break the chain. */
	tail = topdev->do_attacheddev;
	if (tail == NULL) {
		mutex_exit(&ntoskrnl_dispatchlock);
		RESTORE();
		return;
	}
	topdev->do_attacheddev = tail->do_attacheddev;
	topdev->do_refcnt--;

	/* Now reduce the stacksize count for the tail objects. */

	tail = topdev->do_attacheddev;
	while (tail != NULL) {
		tail->do_stacksize--;
		tail = tail->do_attacheddev;
	}

	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return;
}

/* Always called with dispatcher lock held. */
static void
ntoskrnl_wakeup(arg)
	void			*arg;
{
	nt_dispatch_header	*obj;
	wait_block		*w;
	list_entry		*e;
	kthread_t		*td;

	obj = arg;
	__FUNCNAME__;

//	mutex_enter(&ntoskrnl_dispatchlock);
	obj->dh_sigstate = TRUE;
	e = obj->dh_waitlisthead.nle_flink;
	/*
	 * This has asserted with e == 0, however, it probably should never
	 * happen.  There was a missing hold of the dispatcher lock which has been
	 * corrected, but just in case, test for a valid pointer as well.
	 */
	while ((e != NULL) && (e != &obj->dh_waitlisthead)) {
		w = (wait_block *)e;
		td = w->wb_kthread;
		ndis_thresume(&td->t_delay_cv);
		/*
		 * For synchronization objects, only wake up
		 * the first waiter.
		 */
		if (obj->dh_type == EVENT_TYPE_SYNC)
			break;
		e = e->nle_flink;
	}
//	mutex_exit(&ntoskrnl_dispatchlock);

	return;
}

static void 
ntoskrnl_time(tval)
	uint64_t                *tval;
{
	time_t now;	
	uint64_t t;	
	
	__FUNCNAME__;	
	now = ddi_get_time();	
	t = (uint64_t) now * TICKSPERSEC;	
	t += TICKS_1601_TO_1970;	
	*tval = t;
	return;
}

/*
 * KeWaitForSingleObject() is a tricky beast, because it can be used
 * with several different object types: semaphores, timers, events,
 * mutexes and threads. Semaphores don't appear very often, but the
 * other object types are quite common. KeWaitForSingleObject() is
 * what's normally used to acquire a mutex, and it can be used to
 * wait for a thread termination.
 *
 * The Windows NDIS API is implemented in terms of Windows kernel
 * primitives, and some of the object manipulation is duplicated in
 * NDIS. For example, NDIS has timers and events, which are actually
 * Windows kevents and ktimers. Now, you're supposed to only use the
 * NDIS variants of these objects within the confines of the NDIS API,
 * but there are some naughty developers out there who will use
 * KeWaitForSingleObject() on NDIS timer and event objects, so we
 * have to support that as well. Conseqently, our NDIS timer and event
 * code has to be closely tied into our ntoskrnl timer and event code,
 * just as it is in Windows.
 *
 * KeWaitForSingleObject() may do different things for different kinds
 * of objects:
 *
 * - For events, we check if the event has been signalled. If the
 *   event is already in the signalled state, we just return immediately,
 *   otherwise we wait for it to be set to the signalled state by someone
 *   else calling KeSetEvent(). Events can be either synchronization or
 *   notification events.
 *
 * - For timers, if the timer has already fired and the timer is in
 *   the signalled state, we just return, otherwise we wait on the
 *   timer. Unlike an event, timers get signalled automatically when
 *   they expire rather than someone having to trip them manually.
 *   Timers initialized with KeInitializeTimer() are always notification
 *   events: KeInitializeTimerEx() lets you initialize a timer as
 *   either a notification or synchronization event.
 *
 * - For mutexes, we try to acquire the mutex and if we can't, we wait
 *   on the mutex until it's available and then grab it. When a mutex is
 *   released, it enters the signaled state, which wakes up one of the
 *   threads waiting to acquire it. Mutexes are always synchronization
 *   events.
 *
 * - For threads, the only thing we do is wait until the thread object
 *   enters a signalled state, which occurs when the thread terminates.
 *   Threads are always notification events.
 *
 * A notification event wakes up all threads waiting on an object. A
 * synchronization event wakes up just one. Also, a synchronization event
 * is auto-clearing, which means we automatically set the event back to
 * the non-signalled state once the wakeup is done.
 */

MSENTRY5(uint32_t KeWaitForSingleObject, obj, reason, mode, alertable, duetime)
	nt_dispatch_header	*obj;
	uint32_t		reason;
	uint32_t		mode;
	uint8_t			alertable;
	int64_t			*duetime;
{
	kthread_t		*td = curthread;
	kmutant			*km;
	wait_block		w;
	int			error = 0;
	uint64_t		curtime;
	uint32_t		tv_usec;

	if (obj == NULL) {
		RESTORE();
		return(STATUS_INVALID_PARAMETER);
	}

	__FUNCNAME__;
	//cmn_err(CE_CONT,"%s: in\n",__func__);
	mutex_enter(&ntoskrnl_dispatchlock);

	/*
	 * See if the object is a mutex. If so, and we already own
	 * it, then just increment the acquisition count and return.
         *
         * For any other kind of object, see if it's already in the
	 * signalled state, and if it is, just return. If the object
         * is marked as a synchronization event, reset the state to
         * unsignalled.
	 */

	if (obj->dh_size == OTYPE_MUTEX) {
		km = (kmutant *)obj;
		if (km->km_ownerthread == NULL ||
		    km->km_ownerthread == curthread) {
			obj->dh_sigstate = FALSE;
			km->km_acquirecnt++;
			km->km_ownerthread = curthread;
			mutex_exit(&ntoskrnl_dispatchlock);
			RESTORE();
			return (STATUS_SUCCESS);
		}
	} else if (obj->dh_sigstate == TRUE) {
		if (obj->dh_type == EVENT_TYPE_SYNC)
			obj->dh_sigstate = FALSE;
		mutex_exit(&ntoskrnl_dispatchlock);
		RESTORE();
		return (STATUS_SUCCESS);
	}

	w.wb_object = obj;
	w.wb_kthread = curthread;

	INSERT_LIST_TAIL((&obj->dh_waitlisthead), (&w.wb_waitlist));

	/*
	 * The timeout value is specified in 100 nanosecond units
	 * and can be a positive or negative number. If it's positive,
	 * then the duetime is absolute, and we need to convert it
	 * to an absolute offset relative to now in order to use it.
	 * If it's negative, then the duetime is relative and we
	 * just have to convert the units.
	 */

	if (duetime != NULL) {
		if (*duetime < 0) {
			tv_usec = -(*duetime) / 10;
		} else {
			ntoskrnl_time(&curtime);
			if (*duetime < (int64_t)curtime)
				tv_usec = 0;
			else
				tv_usec = ((*duetime) - curtime)/10;
		}
	} else {
	    tv_usec = 60000000; /* use 1mn to mean waiting forever */
	}
	/*
	 * If zero, set to a small values, otherwise, it may never time out
	 */
	if(tv_usec == 0)
	    tv_usec = drv_hztousec(1); /* at least one tick */

	mutex_exit(&ntoskrnl_dispatchlock);

	//cmn_err(CE_CONT,"%s: will thsuspend myself\n",__func__);
	error = ndis_thsuspend(&curthread->t_delay_cv, &curthread->t_delay_lock,
		    (clock_t)drv_usectohz(tv_usec));
	//cmn_err(CE_CONT,"%s: will wakeup myself\n",__func__);

	mutex_enter(&ntoskrnl_dispatchlock);

	/* We timed out. Leave the object alone and return status. */

	//if (error == EWOULDBLOCK) {
	if (error == -1) {
		REMOVE_LIST_ENTRY((&w.wb_waitlist));
		mutex_exit(&ntoskrnl_dispatchlock);
		RESTORE();
		return(STATUS_TIMEOUT);
	}
	/*
	 * Mutexes are always synchronization objects, which means
         * if several threads are waiting to acquire it, only one will
         * be woken up. If that one is us, and the mutex is up for grabs,
         * grab it.
	 */

	if (obj->dh_size == OTYPE_MUTEX) {
		km = (kmutant *)obj;
		if (km->km_ownerthread == NULL) {
			km->km_ownerthread = curthread;
			km->km_acquirecnt++;
		}
	}

	if (obj->dh_type == EVENT_TYPE_SYNC)
		obj->dh_sigstate = FALSE;
	REMOVE_LIST_ENTRY((&w.wb_waitlist));

	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return(STATUS_SUCCESS);
}

MSENTRY5(uint32_t KeWaitForMultipleObjects, cnt, obj, wtype, reason, mode,
	alertable, duetime, wb_array)
	uint32_t		cnt;
	nt_dispatch_header	*obj[];
	uint32_t		wtype;
	uint32_t		reason;
	uint32_t		mode;
	uint8_t			alertable;
	int64_t			*duetime;
	wait_block		*wb_array;
{
	__FUNCNAME__;
	RESTORE();
	return(STATUS_SUCCESS);
}

extern uint32_t ndis_iolength[6];

int idx_counter(void *reg)
{
	if ((reg >= curSC->ioreg_ndis[0]) && (reg <= curSC->ioreg_ndis[0] + ndis_iolength[0])){
	//	cmn_err(CE_CONT,"reg = 0x%p, ioreg[0]=0x%p, len = 0x%p\n",
	//		reg, curSC->ioreg_ndis[0], ndis_iolength[0]);
		return 0;
	}
	if ((reg >= curSC->ioreg_ndis[1]) && (reg <= curSC->ioreg_ndis[1] + ndis_iolength[1]))
		return 1;
	if ((reg >= curSC->ioreg_ndis[2]) && (reg <= curSC->ioreg_ndis[2] + ndis_iolength[2]))
		return 2;
	if ((reg >= curSC->ioreg_ndis[3]) && (reg <= curSC->ioreg_ndis[3] + ndis_iolength[3]))
		return 3;
	if ((reg >= curSC->ioreg_ndis[4]) && (reg <= curSC->ioreg_ndis[4] + ndis_iolength[4]))
		return 4;
	if ((reg >= curSC->ioreg_ndis[5]) && (reg <= curSC->ioreg_ndis[5] + ndis_iolength[5]))
		return 5;
	cmn_err(CE_CONT,"Incorrect register mapping \n");
	return (-1);
}

MSENTRY2(void WRITE_REGISTER_USHORT, reg, val)
	uint16_t		*reg;
	uint16_t		val;
{
	int idx=0;
	
	__FUNCNAME__;
	LOCK_REG(&register_mtx);
	idx = idx_counter(reg); 
        ddi_put16(curSC->ndis_PCI_bar2_handle[idx], (uint16_t *)reg, val);
	UNLOCK_REG(&register_mtx);
	RESTORE();
	return;
}

MSENTRY1(uint16_t READ_REGISTER_USHORT, reg)
	uint16_t		*reg;
{
	uint16_t rval=0;
	int idx=0;

	__FUNCNAME__S;
	LOCK_REG(&register_mtx);
	//rval = *(volatile uint16_t *)(reg);
	idx = idx_counter(reg); 
        rval = ddi_get16(curSC->ndis_PCI_bar2_handle[idx], (uint16_t *)reg);
	UNLOCK_REG(&register_mtx);
	RESTORE();
	return (rval);

}

MSENTRY2(void WRITE_REGISTER_ULONG, reg, val)
	uint32_t		*reg;
	uint32_t		val;
{
	__FUNCNAME__S;
	int idx;
	//LOCK_REG(&register_mtx);
	idx = idx_counter(reg); 
	//cmn_err(CE_CONT,"%s(0x%p, 0x%x), idx = %d\n",__func__, reg,val,idx);
        ddi_put32(curSC->ndis_PCI_bar2_handle[idx], (uint32_t *)reg, val);
	//UNLOCK_REG(&register_mtx);
	RESTORE();
	return;
}

MSENTRY1(uint32_t READ_REGISTER_ULONG, reg)
	uint32_t		*reg;
{
	uint32_t rval=0;
	int idx;

	__FUNCNAME__S;
	cmn_err(CE_CONT,"%s(0x%p)\n",__func__, reg);
	//LOCK_REG(&register_mtx);
	idx = idx_counter(reg); 
        rval = ddi_get32(curSC->ndis_PCI_bar2_handle[idx], (uint32_t *)reg);
	//UNLOCK_REG(&register_mtx);	
	RESTORE();
	return (rval);
}

MSENTRY1(uint8_t READ_REGISTER_UCHAR, reg)
	uint8_t			*reg;
{
	uint8_t rval=0;
	int idx=0;
	__FUNCNAME__S;
	//LOCK_REG(&register_mtx);
	idx = idx_counter(reg); 
        rval = ddi_get8(curSC->ndis_PCI_bar2_handle[idx], (uint8_t *)reg);
	//UNLOCK_REG(&register_mtx);
	RESTORE();
	return (rval);

}

MSENTRY2(void WRITE_REGISTER_UCHAR, reg, val)
	uint8_t			*reg;
	uint8_t			val;
{
	int idx= 0;
	__FUNCNAME__S;
	//LOCK_REG(&register_mtx);
	idx = idx_counter(reg); 
        ddi_put8(curSC->ndis_PCI_bar2_handle[idx], (uint8_t *)reg, val);
	//UNLOCK_REG(&register_mtx);
	RESTORE();
	return;
}

MSENTRY2(int64_t _allmul, a, b)
	int64_t			a;
	int64_t			b;
{
	uint64_t ret;

	__FUNCNAME__;
	ret = a * b;
	RESTORE();
	return (ret);
}

MSENTRY2(int64_t _alldiv, a, b)
	int64_t			a;
	int64_t			b;
{
	uint64_t ret;

	__FUNCNAME__;
	ret = a / b;
	RESTORE();
	return (ret);
}

MSENTRY2(int64_t _allrem, a, b)
	int64_t			a;
	int64_t			b;
{
	uint64_t ret;

	__FUNCNAME__;
	ret = a % b;
	RESTORE();
	return (ret);
}

MSENTRY2(uint64_t _aullmul, a, b)
	uint64_t		a;
	uint64_t		b;
{
	uint64_t ret;

	__FUNCNAME__;
	ret = a * b;
	RESTORE();
	return (ret);
}

MSENTRY2(uint64_t _aulldiv, a, b)
	uint64_t		a;
	uint64_t		b;
{
	uint64_t ret;

	__FUNCNAME__;
	ret = a / b;
	RESTORE();
	return (ret);
}

MSENTRY2(uint64_t _aullrem,a, b)
	uint64_t		a;
	uint64_t		b;
{
	uint64_t ret;

	__FUNCNAME__;
	ret = a % b;
	RESTORE();
	return (ret);
}

__regparm static int64_t
_allshl(a, b)
	int64_t			a;
	uint8_t			b;
{
	__FUNCNAME__;
	return (a << b);
}

__regparm static uint64_t
_aullshl(a, b)
	uint64_t		a;
	uint8_t			b;
{
	__FUNCNAME__;
	return (a << b);
}

__regparm static int64_t
_allshr(a, b)
	int64_t			a;
	uint8_t			b;
{
	__FUNCNAME__;
	return (a >> b);
}

__regparm static uint64_t
_aullshr(a, b)
	uint64_t		a;
	uint8_t			b;
{
	__FUNCNAME__;
	return (a >> b);
}

static slist_entry *
ntoskrnl_pushsl(head, entry)
	slist_header		*head;
	slist_entry		*entry;
{
	slist_entry		*oldhead;
	__FUNCNAME__;

	oldhead = head->slh_list.slh_next;
	entry->sl_next = head->slh_list.slh_next;
	head->slh_list.slh_next = entry;
	head->slh_list.slh_depth++;
	head->slh_list.slh_seq++;

	return(oldhead);
}

static slist_entry *
ntoskrnl_popsl(head)
	slist_header		*head;
{
	slist_entry		*first;
	__FUNCNAME__;

	first = head->slh_list.slh_next;
	if (first != NULL) {
		head->slh_list.slh_next = first->sl_next;
		head->slh_list.slh_depth--;
		head->slh_list.slh_seq++;
	}

	return(first);
}

/*
 * We need this to make lookaside lists work for amd64.
 * We pass a pointer to ExAllocatePoolWithTag() the lookaside
 * list structure. For amd64 to work right, this has to be a
 * pointer to the wrapped version of the routine, not the
 * original. Letting the Windows driver invoke the original
 * function directly will result in a convention calling
 * mismatch and a pretty crash. On x86, this effectively
 * becomes a no-op since ipt_func and ipt_wrap are the same.
 */

static funcptr
ntoskrnl_findwrap(func)
	funcptr			func;
{
	image_patch_table	*patch;

	patch = ntoskrnl_functbl;
	while (patch->ipt_func != NULL) {
		if ((funcptr)patch->ipt_func == func)
			return((funcptr)patch->ipt_wrap);
		patch++;
	}

	return(NULL);
}

MSENTRY5(void ExInitializePagedLookasideList, lookaside, allocfunc, freefunc,
    flags, size, tag, depth)
	paged_lookaside_list	*lookaside;
	lookaside_alloc_func	allocfunc; /* JPA was *allocfunc; */
	lookaside_free_func	freefunc; /* JPA was *freefunc; */
	uint32_t		flags;
	size_t			size;
	uint32_t		tag;
	uint16_t		depth;
{
	__FUNCNAME__;
	bzero((char *)lookaside, sizeof(paged_lookaside_list));

	if (size < sizeof(slist_entry))
		lookaside->nll_l.gl_size = sizeof(slist_entry);
	else
		lookaside->nll_l.gl_size = size;
	lookaside->nll_l.gl_tag = tag;
	if (allocfunc == NULL) {
#ifdef LINKEDIMAGE
		lookaside->nll_l.gl_allocfunc = ExAllocatePoolWithTag;
#else
		lookaside->nll_l.gl_allocfunc =
		    ntoskrnl_findwrap((funcptr)ExAllocatePoolWithTag);
#endif
	} else
		lookaside->nll_l.gl_allocfunc = allocfunc;

	if (freefunc == NULL) {
#ifdef LINKEDIMAGE
		lookaside->nll_l.gl_freefunc = ExFreePool;
#else
		lookaside->nll_l.gl_freefunc =
		    ntoskrnl_findwrap((funcptr)ExFreePool);
#endif
	} else
		lookaside->nll_l.gl_freefunc = freefunc;

#ifdef __i386__
	KiInitializeSpinLock(&lookaside->nll_obsoletelock);
#endif

	lookaside->nll_l.gl_type = NonPagedPool;
	lookaside->nll_l.gl_depth = depth;
	lookaside->nll_l.gl_maxdepth = LOOKASIDE_DEPTH;

	RESTORE();
	return;
}

MSENTRY1(void ExDeletePagedLookasideList, lookaside)
	paged_lookaside_list   *lookaside;
{
	void			*buf;
	lookaside_free_func	freefunc;
	__FUNCNAME__;

	freefunc = lookaside->nll_l.gl_freefunc;
	while((buf = ntoskrnl_popsl(&lookaside->nll_l.gl_listhead)) != NULL)
		MSCALL1(freefunc, buf);

	RESTORE();
	return;
}

MSENTRY5(void ExInitializeNPagedLookasideList, lookaside, allocfunc, freefunc,
    flags, size, tag, depth)
	npaged_lookaside_list	*lookaside;
	lookaside_alloc_func	allocfunc; /* JPA was *allocfunc; */
	lookaside_free_func	freefunc; /* JPA was *freefunc; */
	uint32_t		flags;
	size_t			size;
	uint32_t		tag;
	uint16_t		depth;
{
	__FUNCNAME__;
	bzero((char *)lookaside, sizeof(npaged_lookaside_list));

	if (size < sizeof(slist_entry))
		lookaside->nll_l.gl_size = sizeof(slist_entry);
	else
		lookaside->nll_l.gl_size = size;
	lookaside->nll_l.gl_tag = tag;
	if (allocfunc == NULL) {
#ifdef LINKEDIMAGE
		lookaside->nll_l.gl_allocfunc = ExAllocatePoolWithTag;
#else
		lookaside->nll_l.gl_allocfunc =
		    ntoskrnl_findwrap((funcptr)ExAllocatePoolWithTag);
#endif
	} else {
		/* JPA unsure whether allocfunc or *allocfunc, no known call */
		cmn_err(CE_WARN,"Suspicious arg allocfunc 0x%lx\n",(long)allocfunc);
		lookaside->nll_l.gl_allocfunc = allocfunc;
	}

	if (freefunc == NULL) {
#ifdef LINKEDIMAGE
		lookaside->nll_l.gl_freefunc = ExFreePool;
#else
		lookaside->nll_l.gl_freefunc =
		    ntoskrnl_findwrap((funcptr)ExFreePool);
#endif
	} else {
		/* JPA unsure whether freefunc or *freefunc, no known call */
		cmn_err(CE_WARN,"Suspicious arg freefunc 0x%lx\n",(long)freefunc);
		lookaside->nll_l.gl_freefunc = freefunc;
	}

#ifdef __i386__
	KiInitializeSpinLock(&lookaside->nll_obsoletelock);
#endif

	lookaside->nll_l.gl_type = NonPagedPool;
	lookaside->nll_l.gl_depth = depth;
	lookaside->nll_l.gl_maxdepth = LOOKASIDE_DEPTH;

	RESTORE();
	return;
}

MSENTRY1(void ExDeleteNPagedLookasideList, lookaside)
	npaged_lookaside_list   *lookaside;
{
	void			*buf;
	lookaside_free_func	freefunc;
	__FUNCNAME__;
	freefunc = lookaside->nll_l.gl_freefunc;
	while((buf = ntoskrnl_popsl(&lookaside->nll_l.gl_listhead)) != NULL)
		MSCALL1(freefunc, buf);
	RESTORE();
	return;
}

/*
 * Note: the interlocked slist push and pop routines are
 * declared to be _fastcall in Windows. gcc 3.4 is supposed
 * to have support for this calling convention, however we
 * don't have that version available yet, so we kludge things
 * up using __regparm__(3) and some argument shuffling.
 */

__fastcall static slist_entry *
InterlockedPushEntrySList(REGARGS2(slist_header *head, slist_entry *entry))
{
	slist_entry		*oldhead;
	__FUNCNAME__;

	oldhead = (slist_entry *)FASTCALL3(ExInterlockedPushEntrySList,
	    head, entry, &ntoskrnl_global);
	return(oldhead);
}

__fastcall static slist_entry *
InterlockedPopEntrySList(REGARGS1(slist_header *head))
{
	slist_entry		*first;
	__FUNCNAME__;
	first = (slist_entry *)FASTCALL2(ExInterlockedPopEntrySList,
	    head, &ntoskrnl_global);

	return(first);
}

__fastcall static slist_entry *
ExInterlockedPushEntrySList(REGARGS2(slist_header *head,
	slist_entry *entry), kspin_lock *lock)
{
	slist_entry		*oldhead;
	uint8_t			irql;
	__FUNCNAME__;

	KmAcquireSpinLock(lock, &irql);
	oldhead = ntoskrnl_pushsl(head, entry);
	KmReleaseSpinLock(lock, irql);

	return(oldhead);
}

__fastcall static slist_entry *
ExInterlockedPopEntrySList(REGARGS2(slist_header *head, kspin_lock *lock))
{
	slist_entry		*first;
	uint8_t			irql;

	__FUNCNAME__;
	KmAcquireSpinLock(lock, &irql);
	first = ntoskrnl_popsl(head);
	KmReleaseSpinLock(lock, irql);

	return(first);
}

MSENTRY1(uint16_t ExQueryDepthSList, head)
	slist_header		*head;
{
	uint16_t		depth;
	uint8_t			irql;
	
	__FUNCNAME__;
	KmAcquireSpinLock(&ntoskrnl_global, &irql);
	depth = head->slh_list.slh_depth;
	KmReleaseSpinLock(&ntoskrnl_global, irql);

	RESTORE();
	return(depth);
}
/*
 * The KeInitializeSpinLock(), KefAcquireSpinLockAtDpcLevel()
 * and KefReleaseSpinLockFromDpcLevel() appear to be analagous
 * to splnet()/splx() in their use. We can't create a new mutex
 * lock here because there is no complimentary KeFreeSpinLock()
 * function. Instead, we grab a mutex from the mutex pool.
 */

MSENTRY1(void KeInitializeSpinLock, lock)
	kspin_lock		*lock;
{
	__FUNCNAME__;
	LOCK_INIT_CLEAR((lock_t *)lock);
	mutex_init((kmutex_t *)lock, "ntos spin lock mutex",
		MUTEX_DEFAULT, NULL);	
	RESTORE();
	return;
}

void KiInitializeSpinLock(lock) /* internal use only */
	kspin_lock		*lock;
{
	__FUNCNAME__;
	LOCK_INIT_CLEAR((lock_t *)lock);
	mutex_init((kmutex_t *)lock, "ntos spin lock mutex",
		MUTEX_DEFAULT, NULL);	
	return;
}

#ifdef __i386__
__fastcall void
KefAcquireSpinLockAtDpcLevel(REGARGS1(kspin_lock *lock))
{
	__FUNCNAME__;
	//lock_set((lock_t *)lock);
	mutex_enter((kmutex_t *)lock);
	return;
}

__fastcall void
KefReleaseSpinLockFromDpcLevel(REGARGS1(kspin_lock *lock))
{
	__FUNCNAME__;
	//lock_clear((lock_t *)lock);
	mutex_exit((kmutex_t *)lock);
	return;
}

// try merging with the amd64 variant
MSENTRY1(uint8_t KeAcquireSpinLockRaiseToDpc, lock)
	kspin_lock *lock;
{
        uint8_t                 oldirql;
	__FUNCNAME__;

        if (KfGetCurrentIrql() > DISPATCH_LEVEL)
                panic("IRQL_NOT_LESS_THAN_OR_EQUAL");

        oldirql = KmRaiseIrql(DISPATCH_LEVEL);
        KeAcquireSpinLockAtDpcLevel(lock);

	RESTORE();
        return(oldirql);
}
#else

MSENTRY1(void KeAcquireSpinLockAtDpcLevel, lock)
	kspin_lock *lock;
{
	__FUNCNAME__;
	//lock_set((lock_t *)lock);
	mutex_enter((kmutex_t *)lock);
	RESTORE();
	return;
}

MSENTRY1(void KeReleaseSpinLockFromDpcLevel, lock)
	kspin_lock *lock;
{
	__FUNCNAME__;
	//lock_clear((lock_t *)lock);
	mutex_exit((kmutex_t *)lock);
	RESTORE();
	return;
}

MSENTRY1(uint8_t KeAcquireSpinLockRaiseToDpc, lock)
	kspin_lock *lock;
{
	int8_t res;

	res = KfAcquireSpinLock(lock);
	__FUNCNAME__;
	RESTORE();
	return (res);
}
#endif /* __i386__ */

__fastcall uintptr_t
InterlockedExchange(REGARGS2(volatile uint32_t *dst, uintptr_t val))
{
	uint8_t			irql;
	uintptr_t		r;

	__FUNCNAME__;
	KmAcquireSpinLock(&ntoskrnl_global, &irql);
	r = *dst;
	*dst = val;
	KmReleaseSpinLock(&ntoskrnl_global, irql);

	return(r);
}

__fastcall static uint32_t
InterlockedIncrement(REGARGS1(volatile uint32_t *addend))
{
	__FUNCNAME__;
/*
	atomic_add_long((volatile u_long *)addend, 1);
*/	return(*addend);
}

__fastcall static uint32_t
InterlockedDecrement(REGARGS1(volatile uint32_t *addend))
{
	__FUNCNAME__;
/*
	atomic_subtract_long((volatile u_long *)addend, 1);
*/	return(*addend);
}

__fastcall static void
ExInterlockedAddLargeStatistic(REGARGS2(uint64_t *addend, uint32_t inc))
{
	uint8_t			irql;

	__FUNCNAME__;
	KmAcquireSpinLock(&ntoskrnl_global, &irql);
	*addend += inc;
	KmReleaseSpinLock(&ntoskrnl_global, irql);

	return;
};

MSENTRY5(mdl *IoAllocateMdl, vaddr, len, secondarybuf, chargequota, iopkt)
	void			*vaddr;
	uint32_t		len;
	uint8_t			secondarybuf;
	uint8_t			chargequota;
	irp			*iopkt;
{
	mdl			*m;
	int			zone = 0;

	__FUNCNAME__;	

	if (MSCALL2(MmSizeOfMdl, vaddr, len) > MDL_ZONE_SIZE){
		m = myExAllocatePoolWithTag(NonPagedPool,
		    MSCALL2(MmSizeOfMdl, vaddr, len), 0);
	}else {
		//cmn_err(CE_CONT,"%s: correct1\n",__func__);
		//m = kmem_cache_alloc(mdl_cache, KM_NOSLEEP);
		m = myExAllocatePoolWithTag(NonPagedPool,
			MSCALL2(MmSizeOfMdl, vaddr, len), 0);
		zone++;
	}

	if (m == NULL) {
		RESTORE();
		return (NULL);
	}
	//cmn_err(CE_CONT,"%s: correct2\n",__func__);
	MmInitializeMdl(m, vaddr, len);

	/*
	 * MmInitializMdl() clears the flags field, so we
	 * have to set this here. If the MDL came from the
	 * MDL UMA zone, tag it so we can release it to
	 * the right place later.
	 */
	if (zone)
		m->mdl_flags = MDL_ZONE_ALLOCED;

	if (iopkt != NULL) {
		if (secondarybuf == TRUE) {
			mdl			*last;
			last = iopkt->irp_mdl;
			while (last->mdl_next != NULL)
				last = last->mdl_next;
			last->mdl_next = m;
		} else {
			if (iopkt->irp_mdl != NULL)
				panic("leaking an MDL in IoAllocateMdl()");
			iopkt->irp_mdl = m;
		}
	}

	RESTORE();
	return (m);
}

#if 0 /* Unused, probably experimental */

mdl * 
TxAllocateMdl(vaddr, len, secondarybuf, chargequota, iopkt)
	void			*vaddr;
	uint32_t		len;
	uint8_t			secondarybuf;
	uint8_t			chargequota;
	irp			*iopkt;
{
	mdl			*m;
	int			zone = 0;

	__FUNCNAME__;	

	if (MSCALL2(MmSizeOfMdl, vaddr, len) > MDL_ZONE_SIZE){
		//cmn_err(CE_CONT,"%s: >MDL_SIZE, input len %d, alloc len %d\n",
		//	__func__,len, MmSizeOfMdl(vaddr,len));
		m = TxExAllocatePoolWithTag(NonPagedPool,
		    MSCALL2(MmSizeOfMdl, vaddr, len), 0);
	}else {
		//cmn_err(CE_CONT,"%s: <MDL_SIZE %d, input len %d, alloc len %d\n",
		//	__func__,MDL_ZONE_SIZE, len, MmSizeOfMdl(vaddr,len));
		m = kmem_cache_alloc(tx_mdl_cache, KM_NOSLEEP);
		//m = TxExAllocatePoolWithTag(NonPagedPool,
		   // MmSizeOfMdl(vaddr, len), 0);
		zone++;
	}

	if (m == NULL)
		return (NULL);
	//cmn_err(CE_CONT,"%s: correct2\n",__func__);
	MmInitializeMdl(m, vaddr, len);

	/*
	 * MmInitializMdl() clears the flags field, so we
	 * have to set this here. If the MDL came from the
	 * MDL UMA zone, tag it so we can release it to
	 * the right place later.
	 */
	if (zone)
		m->mdl_flags = MDL_ZONE_ALLOCED;

	if (iopkt != NULL) {
		if (secondarybuf == TRUE) {
			mdl			*last;
			last = iopkt->irp_mdl;
			while (last->mdl_next != NULL)
				last = last->mdl_next;
			last->mdl_next = m;
		} else {
			if (iopkt->irp_mdl != NULL)
				panic("leaking an MDL in IoAllocateMdl()");
			iopkt->irp_mdl = m;
		}
	}

	return (m);
}

void TxFreeMdl(m)
	mdl	*m;
{

	__FUNCNAME__;
	if (m == NULL)
		return;

	if (m->mdl_flags & MDL_ZONE_ALLOCED){
		//cmn_err(CE_CONT,"%s:",__func__);
		kmem_cache_free(tx_mdl_cache, m);
		//TxExFreePool(m);
	}else
		TxExFreePool(m);

    return;
}

#endif /* Unused... */

MSENTRY1(void IoFreeMdl, m)
	mdl	*m;
{

	__FUNCNAME__;
	if (m == NULL) {
		RESTORE();
		return;
	}

	if (m->mdl_flags & MDL_ZONE_ALLOCED){
		//cmn_err(CE_CONT,"%s:",__func__);
		//kmem_cache_free(mdl_cache, m);
		MSCALL1(ExFreePool, m);
	}else
		MSCALL1(ExFreePool, m);

	RESTORE();
	return;
}

MSENTRY2(uint32_t MmSizeOfMdl, vaddr, len)
	void			*vaddr;
	size_t			len;
{
	uint32_t		l;
	__FUNCNAME__;
        l = sizeof(struct mdl) +
	    (sizeof(vm_offset_t *) * SPAN_PAGES(vaddr, len));

	RESTORE();
	return(l);
}

/*
 * The Microsoft documentation says this routine fills in the
 * page array of an MDL with the _physical_ page addresses that
 * comprise the buffer, but we don't really want to do that here.
 * Instead, we just fill in the page array with the kernel virtual
 * addresses of the buffers.
 */
MSENTRY1(void MmBuildMdlForNonPagedPool, m)
	mdl			*m;
{
	vm_offset_t		*mdl_pages;
	uint32_t		pagecnt, i;

	__FUNCNAME__;
	pagecnt = SPAN_PAGES(m->mdl_byteoffset, m->mdl_bytecount);

	if (pagecnt > (m->mdl_size - sizeof(mdl)) / sizeof(vm_offset_t *))
		panic("not enough pages in MDL to describe buffer");

	mdl_pages = MmGetMdlPfnArray(m);

	for (i = 0; i < pagecnt; i++)
		*mdl_pages = (vm_offset_t)m->mdl_startva + (i * PAGE_SIZE);

	m->mdl_flags |= MDL_SOURCE_IS_NONPAGED_POOL;
	m->mdl_mappedsystemva = MmGetMdlVirtualAddress(m);

	RESTORE();
	return;
}

MSENTRY2(void *MmMapLockedPages, buf, accessmode)
	mdl			*buf;
	uint8_t			accessmode;
{
	void *ret;

	__FUNCNAME__;
	buf->mdl_flags |= MDL_MAPPED_TO_SYSTEM_VA;
	ret = MmGetMdlVirtualAddress(buf);
	RESTORE();
	return (ret);
}

MSENTRY5(void *MmMapLockedPagesSpecifyCache, buf, accessmode, cachetype, vaddr,
    bugcheck, prio)
	mdl			*buf;
	uint8_t			accessmode;
	uint32_t		cachetype;
	void			*vaddr;
	uint32_t		bugcheck;
	uint32_t		prio;
{
	void *res;

	__FUNCNAME__;
	res = (void*)MSCALL2(MmMapLockedPages, buf, accessmode);
	RESTORE();
	return (res);
}

MSENTRY2(void MmUnmapLockedPages, vaddr, buf)
	void			*vaddr;
	mdl			*buf;
{
	__FUNCNAME__;
	buf->mdl_flags &= ~MDL_MAPPED_TO_SYSTEM_VA;
	RESTORE();
	return;
}

MSENTRY3 (size_t RtlCompareMemory, s1, s2, len)
	const void		*s1;
	const void		*s2;
	size_t			len;
{
	size_t ret;
	const uchar_t *ps1 = s1;
	const uchar_t *ps2 = s2;

	__FUNCNAME__;
	if (s1 != s2 && len != 0) {
		do {
			if (*ps1++ != *ps2++) {
				ret = ps1[-1] - ps2[-1];
				RESTORE();
				return (ret);
				}
		} while (--len != 0);
	}

	RESTORE();
	return (0);
}

MSENTRY2(void RtlInitAnsiString, dst, src)
	ndis_ansi_string	*dst;
	char			*src;
{
	ndis_ansi_string	*a;

	__FUNCNAME__;
	a = dst;
	if (a == NULL) {
		RESTORE();
		return;
	}
	if (src == NULL) {
		a->nas_len = a->nas_maxlen = 0;
		a->nas_buf = NULL;
	} else {
		a->nas_buf = src;
		a->nas_len = a->nas_maxlen = strlen(src);
	}

	RESTORE();
	return;
}

#if defined(__amd64__) & defined(LINKEDIMAGE)
MSENTRY2(void RtlInitString, dst, src)
	ndis_ansi_string	*dst;
	char			*src;
{
	MSCALL2(RtlInitAnsiString, dst, src);
	RESTORE();
	return;
}
#endif

MSENTRY2(void RtlInitUnicodeString, dst, src)
	ndis_unicode_string	*dst;
	uint16_t		*src;
{
	ndis_unicode_string	*ux;
	int			i;

	__FUNCNAME__;
	ux = dst;
	if (ux == NULL) {
		RESTORE();
		return;
	}
	if (src == NULL) {
		ux->us_len = ux->us_maxlen = 0;
		ux->us_buf = NULL;
	} else {
		i = 0;
		while(src[i] != 0)
			i++;
		ux->us_buf = src;
		ux->us_len = ux->us_maxlen = i * 2;
	}

	RESTORE();
	return;
}

MSENTRY3(ndis_status RtlUnicodeStringToInteger, ustr, base, val)
	ndis_unicode_string	*ustr;
	uint32_t		base;
	uint32_t		*val;
{
	uint16_t		*uchr;
	int			len, neg = 0;
	char			abuf[64];
	char			*astr;

	__FUNCNAME__;
	uchr = ustr->us_buf;
	len = ustr->us_len;
	bzero(abuf, sizeof(abuf));

	if ((char)((*uchr) & 0xFF) == '-') {
		neg = 1;
		uchr++;
		len -= 2;
	} else if ((char)((*uchr) & 0xFF) == '+') {
		neg = 0;
		uchr++;
		len -= 2;
	}

	if (base == 0) {
		if ((char)((*uchr) & 0xFF) == 'b') {
			base = 2;
			uchr++;
			len -= 2;
		} else if ((char)((*uchr) & 0xFF) == 'o') {
			base = 8;
			uchr++;
			len -= 2;
		} else if ((char)((*uchr) & 0xFF) == 'x') {
			base = 16;
			uchr++;
			len -= 2;
		} else
			base = 10;
	}

	astr = abuf;
	if (neg) {
		strcpy(astr, "-");
		astr++;
	}

	ndis_unicode_to_ascii(uchr, len, &astr);
	ddi_strtol(abuf, NULL, base,(long *)val);

	RESTORE();
	return(NDIS_STATUS_SUCCESS);
}

MSENTRY1(void RtlFreeUnicodeString, ustr)
	ndis_unicode_string	*ustr;
{
	__FUNCNAME__;
	if (ustr->us_buf == NULL) {
		RESTORE();
		return;
	}
	kmem_free(ustr->us_buf, ustr->us_len);
	ustr->us_buf = NULL;
	RESTORE();
	return;
}

MSENTRY1(void RtlFreeAnsiString, astr)
	ndis_ansi_string	*astr;
{
	__FUNCNAME__;
	if (astr->nas_buf == NULL) {
		RESTORE();
		return;
	}
	kmem_free(astr->nas_buf, astr->nas_len);
	astr->nas_buf = NULL;
	RESTORE();
	return;
}

static int
atoi(str)
	const char		*str;
{
	long val;
	__FUNCNAME__;
	ddi_strtol(str,(char **)NULL, 10,&val);
	return((int)val);
}

static long
atol(str)
	const char		*str;
{
	long val;
	__FUNCNAME__;
	ddi_strtol(str,(char **)NULL, 10,&val);
	return ((long)val);
}
static uint_t rand_seed = 1;

int
rand_r(uint_t *seed)
{
	int res;
	__FUNCNAME__;
	res = ((*seed = *seed * 1103515245 + 12345) >> 16) & 0x7fff;
	return (res);
}

static void
srand(uint_t seed)
{
	__FUNCNAME__;
	rand_seed = seed;
}

static int
rand(void)
{
	__FUNCNAME__;
	return (rand_r(&rand_seed));
}

MSENTRY2(uint8_t IoIsWdmVersionAvailable, major, minor)
	uint8_t			major;
	uint8_t			minor;
{
	__FUNCNAME__;
	if (major == WDM_MAJOR && minor == WDM_MINOR_WINXP) {
		RESTORE();
		return(TRUE);
	}
	RESTORE();
	return(FALSE);
}

MSENTRY5(ndis_status IoGetDeviceProperty, devobj, regprop, buflen, prop, reslen)
	device_object		*devobj;
	uint32_t		regprop;
	uint32_t		buflen;
	void			*prop;
	uint32_t		*reslen;
{
	driver_object		*drv;
	uint16_t		**name;

	drv = devobj->do_drvobj;

	__FUNCNAME__;
	switch (regprop) {
	case DEVPROP_DRIVER_KEYNAME:
		name = prop;
		*name = drv->dro_drivername.us_buf;
		*reslen = drv->dro_drivername.us_len;
		break;
	default:
		RESTORE();
		return(STATUS_INVALID_PARAMETER_2);
		break;
	}

	RESTORE();
	return(STATUS_SUCCESS);
}

MSENTRY2(void KeInitializeMutex, kmutex, level)
	kmutant			*kmutex;
	uint32_t		level;
{
	__FUNCNAME__;

	INIT_LIST_HEAD((&kmutex->km_header.dh_waitlisthead));
	kmutex->km_abandoned = FALSE;
	kmutex->km_apcdisable = 1;
	kmutex->km_header.dh_sigstate = TRUE;
	kmutex->km_header.dh_type = EVENT_TYPE_SYNC;
	kmutex->km_header.dh_size = OTYPE_MUTEX;
	kmutex->km_acquirecnt = 0;
	kmutex->km_ownerthread = NULL;
	RESTORE();
	return;
}

MSENTRY2(uint32_t KeReleaseMutex,kmutex, kwait)
	kmutant			*kmutex;
	uint8_t			kwait;
{
	uint32_t ret;

	__FUNCNAME__;
	mutex_enter(&ntoskrnl_dispatchlock);
	if (kmutex->km_ownerthread != curthread) {
		mutex_exit(&ntoskrnl_dispatchlock);
		RESTORE();
		return(STATUS_MUTANT_NOT_OWNED);
	}
	kmutex->km_acquirecnt--;
	if (kmutex->km_acquirecnt == 0) {
		kmutex->km_ownerthread = NULL;
		ntoskrnl_wakeup(&kmutex->km_header);
	}
	mutex_exit(&ntoskrnl_dispatchlock);

	ret = kmutex->km_acquirecnt;
	RESTORE();
	return (ret);
}

MSENTRY1(uint32_t KeReadStateMutex, kmutex)
	kmutant			*kmutex;
{
	uint32_t ret;

	__FUNCNAME__;
	ret = kmutex->km_header.dh_sigstate;
	RESTORE();
	return (ret);
}

MSENTRY3(void KeInitializeEvent, kevent, type, state)
	nt_kevent		*kevent;
	uint32_t		type;
	uint8_t			state;
{
	__FUNCNAME__;

	INIT_LIST_HEAD((&kevent->k_header.dh_waitlisthead));
	kevent->k_header.dh_sigstate = state;
	kevent->k_header.dh_type = type;
	kevent->k_header.dh_size = OTYPE_EVENT;
	RESTORE();
	return;
}

MSENTRY1(uint32_t KeResetEvent, kevent)
	nt_kevent		*kevent;
{
	uint32_t		prevstate;

	__FUNCNAME__;
	mutex_enter(&ntoskrnl_dispatchlock);
	prevstate = kevent->k_header.dh_sigstate;
	kevent->k_header.dh_sigstate = FALSE;
	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return(prevstate);
}

MSENTRY3(uint32_t KeSetEvent, kevent, increment, kwait)
	nt_kevent		*kevent;
	uint32_t		increment;
	uint8_t			kwait;
{
	uint32_t		prevstate;
	__FUNCNAME__;
	mutex_enter(&ntoskrnl_dispatchlock);
	prevstate = kevent->k_header.dh_sigstate;
	ntoskrnl_wakeup(&kevent->k_header);
	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return(prevstate);
}

MSENTRY1(void KeClearEvent, kevent)
	nt_kevent		*kevent;
{
	__FUNCNAME__;
	kevent->k_header.dh_sigstate = FALSE;
	RESTORE();
	return;
}

MSENTRY1(uint32_t KeReadStateEvent, kevent)
	nt_kevent		*kevent;
{
	uint32_t state;

	__FUNCNAME__;
	state = kevent->k_header.dh_sigstate;
	RESTORE();
	return (state);
}

MSENTRY5(ndis_status ObReferenceObjectByHandle, handle, reqaccess, otype,
    accessmode, object, handleinfo)
	ndis_handle		handle;
	uint32_t		reqaccess;
	void			*otype;
	uint8_t			accessmode;
	void			**object;
	void			**handleinfo;
{
	__FUNCNAME__;
	RESTORE();
	return(NDIS_STATUS_SUCCESS);
}

#if defined(__amd64__) & defined(LINKEDIMAGE)
MSENTRY1(void ObfDereferenceObject, object)
	void *object;
#else
__fastcall static void
ObfDereferenceObject(REGARGS1(void *object))
#endif
{
	__FUNCNAME__;
	RESTORE();
	return;
}

MSENTRY5(uint32_t ZwCreateFile, handle, access, oattributes, status,
		allocsize, fattributes, share, disposition, options,
		eabuffer, ealength)
	void **handle;
	uint32_t access;
	void *oattributes;
	io_status_block *status;
	long long *allocsize;
	uint32_t fattributes;
	uint32_t share;
	uint32_t disposition;
	uint32_t options;
	void *eabuffer;
	uint32_t ealength;
{
	__FUNCNAME__;
	if (status)
		status->isb_status = STATUS_FILE_NOT_FOUND;
	RESTORE();
	return (STATUS_FILE_NOT_FOUND);
}

MSENTRY1(uint32_t ZwClose, handle)
	ndis_handle		handle;
{
	__FUNCNAME__;
	RESTORE();
	return(STATUS_SUCCESS);
}

MSENTRY5(uint32_t ZwReadFile, handle, event, routine, context, status,
		buffer, length, offset, key)
	void *handle;
	void *event;
	uint32_t (*routine)();
	void *context;
	io_status_block *status;
	void *buffer;
	uint32_t length;
	long long *offset;
	uint32_t *key;
{
	__FUNCNAME__;
	if (status)
		status->isb_status = STATUS_END_OF_FILE;
	RESTORE();
	return (STATUS_END_OF_FILE);
}

MSENTRY5(uint32_t ZwQueryInformationFile, handle, status, information,
		length, class)
	void *handle;
	io_status_block *status;
	void *information;
	uint32_t length;
	void *class;
{
	__FUNCNAME__;
	if (information && (length >= 4))
		bzero(information,4);
	status->isb_status = STATUS_SUCCESS;
	status->isb_info = (register_t)4;
	RESTORE();
	return (STATUS_SUCCESS);
}

MSENTRY1(uint32_t IoUnregisterPlugPlayNotification, notification)
	void *notification;
{
	__FUNCNAME__;
 	RESTORE();
	return (STATUS_SUCCESS);
}
 
/*
 * This is here just in case the thread returns without calling
 * PsTerminateSystemThread().
 */
static void
ntoskrnl_thrfunc(arg)
	void			*arg;
{
	__FUNCNAME__;
/*
	thread_context		*thrctx;
	__stdcall uint32_t (*tfunc)(void *);
	void			*tctx;
	uint32_t		rval;

	thrctx = arg;
	tfunc = thrctx->tc_thrfunc;
	tctx = thrctx->tc_thrctx;
	free(thrctx, M_TEMP);

	rval = MSCALL1(tfunc, tctx);

	PsTerminateSystemThread(rval);
*/
	return; /* notreached */
}

MSENTRY5(ndis_status PsCreateSystemThread, handle, reqaccess, objattrs, phandle,
	clientid, thrfunc, thrctx)
	ndis_handle		*handle;
	uint32_t		reqaccess;
	void			*objattrs;
	ndis_handle		phandle;
	void			*clientid;
	void			*thrfunc;
	void			*thrctx;
{
	int			error;
/*	char			tname[128];
	thread_context		*tc;
	struct proc		*p;

	tc = malloc(sizeof(thread_context), M_TEMP, M_NOWAIT);
	if (tc == NULL) {
		RESTORE();
		return(NDIS_STATUS_FAILURE);
	}

	tc->tc_thrctx = thrctx;
	tc->tc_thrfunc = thrfunc;

	sprintf(tname, "windows kthread %d", ntoskrnl_kth);
	error = kthread_create(ntoskrnl_thrfunc, tc, &p,
	    RFHIGHPID, NDIS_KSTACK_PAGES, tname);
	*handle = p;

	ntoskrnl_kth++;
*/
	__FUNCNAME__;

	RESTORE();
	return(error);
}

/*
 * In Windows, the exit of a thread is an event that you're allowed
 * to wait on, assuming you've obtained a reference to the thread using
 * ObReferenceObjectByHandle(). Unfortunately, the only way we can
 * simulate this behavior is to register each thread we create in a
 * reference list, and if someone holds a reference to us, we poke
 * them.
 */
MSENTRY1(ndis_status PsTerminateSystemThread, status)
	ndis_status		status;
{
/*
	struct nt_objref	*nr;

	TAILQ_FOREACH(nr, &ntoskrnl_reflist, link) {
		if (nr->no_obj != curthread->td_proc)
			continue;
		ntoskrnl_wakeup(&nr->no_dh);
		break;
	}

	ntoskrnl_kth--;
	kthread_exit(0);
*/
	__FUNCNAME__;
	RESTORE();
	return(0);	/* notreached */
}

#if defined(__amd64__) & defined(LINKEDIMAGE)

MSENTRY5(uint32_t DbgPrint, fmt, a1, a2, a3, a4, a5)
char *fmt;
uint64_t a1, a2, a3, a4, a5;
{
	__FUNCNAME__;
	cmn_err(CE_WARN, fmt, a1, a2, a3, a4, a5);
	RESTORE();
	return(STATUS_SUCCESS);
}

#else

static uint32_t
DbgPrint(char *fmt, ...)
{
	__FUNCNAME__;
	return(STATUS_SUCCESS);
}

#endif

MSENTRY0(void DbgBreakPoint)
{
	__FUNCNAME__;
	RESTORE();
	return;
}

MSENTRY5(void KeBugCheckEx, code, param1, param2, param3, param4)
    uint32_t			code;
    u_long			param1;
    u_long			param2;
    u_long			param3;
    u_long			param4;
{
	panic("KeBugCheckEx: STOP 0x%X", code);
	RESTORE();
}

/* Only called as defined by timeout() according to Unix rules */
static void
ntoskrnl_timercall(arg)
	void			*arg;
{
	ktimer			*timer;
	uint64_t		tv_usec;
	
	__FUNCNAME__;
	timer = arg;
	mutex_enter(&ntoskrnl_dispatchlock);
	timer->k_header.dh_inserted = FALSE;

	/*
	 * If this is a periodic timer, re-arm it
	 * so it will fire again. We do this before
	 * calling any deferred procedure calls because
	 * it's possible the DPC might cancel the timer,
	 * in which case it would be wrong for us to
	 * re-arm it again afterwards.
	 */

	if (timer->k_period)
		timer->k_header.dh_inserted = 2;
	/*
	 * Not really sure whether or not lock can be held across InsertQueue.
	 * However, there was a bug where the lock wasn't being held across wakeup,
	 * so it's been acquired for this routine.  It doesn't seem to cause any
	 * deadlocks.
	 */
	if (timer->k_dpc != NULL)
		MSCALL3(KeInsertQueueDpc, (kdpc *)timer->k_dpc, NULL, NULL);

	ntoskrnl_wakeup(&timer->k_header);
	/*
	 * if dh_inserted is still 2, then reschedule, otherwise, the
	 * timer has been canceled
	 */
	mutex_exit(&ntoskrnl_dispatchlock);
	if(timer->k_header.dh_inserted == 2) {
		mutex_enter(&ntoskrnl_dispatchlock);
		tv_usec = timer->k_period * 1000;
		timer->k_handle = timeout(ntoskrnl_timercall, timer, drv_usectohz(tv_usec));
		timer->k_header.dh_inserted = TRUE;
		mutex_exit(&ntoskrnl_dispatchlock);
	}

	return;
}

MSENTRY1(void KeInitializeTimer, timer)
	ktimer			*timer;
{
	__FUNCNAME__;
	if (timer == NULL) {
		RESTORE();
		return;
	}
	MSCALL2(KeInitializeTimerEx, timer,  EVENT_TYPE_NOTIFY);
	RESTORE();
	return;
}

MSENTRY2(void KeInitializeTimerEx, timer, type)
	ktimer			*timer;
	uint32_t		type;
{
	__FUNCNAME__;
	if (timer == NULL) {
		RESTORE();
		return;
	}
	INIT_LIST_HEAD((&timer->k_header.dh_waitlisthead));
	timer->k_header.dh_sigstate = FALSE;
	timer->k_header.dh_inserted = FALSE;
	timer->k_header.dh_type = type;
	timer->k_header.dh_size = OTYPE_TIMER;
	timer->k_handle = NULL;
	RESTORE();
	return;
}

/*
 * This is a wrapper for Windows deferred procedure calls that
 * have been placed on an NDIS thread work queue. We need it
 * since the DPC could be a _stdcall function. Also, as far as
 * I can tell, defered procedure calls must run at DISPATCH_LEVEL.
 */
static void
ntoskrnl_run_dpc(arg)
	void			*arg;
{
	dpc_timer_function	dpcfunc;
	kdpc			*dpc;
	uint8_t			irql;

	dpc = arg;
	__FUNCNAME__;
	dpcfunc = dpc->k_deferedfunc;
	irql = KmRaiseIrql(DISPATCH_LEVEL);
          /* JPA can only call functions from the Windows driver */
	MSCALL4(dpcfunc, dpc, dpc->k_deferredctx, dpc->k_sysarg1, dpc->k_sysarg2);
	KmLowerIrql(irql);

	return;
}

MSENTRY3(void KeInitializeDpc, dpc, dpcfunc, dpcctx)
	kdpc			*dpc;
	dpc_timer_function	dpcfunc;
	void			*dpcctx;
{

	if (dpc == NULL) {
		RESTORE();
		return;
	}

	__FUNCNAME__;

	dpc->k_deferedfunc = dpcfunc;
	dpc->k_deferredctx = dpcctx;

	RESTORE();
	return;
}

MSENTRY3(uint8_t KeInsertQueueDpc, dpc, sysarg1, sysarg2)
	kdpc			*dpc;
	void			*sysarg1;
	void			*sysarg2;
{
	
	__FUNCNAME__;
	dpc->k_sysarg1 = sysarg1;
	dpc->k_sysarg2 = sysarg2;

	if (ndis_sched(ntoskrnl_run_dpc, dpc, NDIS_SWI)) {
		RESTORE();
		return(FALSE);
	}
	RESTORE();
	return(TRUE);
}

MSENTRY1(uint8_t KeRemoveQueueDpc, dpc)
	kdpc			*dpc;
{
	if (ndis_unsched(ntoskrnl_run_dpc, dpc, NDIS_SWI)) {
		RESTORE();
		return(FALSE);
	}
	__FUNCNAME__;
	RESTORE();
	return(TRUE);
}
/*
 * duetime is in 100 ns increments, period is in milliseconds
 */
MSENTRY4(uint8_t KeSetTimerEx, timer, duetime, period, dpc)
	ktimer			*timer;
	int64_t			duetime;
	uint32_t		period;
	kdpc			*dpc;
{
	uint64_t		curtime;
	uint8_t			pending;
	uint64_t		tv_usec;

	__FUNCNAME__;
	if (timer == NULL) {
		RESTORE();
		return(FALSE);
	}

	mutex_enter(&ntoskrnl_dispatchlock);
	/*
	 * Note that the lock can not be held during untimeout since if the
	 * handler is already called, this will wait until the caller finishes
	 * and since the handler grabs the lock, this can cause a deadlock.
	 */
	if (timer->k_header.dh_inserted == TRUE) {
		timer->k_header.dh_inserted = FALSE;
		mutex_exit(&ntoskrnl_dispatchlock);
		untimeout(timer->k_handle);
		mutex_enter(&ntoskrnl_dispatchlock);
		pending = TRUE;
	} else
		pending = FALSE;

	timer->k_duetime = duetime;
	timer->k_period = period;
	timer->k_header.dh_sigstate = FALSE;
	timer->k_dpc = dpc;

	if (duetime != 0) {
		if ( duetime < 0 ) {
		        tv_usec = -duetime / 10;
	        } else {
                        ntoskrnl_time(&curtime);
                        if (duetime < (int64_t)curtime)
				tv_usec = 0;
			else
				tv_usec = (duetime - curtime) / 10;
                }
	} else {
/*
 * make sure tv is initialized reasonably (why would duetime be 0 anyway?)
 */
	    tv_usec = 0;
	}
/*
 * timer should do this anyway ...
 */
	if(tv_usec == 0)
	    tv_usec = 10000;	    /* one tic at 100 Hz */
	timer->k_handle = timeout(ntoskrnl_timercall, timer, drv_usectohz(tv_usec));
	timer->k_header.dh_inserted = TRUE;
	mutex_exit(&ntoskrnl_dispatchlock);
	RESTORE();
	return(pending);
}

MSENTRY3(uint8_t KeSetTimer, timer, duetime, dpc)
	ktimer			*timer;
	int64_t			duetime;
	kdpc			*dpc;
{
	uint8_t t;

	__FUNCNAME__;
	t = MSCALL4(KeSetTimerEx, timer, duetime, 0, dpc);
	RESTORE();
	return (t);
}

MSENTRY1(uint8_t KeCancelTimer, timer)
	ktimer			*timer;
{
	uint8_t			pending;
	
	__FUNCNAME__;
	if (timer == NULL) {
		RESTORE();
		return(FALSE);
	}

	mutex_enter(&ntoskrnl_dispatchlock);
	/*
	 * Note that the lock can not be held during untimeout since if the
	 * handler is already called, this will wait until the caller finishes
	 * and since the handler grabs the lock, this can cause a deadlock.
	 */
	if (timer->k_header.dh_inserted == TRUE) {
		timer->k_header.dh_inserted = FALSE;
		mutex_exit(&ntoskrnl_dispatchlock);
		untimeout(timer->k_handle);
		mutex_enter(&ntoskrnl_dispatchlock);
		pending = TRUE;
	} else
		pending = FALSE;
	
	//is there an issue related to following two lines?
	if (pending && (timer->k_dpc != NULL))
		ndis_unsched(ntoskrnl_run_dpc, timer->k_dpc, NDIS_SWI);

	mutex_exit(&ntoskrnl_dispatchlock);

	RESTORE();
	return(pending);
}

MSENTRY1(uint8_t KeReadStateTimer, timer)
	ktimer			*timer;
{
	uint8_t t;

	__FUNCNAME__;
	t = timer->k_header.dh_sigstate;
	RESTORE();
	return(t);
}

MSENTRY1(void KeQuerySystemTime, current_time)
	uint64_t		*current_time;
{
	ntoskrnl_time(current_time);
	RESTORE();
}

MSENTRY0(uint32_t KeTickCount)
{
	clock_t tv;
	uint32_t t;

	tv = ddi_get_lbolt();
	t = drv_usectohz(tv);
	RESTORE();
	return (t);
}

char *
ndis_rindex(p, ch)
	const char *p;
	int ch;
{
	union {
		const char *cp;
		char *p;
	} utb;
	char *save;

	//cmn_err(CE_CONT,"%s: p= [%s], ch =[%c]\n",__func__,p, ch);
	utb.cp = p;
	for (save = NULL;; ++utb.p) {
		if (*utb.p == ch)
			save = utb.p;
		if (*utb.p == '\0'){
	//		cmn_err(CE_CONT,"%s::: %s\n",__func__, save);
			return(save);
		}
	}
	/* NOTREACHED */
}

/*
 * Describes the chip's DMA engine
 */
static ddi_dma_attr_t ntos_dma_attr = {
	DMA_ATTR_V0,			/* dma_attr version */
	0x0000000000000000ull,		/* dma_attr_addr_lo */
	0x3ffFFFFF,			/* dma_attr_addr_hi */
	0x00000000FFFFFFFFull,		/* dma_attr_count_max */
	0x0000000000000001ull,		/* dma_attr_align */
	0x00000FFF,			/* dma_attr_burstsizes */
	0x00000001,			/* dma_attr_minxfer */
	0x000000000000FFFFull,		/* dma_attr_maxxfer */
	0xFFFFFFFFFFFFFFFFull,		/* dma_attr_seg */
	1,				/* dma_attr_sgllen */
	0x00000001,			/* dma_attr_granular */
	0				/* dma_attr_flags */
};

ddi_device_acc_attr_t ndis_cache_accattr= {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};

/*
 * This maps to bus_dmamem_alloc().
 */
static void
NtoskernelPageCache(len, vaddr)
	uint32_t		len;
	void			**vaddr;
{

	struct ndis_softc	*sc;
	struct ndis_shmem	*sh;
	int			error;
	dev_info_t 	*dip;

	__FUNCNAME__;

	sc = curSC;
	dip = sc->id.id_dev;
	
	sh = kmem_zalloc(sizeof(struct ndis_shmem), KM_NOSLEEP);
	if (sh == NULL)
		return;

	sh->bf_dma.size = len;
	cmn_err(CE_CONT,"%s: len = %d\n",__func__, len);
	sh->accattr = &ntos_dma_attr;

	error = ddi_dma_alloc_handle(dip, sh->accattr,
		    DDI_DMA_SLEEP, NULL, &sh->bf_dma.dma_hdl);

	if (error != DDI_SUCCESS){
		cmn_err(CE_WARN,"ndis get no dma_hdl\n");
		return ;
	}
	
	error = ddi_dma_mem_alloc(sh->bf_dma.dma_hdl,
	    sh->bf_dma.size,
	    &ndis_cache_accattr,
	    DDI_DMA_STREAMING,
	    DDI_DMA_SLEEP, NULL,
	    &sh->bf_dma.mem_va,
	    &sh->bf_dma.alength,
	    &sh->bf_dma.acc_hdl);

	if (error != DDI_SUCCESS){
		cmn_err(CE_WARN,"ndis mem alloc failed\n");
		return ;
	}
	
	error = ddi_dma_addr_bind_handle(sh->bf_dma.dma_hdl,
	    NULL,
	    sh->bf_dma.mem_va,
	    sh->bf_dma.alength,
	    DDI_DMA_RDWR | DDI_DMA_STREAMING,
	    DDI_DMA_SLEEP,
	    NULL,
	    &sh->bf_dma.cookie,
	    &sh->bf_dma.ncookies);

	if (error != DDI_SUCCESS){
		cmn_err(CE_WARN,"ndis bind addr failed\n");
		return;
	}
	
	*vaddr = sh->bf_dma.mem_va;
	
	sh->ndis_saddr = *vaddr;
	sh->shm_len = sh->bf_dma.alength;
	sh->phyaddr.np_quad = sh->bf_dma.cookie.dmac_laddress;
	sh->ndis_next = sc->ndis_shlist;
	sc->ndis_shlist = sh;
	return;
}

static __inline vm_paddr_t
pmap_kextract(vm_offset_t va)
{
	struct ndis_shmem	*prev, *sh = NULL;
	int	found=0;
	struct ndis_softc	*sc;

	mutex_enter(&register_mtx);
	sc = curSC;
	sh = sc->ndis_shlist;
	while (sh) {
		prev = sh;
		if ((uintptr_t *)sh->ndis_saddr == (uintptr_t *)va){
			cmn_err(CE_CONT,"Hit 0x%x\n",(uint32_t)va);
			found = 0x1;
			break;
		}
		sh = sh->ndis_next;
	}

	if(!found){
		cmn_err(CE_CONT,"Yeah? nothing found\n");
		mutex_exit(&register_mtx);
		return (NULL);
	}

	cmn_err(CE_CONT,"%s:return 0x%x,0x%x\n",__func__,
		(uint32_t)( (sh->bf_dma.cookie.dmac_laddress >> 32) & 0xffffffff),
		(uint32_t)( (sh->bf_dma.cookie.dmac_laddress) & 0xffffffff));
	mutex_exit(&register_mtx);
	return (sh->bf_dma.cookie.dmac_laddress);
}

#ifdef LINKEDIMAGE
MSENTRY1(vm_paddr_t MmGetPhysicalAddress, va)
	vm_offset_t va;
{
	vm_paddr_t pa;
	pa = pmap_kextract(va);
	RESTORE();
	return (pa);
}
#endif

MSENTRY5(void *MmAllocateContiguousMemorySpecifyCache, size, lowest, highest,
    boundary, cachetype)
	uint32_t		size;
	uint64_t		lowest;
	uint64_t		highest;
	uint64_t		boundary;
	uint32_t		cachetype;

#define	page_roundup(x, y)	((((x)+(y)-1)/(y))*(y))
{
	void *addr=NULL;
	uint32_t pagelength = page_roundup(size, PAGE_SIZE);
	cmn_err(CE_CONT,"%s: pagelength = %d, PAGE_SIZE = %d\n",
			__func__, pagelength, PAGE_SIZE);
	mutex_enter(&register_mtx);
	NtoskernelPageCache(pagelength,&addr);
	mutex_exit(&register_mtx);
	RESTORE();
	return(addr);
}

MSENTRY3(void MmFreeContiguousMemorySpecifyCache, base, size, cachetype)
	void			*base;
	uint32_t		size;
	uint32_t		cachetype;
{
	struct ndis_shmem	*prev, *sh = NULL;
	int	found=0;
	struct ndis_softc	*sc=curSC;

	mutex_enter(&register_mtx);
	sh = sc->ndis_shlist;
	while (sh) {
		prev = sh;
		if ((uintptr_t *)sh->ndis_saddr == (uintptr_t *)base){
			found = 0x1;
			break;
		}
		sh = sh->ndis_next;
	}
	
	if(!found){
		cmn_err(CE_CONT,"Yeah? nothing found\n");
		mutex_exit(&register_mtx);
		RESTORE();
		return;
	}

	if (sh->bf_dma.dma_hdl != NULL) {
		ddi_dma_unbind_handle(sh->bf_dma.dma_hdl);
		ddi_dma_mem_free(&sh->bf_dma.acc_hdl);
		ddi_dma_free_handle(&sh->bf_dma.dma_hdl);
		sh->bf_dma.dma_hdl = NULL;
	}

	if (sh == sc->ndis_shlist)
		sc->ndis_shlist = sh->ndis_next;
	else
		prev->ndis_next = sh->ndis_next;

	kmem_free(sh, sizeof(struct ndis_shmem));

	mutex_exit(&register_mtx);
	RESTORE();
	return;
}

MSENTRY3(void *MmMapIoSpace, physical, size, type)
	void *physical;
	unsigned int size;
	int type;
{
	cmn_err(CE_WARN,"** MmMapIoSpace is not implemented\n");
	cmn_err(CE_WARN,"physical 0x%lx size 0x%x type 0x%x\n",
			(long)physical,(int)size,(int)type);
	RESTORE();
	return ((void*)NULL);
}

MSENTRY2(void MmUnmapIoSpace, virtual, size)
	void *virtual;
	unsigned int size;
{
	cmn_err(CE_WARN,"** MmUnmapIoSpace is not implemented\n");
	cmn_err(CE_WARN,"virtual 0x%lx size 0x%x\n",
			(long)virtual,(int)size);
	RESTORE();
}

MSENTRY3(uint32_t IoWMIOpenBlock, guid, access, object)
	void *guid;
	uint32_t access;
	void *object;
{
	cmn_err(CE_WARN,"** IoWMIOpenBlock is not implemented\n");
	cmn_err(CE_WARN,"guid 0x%lx access 0x%lx object 0x%lx\n",
			(long)guid, (long)access, (long)object);
	RESTORE();
	return (STATUS_WMI_GUID_NOT_FOUND);
}

MSENTRY3(uint32_t IoWMIQueryAllData, block, size, buffer)
	void *block;
	uint32_t *size;
	void *buffer;
{
	cmn_err(CE_WARN,"** IoWMIQueryAllData is not implemented\n");
	cmn_err(CE_WARN,"block 0x%lx size 0x%lx buffer 0x%lx\n",
			(long)block, (long)size, (long)buffer);
	RESTORE();
	return (STATUS_WMI_GUID_NOT_FOUND);
}

MSENTRY0(void dummy_ntoskrnl)
{
	cmn_err(CE_WARN,"ntoskrnl lost API\n");
	RESTORE();
	return;
}

image_patch_table ntoskrnl_functbl[] = {
	IMPORT_FUNC(RtlCompareMemory),
	IMPORT_FUNC(RtlEqualUnicodeString),
	IMPORT_FUNC(RtlCopyUnicodeString),
	IMPORT_FUNC(RtlUnicodeStringToAnsiString),
	IMPORT_FUNC(RtlAnsiStringToUnicodeString),
	IMPORT_FUNC(RtlInitAnsiString),
	IMPORT_FUNC(IoGetDeviceObjectPointer),
	IMPORT_FUNC_MAP(RtlInitString, RtlInitAnsiString),
	IMPORT_FUNC_MAP(strrchr, ndis_rindex),
	IMPORT_FUNC_MAP(MmGetPhysicalAddress, pmap_kextract),
	IMPORT_FUNC(MmAllocateContiguousMemorySpecifyCache),
	IMPORT_FUNC(MmFreeContiguousMemorySpecifyCache),
	IMPORT_FUNC(RtlInitUnicodeString),
	IMPORT_FUNC(RtlFreeAnsiString),
	IMPORT_FUNC(RtlFreeUnicodeString),
	IMPORT_FUNC(RtlUnicodeStringToInteger),
	IMPORT_FUNC_MAP(sprintf, win_sprintf),
	IMPORT_FUNC(vsprintf),
	IMPORT_FUNC_MAP(_snprintf, snprintf),
	IMPORT_FUNC_MAP(_vsnprintf, vsnprintf),
	IMPORT_FUNC(DbgPrint),
	IMPORT_FUNC(DbgBreakPoint),
	IMPORT_FUNC(KeBugCheckEx),
	IMPORT_FUNC(KeQuerySystemTime),
	IMPORT_FUNC(KeTickCount),
	IMPORT_FUNC(strncmp),
	IMPORT_FUNC(strcmp),
	IMPORT_FUNC(strncpy),
	IMPORT_FUNC(strcpy),
	IMPORT_FUNC(strlen),
	IMPORT_FUNC_MAP(RtlCopyMemory, win_memcpy),
	IMPORT_FUNC_MAP(RtlZeroMemory, win_bzero),	
	IMPORT_FUNC_MAP(memcpy, win_memcpy),
	IMPORT_FUNC_MAP(memmove, ntoskrnl_memset),
	IMPORT_FUNC_MAP(memset, ntoskrnl_memset),
	IMPORT_FUNC(IoAllocateDriverObjectExtension),
	IMPORT_FUNC(IoGetDriverObjectExtension),
	IMPORT_FUNC(IofCallDriver),
	IMPORT_FUNC(IofCompleteRequest),
	IMPORT_FUNC(IoCreateDevice),
	IMPORT_FUNC(IoDeleteDevice),
	IMPORT_FUNC(IoGetAttachedDevice),
	IMPORT_FUNC(IoAttachDeviceToDeviceStack),
	IMPORT_FUNC(IoDetachDevice),
	IMPORT_FUNC(IoBuildSynchronousFsdRequest),
	IMPORT_FUNC(IoBuildAsynchronousFsdRequest),
	IMPORT_FUNC(IoBuildDeviceIoControlRequest),
	IMPORT_FUNC(IoAllocateIrp),
	IMPORT_FUNC(IoReuseIrp),
	IMPORT_FUNC(IoMakeAssociatedIrp),
	IMPORT_FUNC(IoFreeIrp),
	IMPORT_FUNC(IoInitializeIrp),
	IMPORT_FUNC(KeWaitForSingleObject),
	IMPORT_FUNC(KeWaitForMultipleObjects),
	IMPORT_FUNC(_allmul),
	IMPORT_FUNC(_alldiv),
	IMPORT_FUNC(_allrem),
	IMPORT_FUNC(_allshr),
	IMPORT_FUNC(_allshl),
	IMPORT_FUNC(_aullmul),
	IMPORT_FUNC(_aulldiv),
	IMPORT_FUNC(_aullrem),
	IMPORT_FUNC(_aullshr),
	IMPORT_FUNC(_aullshl),
	IMPORT_FUNC(atoi),
	IMPORT_FUNC(atol),
	IMPORT_FUNC(rand),
	IMPORT_FUNC(srand),
	IMPORT_FUNC(WRITE_REGISTER_USHORT),
	IMPORT_FUNC(READ_REGISTER_USHORT),
	IMPORT_FUNC(WRITE_REGISTER_ULONG),
	IMPORT_FUNC(READ_REGISTER_ULONG),
	IMPORT_FUNC(READ_REGISTER_UCHAR),
	IMPORT_FUNC(WRITE_REGISTER_UCHAR),
	IMPORT_FUNC(ExInitializePagedLookasideList),
	IMPORT_FUNC(ExDeletePagedLookasideList),
	IMPORT_FUNC(ExInitializeNPagedLookasideList),
	IMPORT_FUNC(ExDeleteNPagedLookasideList),
	IMPORT_FUNC(InterlockedPopEntrySList),
	IMPORT_FUNC(InterlockedPushEntrySList),
	IMPORT_FUNC(ExQueryDepthSList),
	IMPORT_FUNC_MAP(ExpInterlockedPopEntrySList, InterlockedPopEntrySList),
	IMPORT_FUNC_MAP(ExpInterlockedPushEntrySList,
		InterlockedPushEntrySList),
	IMPORT_FUNC(ExInterlockedPopEntrySList),
	IMPORT_FUNC(ExInterlockedPushEntrySList),
	IMPORT_FUNC(ExAllocatePoolWithTag),
	IMPORT_FUNC(ExFreePool),
#ifdef __i386__
	IMPORT_FUNC(KefAcquireSpinLockAtDpcLevel),
	IMPORT_FUNC(KefReleaseSpinLockFromDpcLevel),
	IMPORT_FUNC(KeAcquireSpinLockRaiseToDpc),
#else
	/*
	 * For AMD64, we can get away with just mapping
	 * KeAcquireSpinLockRaiseToDpc() directly to KfAcquireSpinLock()
	 * because the calling conventions end up being the same.
	 * On i386, we have to be careful because KfAcquireSpinLock()
	 * is _fastcall but KeAcquireSpinLockRaiseToDpc() isn't.
	 */
	IMPORT_FUNC(KeAcquireSpinLockAtDpcLevel),
	IMPORT_FUNC(KeReleaseSpinLockFromDpcLevel),
	IMPORT_FUNC_MAP(KeAcquireSpinLockRaiseToDpc, KfAcquireSpinLock),
#endif
	IMPORT_FUNC_MAP(KeReleaseSpinLock, KfReleaseSpinLock),
	IMPORT_FUNC(InterlockedIncrement),
	IMPORT_FUNC(InterlockedDecrement),
	IMPORT_FUNC(ExInterlockedAddLargeStatistic),
	IMPORT_FUNC(IoAllocateMdl),
	IMPORT_FUNC(IoFreeMdl),
	IMPORT_FUNC(MmSizeOfMdl),
	IMPORT_FUNC(MmMapLockedPages),
	IMPORT_FUNC(MmMapLockedPagesSpecifyCache),
	IMPORT_FUNC(MmUnmapLockedPages),
	IMPORT_FUNC(MmBuildMdlForNonPagedPool),
	IMPORT_FUNC(KeInitializeSpinLock),
	IMPORT_FUNC(IoIsWdmVersionAvailable),
	IMPORT_FUNC(IoGetDeviceProperty),
	IMPORT_FUNC(KeInitializeMutex),
	IMPORT_FUNC(KeReleaseMutex),
	IMPORT_FUNC(KeReadStateMutex),
	IMPORT_FUNC(KeInitializeEvent),
	IMPORT_FUNC(KeSetEvent),
	IMPORT_FUNC(KeResetEvent),
	IMPORT_FUNC(KeClearEvent),
	IMPORT_FUNC(KeReadStateEvent),
	IMPORT_FUNC(KeInitializeTimer),
	IMPORT_FUNC(KeInitializeTimerEx),
	IMPORT_FUNC(KeSetTimer),
	IMPORT_FUNC(KeSetTimerEx),
	IMPORT_FUNC(KeCancelTimer),
	IMPORT_FUNC(KeReadStateTimer),
	IMPORT_FUNC(KeInitializeDpc),
	IMPORT_FUNC(KeInsertQueueDpc),
	IMPORT_FUNC(KeRemoveQueueDpc),
	IMPORT_FUNC(ObReferenceObjectByHandle),
	IMPORT_FUNC(ObfDereferenceObject),
	IMPORT_FUNC(ZwClose),
	IMPORT_FUNC(PsCreateSystemThread),
	IMPORT_FUNC(PsTerminateSystemThread),

	/*
	 * This last entry is a catch-all for any function we haven't
	 * implemented yet. The PE import list patching routine will
	 * use it for any function that doesn't have an explicit match
	 * in this table.
	 */

	{ NULL, (FUNC)dummy_ntoskrnl, NULL },

	/* End of list. */

	{ NULL, NULL, NULL }
};
