; /*
;  * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
;  * Use is subject to license terms.
;  */
;
; /*-
;  * Copyright (c) 2005
;  *      Bill Paul <wpaul@windriver.com>.  All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. All advertising materials mentioning features or use of this software
;  *    must display the following acknowledgement:
;  *      This product includes software developed by Bill Paul.
;  * 4. Neither the name of the author nor the names of any co-contributors
;  *    may be used to endorse or promote products derived from this software
;  *    without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
;  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;  * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
;  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
;  * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;  * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;  * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
;  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
;  * THE POSSIBILITY OF SUCH DAMAGE.
;  *
;  * The x86_64 callback routines were written and graciously submitted
;  * by Ville-Pertti Keinonen <will@exomi.com>.
;  *
;  * $FreeBSD: src/sys/compat/ndis/winx64_wrap.S,v 1.3 2005/02/18 03:22:37 wpaul Exp $
;  */
;
; #define ALIGN_TEXT      .p2align 4,090h /* 16-byte alignment, nop filled */
; #define CNAME(csym)             csym
; #define GEN_ENTRY(name)         ALIGN_TEXT; 	public	CNAME(name); \
;                                 .type CNAME(name),@function; CNAME(name):
; #define ENTRY(name)             GEN_ENTRY(name)
;
; /*
;  * Wrapper for handling up to 16 arguments. We can't really
;  * know how many arguments the caller will pass us. I'm taking an
;  * educated guess that we'll never get over 16. Handling too
;  * few arguments is bad. Handling too many is inefficient, but
;  * not fatal. If someone can think of a way to handle an arbitrary
;  * number of arguments with more elegant code, freel free to let
;  * me know.
;  *
;  * Standard amd64 calling conventions specify the following registers
;  * to be used for passing the first 6 arguments:
;  *
;  *   %rdi, %rsi, %rdx, %rcx, r8, r9
;  *
;  * Further arguments are passed on the stack (the 7th argument is
;  * located immediately after the return address).
;  *
;  * Windows x86_64 calling conventions only pass the first 4
;  * arguments in registers:
;  *
;  *   %rcx, %rdx, r8, r9
;  *
;  * Even when arguments are passed in registers, the stack must have
;  * space reserved for those arguments.  Thus the 5th argument (the
;  * first non-register argument) is placed 32 bytes after the return
;  * address.  Additionally, rdi and rsi must be preserved. (These
;  * two registers are not scratch registers in the standard convention.)
;  *
;  * Note that in this template, we load a contrived 64 bit address into
;  * r11 to represent our jump address. This is to guarantee that the
;  * assembler leaves enough room to patch in an absolute 64-bit address
;  * later. The idea behind this code is that we want to avoid having to
;  * manually create all the wrapper functions at compile time with
;  * a bunch of macros. This is doable, but a) messy and b) requires
;  * us to maintain two separate tables (one for the UNIX function
;  * pointers and another with the wrappers). This means I'd have to
;  * update two different tables each time I added a function.
;  *
;  * To avoid this, we create the wrappers at runtime instead. The
;  * image patch tables now contain two pointers: one two the normal
;  * routine, and a blank one for the wrapper. To construct a wrapper,
;  * we allocate some memory and copy the template function into it,
;  * then patch the function pointer for the routine we want to wrap
;  * into the newly created wrapper. The subr_pe module can then
;  * simply patch the wrapper routine into the jump table into the
;  * windows image. As a bonus, the wrapper pointer not only serves
;  * as the wrapper entry point address, it's also a data pointer
;  * that we can pass to free() later when we unload the module.
;  */
	.x86_64
	.xmm
_BSS	segment	para public '.bss'
_BSS	ends
_DATA	segment	para public '.data'
_DATA	ends
_TEXT	segment	para public '.text'

	public	x86_64_wrap
	public	x86_64_wrap_call
	public	x86_64_wrap_end
	public  x86_64_call1
	public  x86_64_call2
	public  x86_64_call3
	public  x86_64_call4
	public  x86_64_call5
	public  x86_64_call6
	public  x86_64_call7
	public  restore

common equ 1

    ifdef common

x86_64_wrap proc near
x86_64_wrap_call:
	mov	r10,0FF00FF00FF00FF00h
	jmp	qword ptr wantcomm[rip]
wantcomm dq	commoncode
x86_64_wrap_end:
;
;	This part of code need not be copied, easier to debug
;
commoncode :
	push	rbp
	mov	rbp,rsp
	and	rsp,-16
	sub	rsp,96		; allocate space on stack
	mov     96-8[rsp],rsi	; save rsi
	mov     96-16[rsp],rdi	; save rdi
	mov	rax,rcx		; temporarily save rcx in scratch
	lea	rsi,56+8[rbp]	; source == old stack top (stack+56)
	mov	rdi,rsp		; destination == new stack top
	mov	rcx,10		; count == 10 quadwords
	cld
	rep
	movsq			; copy old stack contents to new location
	mov	rdi,rax		; set up arg0 (rcx -> rdi)
	mov     rsi,rdx		; set up arg1 (rdx -> rsi)
	mov     rdx,r8		; set up arg2 (r8 -> rdx)
	mov     rcx,r9		; set up arg3 (r9 -> rcx)
	mov	r8,40+8[rbp]	; set up arg4 (stack+40 -> r8)
	mov	r9,48+8[rbp]	; set up arg5 (stack+48 -> r9)
	xor	rax,rax		; clear return value
	call	r10		; call routine
	mov	rdi,96-16[rsp]	; restore rdi
	mov	rsi,96-8[rsp]	; restore rsi
	leave
	ret
x86_64_wrap endp

    else

x86_64_wrap proc near
	push	rbp
	mov	rbp,rsp
	and	rsp,-16
	sub	rsp,96		; allocate space on stack
	mov     96-8[rsp],rsi	; save rsi
	mov     96-16[rsp],rdi	; save rdi
	mov	r10,rcx		; temporarily save rcx in scratch
	lea	rsi,56+8[rbp]	; source == old stack top (stack+56)
	mov	rdi,rsp		; destination == new stack top
	mov	rcx,10		; count == 10 quadwords
	cld
	rep
	movsq			; copy old stack contents to new location
	mov	rdi,r10		; set up arg0 (rcx -> rdi)
	mov     rsi,rdx		; set up arg1 (rdx -> rsi)
	mov     rdx,r8		; set up arg2 (r8 -> rdx)
	mov     rcx,r9		; set up arg3 (r9 -> rcx)
	mov	r8,40+8[rbp]	; set up arg4 (stack+40 -> r8)
	mov	r9,48+8[rbp]	; set up arg5 (stack+48 -> r9)
	xor	rax,rax		; clear return value
x86_64_wrap_call:
	mov	r10,0FF00FF00FF00FF00h
	call	r10		; call routine
	mov	rdi,96-16[rsp]	; restore rdi
	mov	rsi,96-8[rsp]	; restore rsi
	leave
	ret
x86_64_wrap_end:
x86_64_wrap endp

    endif


; /*
;  * Functions for invoking x86_64 callbacks.  In each case, the first
;  * argument is a pointer to the function.
;  * JPA : 32 bytes HAVE TO be reserved in the stack, even when there are
;  * no arguments : this space may be used for local variables
;  */

x86_64_call1 proc near
	sub	rsp,40
	mov	rcx,rsi
	call	rdi
	add	rsp,40
	ret
x86_64_call1 endp

x86_64_call2 proc near
	sub	rsp,40
	mov	rcx,rsi
;			/* rdx is already correct */
	call	rdi
	add	rsp,40
	ret
x86_64_call2 endp

x86_64_call3 proc near
	sub	rsp,40
	mov	r8,rcx
	mov	rcx,rsi
	call	rdi
	add	rsp,40
	ret
x86_64_call3 endp

x86_64_call4 proc near
	sub	rsp,40
	mov	r9,r8
	mov	r8,rcx
	mov	rcx,rsi
	call	rdi
	add	rsp,40
	ret
x86_64_call4 endp

x86_64_call5 proc near
	sub	rsp,40
	mov	32[rsp],r9
	mov	r9,r8
	mov	r8,rcx
	mov	rcx,rsi
	call	rdi
	add	rsp,40
	ret
x86_64_call5 endp

x86_64_call6 proc near
	sub	rsp,56
	mov	rax,56+8[rsp]
	mov	32[rsp],r9
	mov	40[rsp],rax
	mov	r9,r8
	mov	r8,rcx
	mov	rcx,rsi
	call	rdi
	add	rsp,56
	ret
x86_64_call6 endp

x86_64_call7 proc near
	sub	rsp,56
	mov	rax,56+8[rsp]
	mov	32[rsp],r9
	mov	40[rsp],rax
	mov	rax,56+16[rsp]
	mov	48[rsp],rax
	mov	r9,r8
	mov	r8,rcx
	mov	rcx,rsi
	call	rdi
	add	rsp,56
	ret
x86_64_call7 endp
;
;	restore() is useful to force restoring the registers rdi and rsi
;
restore proc near
	ret
restore endp

_TEXT	ends
	end
